c
c
      subroutine preact(n)
c
c  this routine loads data into the active
c  element common block.  That data is then
c  given to the user to work with at will
c
      implicit none
c
      include 'domain.inc'
      include 'active.inc'
      include 'ibp.inc'
c
      integer n
c
      integer i,k
      double precision h2
c
c  check to see if entity or a portion of 
c    that entity is active or not
c
      k = 0
      h2 = h*h
      do i = activ3(n), activ4(n)
        k = k+1
        ibptx(k)  = xb(2*i-1)+xmin
        ibpty(k)  = xb(2*i)+ymin
        vibptx(k) = usb1(i)*h2
        vibpty(k) = usb2(i)*h2
        tethx(k)  = xteth(i)+xmin
        tethy(k)  = yteth(i)+ymin
        stif(k)   = swp(i)
        restd(k)  = dlp(i)
        bndstf(k) = sp(i)
        resta(k)  = csp(i)
        tstiff(k) = sthp(i)
      enddo
      nels = k
      elab = activ2(n)

      return
      end     
c
c---------------------------------------------------------------
c
      subroutine afact(n)
c
c  this routine reloads data after the active 
c  motion.  The data is passed back to the 
c  driver.
c
      implicit none
c
      include 'domain.inc'
      include 'active.inc'
      include 'ibp.inc'
c
      integer n
c
      integer i,k
c
c  check to see if entity or a portion of 
c    that entity is active or not
c
      k = 0
      do i = activ3(n), activ4(n)
        k = k+1
        xb(2*i-1) = ibptx(k)-xmin
        xb(2*i)   = ibpty(k)-ymin
        xteth(i)  = tethx(k)-xmin
        yteth(i)  = tethy(k)-ymin
        swp(i)    = stif(k)
        dlp(i)    = restd(k)
        sp(i)     = bndstf(k)
        csp(i)    = resta(k)
        sthp(i)   = tstiff(k)
      enddo
c
      return
      end     
c
c------------------------------------------------------
c
      subroutine chksrc(ntsrc,nsrc)
c
c  global variables
c
      implicit none
c
      include 'domain.inc'
      include 'sink.inc'
c
      integer ntsrc,nsrc
c
      integer j
      double precision sumsrc
c
c  sum the src variables
c
      sumsrc = 0.0d0
      do j = 1, ntsrc
         sumsrc = sumsrc + src(j)
      enddo
c
      if ( sumsrc.ge.1.0d-6 ) then
        write (6,*) ' '
        write (6,*) 'Error:'
        write (6,*) 'Source/sink do not sum to zero'
        write (6,*) 'Program terminating prematurely'
        write (6,*) ' '
        stop
      elseif ( nsrc.lt.ntsrc ) then
        write (6,*) ' '
        write (6,*) 'Error:'
        write (6,*) 'More source and sinks than specified'
        write (6,*) 'Program terminating prematurely'
        write (6,*) ' '
        stop
      endif
c
      do j=1,ntsrc
        xsourc(j) = xsourc(j) - xmin
        ysourc(j) = ysourc(j) - ymin
        if ( xsourc(j).gt.xlength .or. xsourc(j).lt.0.0d0 .or.
     x       ysourc(j).gt.ylength .or. ysourc(j).lt.0.0d0) then
          write (6,*) ' '
          write (6,*) 'Error:'
          write (6,*) 'source or sinks outside the domain'
          write (6,*) 'Program terminating prematurely'
          write (6,*) ' '
          stop
        endif
      enddo
c
      return
      end
c
c------------------------------------------------------------
c
      subroutine sprsrc(ntsrc,nxp2,nyp1,psi)
c
      implicit none
c
c  global include
c
      include 'domain.inc'
      include 'sink.inc'
c
      integer jx,jy
      common/sprdmap/ jx(16),jy(16)
c
c  global variables
c
      integer ntsrc,nxp2,nyp1
      double precision psi(0:nxp2,0:nyp1)
c
c  local variables
c 
      integer jmarr,kmarr
      double precision f1,xsprd,ysprd
      double precision xwt(4)
      double precision ywt(4)
      double precision wt(16)
c
c  local variables
c
      integer i,ii
      integer j,jmp,jxii,jyii
      integer k,kmp,n,nx,ny
c
      double precision wt1x,wt1y,wt2x,wt2y
      double precision argx,argy
      double precision cy1,cy2,sy1,sy2
      double precision cx1,cx2,sx1,sx2
      double precision cx,cy,rh,sx,sy,xper100,yper100
c
      nx   = nxp2 - 2
      ny   = nyp1 - 1
c
c once only ccccccccccccccccccccccccccccccccccccccccc
c
      xper100 = xlength*100.d0
      yper100 = ylength*100.d0
      rh   = 1.d0/h
      wt1x = 2.0d0*datan(1.0d0)*rh
      wt1y = 2.0d0*datan(1.0d0)*rh
      wt2x = .25d0*rh
      wt2y = .25d0*rh
      cy1  = dcos(wt1y*h)
      sy1  = dsin(wt1y*h)
      cy2  = dcos(2.d0*wt1y*h)
      sy2  = dsin(2.d0*wt1y*h)
      cx1  = dcos(wt1x*h)
      sx1  = dsin(wt1x*h)
      cx2  = dcos(2.d0*wt1x*h)
      sx2  = dsin(2.d0*wt1x*h)
c
c once only done ccccccccccccccccccccccccccccccccccc
c
      do j=0,nx
        do k=0,ny
          psi(j,k)=0.0D+00
        enddo
      enddo
c
c loop over the sources and sinks
c
      do n=1,ntsrc
         xsprd=xsourc(n)+xper100
         ysprd=ysourc(n)+yper100
         jmarr=idint(xsprd*rh)
         kmarr=idint(ysprd*rh)
         f1 = dlibp*src(n)
c
c  compute weights
c
         argy=wt1y*(ysprd-dble(idint(ysprd*rh))*h)
         cy = dcos(argy)
         sy = dsin(argy)
         ywt(1)=wt2y*(1.0d0 + cy*cy1 - sy*sy1)
         ywt(2)=wt2y*(1.0d0 + cy)
         ywt(3)=wt2y*(1.0d0 + cy*cy1 + sy*sy1)
         ywt(4)=wt2y*(1.0d0 + cy*cy2 + sy*sy2)
         argx=wt1x*(xsprd-dble(idint(xsprd*rh))*h)
         cx = dcos(argx)
         sx = dsin(argx)
         xwt(1)=wt2x*(1.0d0 + cx*cx1 - sx*sx1)
         xwt(2)=wt2x*(1.0d0 + cx)
         xwt(3)=wt2x*(1.0d0 + cx*cx1 + sx*sx1)
         xwt(4)=wt2x*(1.0d0 + cx*cx2 + sx*sx2)
c
c  compute the spread out source
c
       do ii=1,16
          jxii = jx(ii)  
          jyii = jy(ii)  
          wt(ii) = xwt(jxii)*ywt(jyii)
          jmp=mod((jmarr-2+jxii)+100*nx,nx)
          kmp=mod((kmarr-2+jyii)+100*ny,ny)
          psi(jmp,kmp)=psi(jmp,kmp)+wt(ii)*f1
        enddo
      enddo
c
c extend periodically
c
      do j=0,ny
         psi(nx,j)=psi(0,j)
      enddo

      do i=0,nx
         psi(i,ny)=psi(i,0)
      enddo  

      psi(nx+1,nyp1)=psi(1,1)
c
      return
      end
