c
      subroutine ibisdrv(nxp2,nyp1,u,f,fout,p,psi,bzero,
     *    sinsq,z,b,c,r,work1,work2,yh,yp,
     *    ments,mlink,mlkpt,mpts,mwhere,
     *    nring,np,npdx,npdy,nsrc,
     *    alpha,nxx,nyy)
c
c this program is the driver routine for IBIS simulations
c get global parameters
c
      implicit none
c
c get local parameter and global commons
c
      include 'ibp.inc'
      include 'active.inc'
      include 'links.inc'
      include 'domain.inc'
      include 'graph.inc'
      include 'fluid.inc'
      include 'plat.inc'
      include 'sink.inc'
c
c  global variables 
c
      integer cpflag,last1,last2,nring
      integer mpts,np,nsrc,ntsrc,nxp2,nxx,nyp1,nyy
      integer ments,mlink,mlkpt,mwhere,npdx,npdy
c
      byte alpha(nxx,nyy)
c
      double precision sfluid(3), a
c
      integer jx,jy
      common/sprdmap/ jx(16),jy(16)
c
      double precision u(0:nxp2,0:nyp1,1:2)
      double precision f(0:nxp2,0:nyp1,1:2)
      double precision p(0:nxp2,0:nyp1)
      double precision psi(0:nxp2,0:nyp1)
      double precision fout(0:nxp2,0:nyp1,1:2)
c
      double precision bzero(0:nxp2,0:nyp1)
      double precision sinsq(0:nxp2,0:nyp1)
      double precision z(0:nxp2,0:nyp1)
c
      double precision b(0:nxp2,0:nyp1)
      double precision c(0:nxp2,0:nyp1)
      double precision r(0:nxp2,0:nyp1)
      double precision work1(0:nxp2,0:nyp1)
      double precision work2(0:nxp2,0:nyp1)
      double precision yh(0:nxp2,0:nyp1)
      double precision yp(0:nxp2,0:nyp1)
c
c  command line variables
c
      character*1 cmdline
      integer*4 iargc,ncmdline
c
c  local variables
c
      integer i,j,k,idebug(0:4),rsflag,nx,ny
c
      data idebug / 0,0,0,0,0 /
c
      data iff,iffp,iplat /0,0,0/
c
      write(6,*)' '
      write(6,*)'IBIS numerical solver'
      write(6,*)'Version 1.0'
      write(6,*)' '
c
      nx = nxp2-2
      ny = nyp1-1
      ment = ments
c
c initialize immersed boundary points and fluid velocities
c
      call init(nxp2,nyp1,f,u,sinsq,bzero,sfluid,a,z,rsflag,
     x    mlink,mlkpt,mpts,mwhere,np,npdx,npdy,nring,nxx,nyy)
c
      time = ntstart*dlt
      nt = 0
c
c initialize array for spread/interp
c
      call sprdinit
c
c plot the initial data
c
      if (rsflag .eq. 0) then
        call cplot(ccount,alpha,nxx,nyy,mlink,np)
        call output(lcount,nxp2,nyp1,f,u,mlink,np)
      endif
c
c check command line for stop or debug options
c
      ncmdline = iargc()
      if (ncmdline .eq. 1) then
        call getarg(1,cmdline)
        if ( cmdline .eq. 'p' ) then
          stop
        elseif ( cmdline .eq. 'd' ) then
          idebug(0) = 1
          open (unit=12,file='ibisdat/debug.dat',status='old')
          do j=1,2
            read (12,*) idebug(j)
          enddo
          close(12)
        endif
      endif
c
c print some junk out on stdout
c
      write(6,*)' nent = ',nent, ' npts = ',npts
      call prout (ntstart,fluxx,fluxy,vmax,(dlt*vmax)-h,time)
c
cccccccccccccccccccccccc main loop ccccccccccccccccccccc
c
      do nt = ntstart+1, ntmax
c
c
        time = time + dlt
c
c compute centers of mass for use in create
c  and insert new rings if necessary
c
        if (iplat .eq. 1) then
          call cmass
          if(fluxx.gt.(0.0D+00))then
            call create(nring)
          endif
        endif
c
c active control of the immersed boundary points
c   pass stuff to user.
c
        do i=1,ment
          if ( activ1(i) .eq. 1 ) then
            call preact(i)
            call actmov
            call afact(i)
          endif
        enddo
c
c perform map of actual entities
c evaluate the force density at immersed boundary points
c
        call map
        call grad
        call gradlink
        if (idebug(0) .eq. 1 ) 
     x     call pdebug (idebug)
c
c calculate entity force density (force/length)
cdir$ ivdep
c
        do j=1,npts
          fs(2*j-1)=-gfrc1(j)
          fs(2*j)=-gfrc2(j)
        enddo
c
c spread this to grid to get fluid force density
c add background force to drive the flow
c
        call spread(npts,nxp2,nyp1,f)
        call fadd(nxp2,nyp1,f)
c
c save force density for output if appropriate
c extend fout periodically to 0:nx 0:nyp1
c
        if(mod(nt,lfreq).eq.0)then
          do j=1,ny
            do i=1,nx
              fout(i,j,1)=f(i,j,1)
              fout(i,j,2)=f(i,j,2)
            enddo
          enddo
          call period (nxp2,nyp1,fout(0,0,1))
          call period (nxp2,nyp1,fout(0,0,2))
        endif
c
c get the fluid sources and sinks, check for volume,
c   number of sources and sinks, and fix particles.
c
        call getsrc (ntsrc)
        if ( ntsrc.gt.0 ) then
          call chksrc(ntsrc,nsrc)
          call sprsrc(ntsrc,nxp2,nyp1,psi)
          call prtsrc(ntsrc,np)
        endif
c
c solve ns equations
c extend u periodically to 0:nx+1  0:nyp1
c compute maximum speed and fluxes
c
        call navs(nxp2,nyp1,u,f,p,psi,
     *        bzero,sinsq,z,
     *        b,c,r,work1,work2,yh,yp,
     *        a,dlt,h,ro,sfluid)
        call period (nxp2,nyp1,u(0,0,1))
        call period (nxp2,nyp1,u(0,0,2))
        call upflux (nxp2,nyp1,u)
c
c move entities by convection - add centers of the
c  entities to the point list and add the fluid marker
c  particles to the point list.  Then remove after moving.
c
        call addmap(xpt,ypt,mpts,centx,centy,nent,npts)
        call addmap(xpt,ypt,mpts,px,py,np,npts+nent)
        call xmove (npts+nent+np,nxp2,nyp1,u)
        call delmap(xpt,ypt,mpts,px,py,np,npts+nent)
        call delmap(xpt,ypt,mpts,centx,centy,nent,npts)
c
c move fluid marker particles via random walk
c evaluate the fluid marker positions
c
        if ( pwalk .gt. 0.0d0 ) call partwalk(np)
        call part(np,npdx,npdy,nxx,nyy)
c
c reverse map of coordinates
c
        call revmap
c
c check on the platelet progress
c
        if (iplat .eq. 1) then
          call chkplat(cpflag,last1,last2,mlink,nring)
        endif
c
c remove the forces on objects have left the domain
c  ADD SOMETHING HERE

c
c print data to screen...
c
        call prout (nt,fluxx,fluxy,vmax,(dlt*vmax)-h,time)
c
c  shut down if failure is about to happen
c
        if( abs((dlt*vmax)-h) .ge. 1.0d0 ) then
          call ohnoo
          call cplot(ccount+1,alpha,nxx,nyy,mlink,np)
          call output(lcount+1,nxp2,nyp1,fout,u,mlink,np)
          stop
        endif
c
c  output the color plot at correct times
c
        if (mod(nt,cfreq).eq.0) then
          ccount = ccount + 1
          call cplot(ccount,alpha,nxx,nyy,mlink,np)
        endif
c
c  output line drawing and restart files
c
        if(mod(nt,lfreq).eq.0)then
          lcount = lcount+1
          call output(lcount,nxp2,nyp1,fout,u,mlink,np)
          call rsdata(13,nxp2,nyp1,u,mlink,np)
        endif
c
c  debug as needed - the immersed boundary points, forces
c
c
cccccccccccccccccccccccc end main loop ccccccccccccccccc
c
      enddo
      return
 1008 format(7(1x,e15.7))
 9990 format(8f15.7)
 9991 format(4f10.3,6e11.3)
      end
c
c
      subroutine chkplat(cpflag,last1,last2,mlink,nring)
c
c  this routine gathers a cluster of subroutine calls for
c    platlets only
c
c get global parameters
c
      implicit none
c
c global commons
c
      include 'plat.inc'
      include 'links.inc'
c
c global variables
c
      integer cpflag,last1,last2,mlink,nring
c
c compute center of mass for each platelet in domain
c   and move single rings by random walk
c
      call cmass
      call rw
c
c make new links between rings and walls
c
      cpflag=0
      call adhere(cpflag)
c
      if(nlink.gt.mlink)then
          write(6,*)' NLINK GT MLINK'
          stop
      endif
c
c make new links between rings
c
      call cohere(cpflag)
c
      if(nlink.gt.mlink)then
          write(6,*)' NLINK GT MLINK'
          stop
      endif
c
c remove rings downstream (and break links)
c
      call remove(cpflag,nring)
c
c if linked pairs have changed, compute new
c list and number of wall-bound platelets
c
      if(cpflag.eq.1)then
        call wallagg(last1,last2)
      endif
c
c compute height of wall-bound aggregates
c
      call height(last1,last2)
c
      return
      end

c
c
      subroutine preact(n)
c
c  this routine loads data into the active
c  element common block.  That data is then
c  given to the user to work with at will
c
      implicit none
c
      include 'domain.inc'
      include 'active.inc'
      include 'ibp.inc'
c
      integer n
c
      integer i,k
      double precision h2
c
c  check to see if entity or a portion of 
c    that entity is active or not
c
      k = 0
      h2 = h*h
      do i = activ3(n), activ4(n)
        k = k+1
        ibptx(k)  = xb(2*i-1)+xmin
        ibpty(k)  = xb(2*i)+ymin
        vibptx(k) = usb1(i)*h2
        vibpty(k) = usb2(i)*h2
        tethx(k)  = xteth(i)+xmin
        tethy(k)  = yteth(i)+ymin
        stif(k)   = swp(i)
        restd(k)  = dlp(i)
        bndstf(k) = sp(i)
        resta(k)  = csp(i)
        tstiff(k) = sthp(i)
      enddo
      nels = k
      elab = activ2(n)

      return
      end     
c
c---------------------------------------------------------------
c
      subroutine afact(n)
c
c  this routine reloads data after the active 
c  motion.  The data is passed back to the 
c  driver.
c
      implicit none
c
      include 'domain.inc'
      include 'active.inc'
      include 'ibp.inc'
c
      integer n
c
      integer i,k
c
c  check to see if entity or a portion of 
c    that entity is active or not
c
      k = 0
      do i = activ3(n), activ4(n)
        k = k+1
        xb(2*i-1) = ibptx(k)-xmin
        xb(2*i)   = ibpty(k)-ymin
        xteth(i)  = tethx(k)-xmin
        yteth(i)  = tethy(k)-ymin
        swp(i)    = stif(k)
        dlp(i)    = restd(k)
        sp(i)     = bndstf(k)
        csp(i)    = resta(k)
        sthp(i)   = tstiff(k)
      enddo
c
      return
      end     
c
c------------------------------------------------------
c
      subroutine chksrc(ntsrc,nsrc)
c
c  global variables
c
      implicit none
c
      include 'domain.inc'
      include 'sink.inc'
c
      integer ntsrc,nsrc
c
      integer j
      double precision sumsrc
c
c  sum the src variables
c
      sumsrc = 0.0d0
      do j = 1, ntsrc
         sumsrc = sumsrc + src(j)
      enddo
c
      if ( sumsrc.ge.1.0d-6 ) then
        write (6,*) ' '
        write (6,*) 'Error:'
        write (6,*) 'Source/sink do not sum to zero'
        write (6,*) 'Program terminating prematurely'
        write (6,*) ' '
        stop
      elseif ( nsrc.lt.ntsrc ) then
        write (6,*) ' '
        write (6,*) 'Error:'
        write (6,*) 'More source and sinks than specified'
        write (6,*) 'Program terminating prematurely'
        write (6,*) ' '
        stop
      endif
c
      do j=1,ntsrc
        xsourc(j) = xsourc(j) - xmin
        ysourc(j) = ysourc(j) - ymin
        if ( xsourc(j).gt.xlength .or. xsourc(j).lt.0.0d0 .or.
     x       ysourc(j).gt.ylength .or. ysourc(j).lt.0.0d0) then
          write (6,*) ' '
          write (6,*) 'Error:'
          write (6,*) 'source or sinks outside the domain'
          write (6,*) 'Program terminating prematurely'
          write (6,*) ' '
          stop
        endif
      enddo
c
      return
      end
c
c------------------------------------------------------------
c
      subroutine sprsrc(ntsrc,nxp2,nyp1,psi)
c
      implicit none
c
c  global include
c
      include 'domain.inc'
      include 'sink.inc'
c
      integer jx,jy
      common/sprdmap/ jx(16),jy(16)
c
c  global variables
c
      integer ntsrc,nxp2,nyp1
      double precision psi(0:nxp2,0:nyp1)
c
c  local variables
c 
      integer jmarr,kmarr
      double precision f1,xsprd,ysprd
      double precision xwt(4)
      double precision ywt(4)
      double precision wt(16)
c
c  local variables
c
      integer i,ii
      integer j,jmp,jxii,jyii
      integer k,kmp,n,nx,ny
c
      double precision wt1x,wt1y,wt2x,wt2y
      double precision argx,argy
      double precision cy1,cy2,sy1,sy2
      double precision cx1,cx2,sx1,sx2
      double precision cx,cy,rh,sx,sy,xper100,yper100
c
      nx   = nxp2 - 2
      ny   = nyp1 - 1
c
c once only ccccccccccccccccccccccccccccccccccccccccc
c
      xper100 = xlength*100.d0
      yper100 = ylength*100.d0
      rh   = 1.d0/h
      wt1x = 2.0d0*datan(1.0d0)*rh
      wt1y = 2.0d0*datan(1.0d0)*rh
      wt2x = .25d0*rh
      wt2y = .25d0*rh
      cy1  = dcos(wt1y*h)
      sy1  = dsin(wt1y*h)
      cy2  = dcos(2.d0*wt1y*h)
      sy2  = dsin(2.d0*wt1y*h)
      cx1  = dcos(wt1x*h)
      sx1  = dsin(wt1x*h)
      cx2  = dcos(2.d0*wt1x*h)
      sx2  = dsin(2.d0*wt1x*h)
c
c once only done ccccccccccccccccccccccccccccccccccc
c
      do j=0,nx
        do k=0,ny
          psi(j,k)=0.0D+00
        enddo
      enddo
c
c loop over the sources and sinks
c
      do n=1,ntsrc
         xsprd=xsourc(n)+xper100
         ysprd=ysourc(n)+yper100
         jmarr=idint(xsprd*rh)
         kmarr=idint(ysprd*rh)
         f1 = dlibp*src(n)
c
c  compute weights
c
         argy=wt1y*(ysprd-dble(idint(ysprd*rh))*h)
         cy = dcos(argy)
         sy = dsin(argy)
         ywt(1)=wt2y*(1.0d0 + cy*cy1 - sy*sy1)
         ywt(2)=wt2y*(1.0d0 + cy)
         ywt(3)=wt2y*(1.0d0 + cy*cy1 + sy*sy1)
         ywt(4)=wt2y*(1.0d0 + cy*cy2 + sy*sy2)
         argx=wt1x*(xsprd-dble(idint(xsprd*rh))*h)
         cx = dcos(argx)
         sx = dsin(argx)
         xwt(1)=wt2x*(1.0d0 + cx*cx1 - sx*sx1)
         xwt(2)=wt2x*(1.0d0 + cx)
         xwt(3)=wt2x*(1.0d0 + cx*cx1 + sx*sx1)
         xwt(4)=wt2x*(1.0d0 + cx*cx2 + sx*sx2)
c
c  compute the spread out source
c
       do ii=1,16
          jxii = jx(ii)  
          jyii = jy(ii)  
          wt(ii) = xwt(jxii)*ywt(jyii)
          jmp=mod((jmarr-2+jxii)+100*nx,nx)
          kmp=mod((kmarr-2+jyii)+100*ny,ny)
          psi(jmp,kmp)=psi(jmp,kmp)+wt(ii)*f1
        enddo
      enddo
c
c extend periodically
c
      do j=0,ny
         psi(nx,j)=psi(0,j)
      enddo

      do i=0,nx
         psi(i,ny)=psi(i,0)
      enddo  

      psi(nx+1,nyp1)=psi(1,1)
c
      return
      end
c
      subroutine sprdinit
c
c  index map of 4x4 square to length 16 one dimensional array.
c  interpolation spread
c
      integer jx,jy
      common/sprdmap/ jx(16),jy(16)
c
      integer i,ix,iy
c
      do 10 i=1,16
         iy = (i - 1)/4
         ix = (i - 4*iy -1)
         jy(i)=iy+1
         jx(i)=ix+1
   10 continue
      return
      end
c
c------------------------------------------------------------
c
c spread5 -- modified from spread4 11/9/98
c
      subroutine spread(npts,nxp2,nyp1,f)
c
      implicit none
c
c  global include
c
      include 'domain.inc'

      integer jx,jy
      common/sprdmap/ jx(16),jy(16)
c
c  global variables
c
      integer npts,nxp2,nyp1
      double precision f(0:nxp2,0:nyp1,2)
c
c  local variables
c 
      integer jmarr,kmarr
      double precision f1,f2,xsprd,ysprd
      double precision xwt(4)
      double precision ywt(4)
      double precision wt(16)
c
c  body force globals
c
      double precision xpt(1)
      common /enxpt/ xpt

      double precision ypt(1)
      common /enypt/ ypt
c
      double precision fs(1)
      common /sfs/ fs
c
c  local variables
c
      integer i,ii
      integer j,jmp,jxii,jyii
      integer k,kmp,n,nx,ny
c
      double precision wt1x,wt1y,wt2x,wt2y
      double precision argx,argy
      double precision cy1,cy2,sy1,sy2
      double precision cx1,cx2,sx1,sx2
      double precision cx,cy,rh,sx,sy,xper100,yper100
c
      nx   = nxp2 - 2
      ny   = nyp1 - 1
c
c once only ccccccccccccccccccccccccccccccccccccccccc
c
      xper100 = xlength*100.d0
      yper100 = ylength*100.d0
      rh   = 1.d0/h
      wt1x = 2.0d0*datan(1.0d0)*rh
      wt1y = 2.0d0*datan(1.0d0)*rh
      wt2x = .25d0*rh
      wt2y = .25d0*rh
      cy1  = dcos(wt1y*h)
      sy1  = dsin(wt1y*h)
      cy2  = dcos(2.d0*wt1y*h)
      sy2  = dsin(2.d0*wt1y*h)
      cx1  = dcos(wt1x*h)
      sx1  = dsin(wt1x*h)
      cx2  = dcos(2.d0*wt1x*h)
      sx2  = dsin(2.d0*wt1x*h)
c
c once only done ccccccccccccccccccccccccccccccccccc
c
      do 1 j=0,nx
        do 1 k=0,ny
          f(j,k,1)=0.0D+00
          f(j,k,2)=0.0D+00
 1    continue
c
      do 100 n=1,npts
         xsprd=xpt(n)+xper100
         ysprd=ypt(n)+yper100
         jmarr=idint(xsprd*rh)
         kmarr=idint(ysprd*rh)
         f1 = dlibp*fs(2*n-1)
         f2 = dlibp*fs(2*n  )
c
c  compute weights
c
         argy=wt1y*(ysprd-dble(idint(ysprd*rh))*h)
         cy = dcos(argy)
         sy = dsin(argy)
         ywt(1)=wt2y*(1.0d0 + cy*cy1 - sy*sy1)
         ywt(2)=wt2y*(1.0d0 + cy)
         ywt(3)=wt2y*(1.0d0 + cy*cy1 + sy*sy1)
         ywt(4)=wt2y*(1.0d0 + cy*cy2 + sy*sy2)
         argx=wt1x*(xsprd-dble(idint(xsprd*rh))*h)
         cx = dcos(argx)
         sx = dsin(argx)
         xwt(1)=wt2x*(1.0d0 + cx*cx1 - sx*sx1)
         xwt(2)=wt2x*(1.0d0 + cx)
         xwt(3)=wt2x*(1.0d0 + cx*cx1 + sx*sx1)
         xwt(4)=wt2x*(1.0d0 + cx*cx2 + sx*sx2)
c
         do 20 ii=1,16
            jxii = jx(ii)  
            jyii = jy(ii)  
            wt(ii) = xwt(jxii)*ywt(jyii)
            jmp=mod((jmarr-2+jxii)+100*nx,nx)
            kmp=mod((kmarr-2+jyii)+100*ny,ny)
            f(jmp,kmp,1)=f(jmp,kmp,1)+wt(ii)*f1
            f(jmp,kmp,2)=f(jmp,kmp,2)+wt(ii)*f2
 20      continue
 100  continue
c
c extend periodically
c
      do 160 j=0,ny
         f(nx,j,1)=f(0,j,1)
         f(nx,j,2)=f(0,j,2)
 160  continue
      do 170 i=0,nx
         f(i,ny,1)=f(i,0,1)
         f(i,ny,2)=f(i,0,2)
 170  continue
      return
      end

c
c------------------------------------------------------------
c
c xmove4 - modified from xmove3  11/8/98
c
      subroutine xmove (npts,nxp2,nyp1,u)
c
      implicit none
c
c  global include
c
      include 'domain.inc'

      integer jx,jy
      common/sprdmap/ jx(16),jy(16)
c
c  global variables
c
      integer npts,nxp2,nyp1
      double precision u(0:nxp2,0:nyp1,2)
c
c
c  body force globals
c
      double precision xpt(1)
      common /enxpt/ xpt

      double precision ypt(1)
      common /enypt/ ypt

      double precision us1(1)
      common /usv1/ us1

      double precision us2(1)
      common /usv2/ us2

c
c  local variables
c
c
c local variables
c
      double precision xwt(4)
      double precision ywt(4)
      double precision wt(16)
      double precision xsprd,ysprd
c
      integer jmarr,kmarr
      integer ii,jmp,jxii,jyii
      integer kmp,nx,ny
      integer n
c
      double precision wt1x,wt1y,wt2x,wt2y
      double precision rh,h2
      double precision argx,argy
      double precision cy1,cy2,sy1,sy2
      double precision cx1,cx2,sx1,sx2
      double precision cx,cy,sx,sy,xper100,yper100
c
      nx   = nxp2 - 2
      ny   = nyp1 - 1
c
c once only ccccccccccccccccccccccccccccccccccccccccc
c
      xper100 = xlength*100.d0
      yper100 = ylength*100.d0
      rh   = 1.d0/h
      wt1x = 2.0d0*datan(1.0d0)*rh
      wt1y = 2.0d0*datan(1.0d0)*rh
      wt2x = .25d0*rh
      wt2y = .25d0*rh
      h2   = h*h
      cy1  = dcos(wt1y*h)
      sy1  = dsin(wt1y*h)
      cy2  = dcos(2.d0*wt1y*h)
      sy2  = dsin(2.d0*wt1y*h)
      cx1  = dcos(wt1x*h)
      sx1  = dsin(wt1x*h)
      cx2  = dcos(2.d0*wt1x*h)
      sx2  = dsin(2.d0*wt1x*h)
c
c once only done ccccccccccccccccccccccccccccccccccc
c
      do 100 n=1,npts
c
c add large multiple of period to x,y so x>0 and y>0
c to avoid testing
c
         xsprd=xpt(n)+xper100
         ysprd=ypt(n)+yper100
         jmarr=idint(xsprd*rh)
         kmarr=idint(ysprd*rh)
c
c  compute weights
c
         argy=wt1y*(ysprd-dble(idint(ysprd*rh))*h)
         cy = dcos(argy)
         sy = dsin(argy)
         ywt(1)=wt2y*(1.0d0 + cy*cy1 - sy*sy1)
         ywt(2)=wt2y*(1.0d0 + cy)
         ywt(3)=wt2y*(1.0d0 + cy*cy1 + sy*sy1)
         ywt(4)=wt2y*(1.0d0 + cy*cy2 + sy*sy2)
         argx=wt1x*(xsprd-dble(idint(xsprd*rh))*h)
         cx = dcos(argx)
         sx = dsin(argx)
         xwt(1)=wt2x*(1.0d0 + cx*cx1 - sx*sx1)
         xwt(2)=wt2x*(1.0d0 + cx)
         xwt(3)=wt2x*(1.0d0 + cx*cx1 + sx*sx1)
         xwt(4)=wt2x*(1.0d0 + cx*cx2 + sx*sx2)
c
        us1(n)=0.d0
        us2(n)=0.d0
        do 20 ii=1,16
           jxii = jx(ii)  
           jyii = jy(ii)  
           wt(ii) = xwt(jxii)*ywt(jyii)
           jmp=mod((jmarr-2+jxii)+100*nx,nx)
           kmp=mod((kmarr-2+jyii)+100*ny,ny)
           us1(n)=us1(n)+wt(ii)*u(jmp,kmp,1)
           us2(n)=us2(n)+wt(ii)*u(jmp,kmp,2)
 20     continue
        xpt(n)=xpt(n)+dlt*us1(n)*h2
        ypt(n)=ypt(n)+dlt*us2(n)*h2
 100  continue
c
      return
      end

c
      subroutine grad
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'domain.inc'
      include 'force.inc'
c
c  local variables
c
      integer i,iend,ione,jent,icat
      double precision dist,sm
c
c  initiate the gradients
c
      do 10 i=1,npts
         gfrc1(i)=0.0D+00
         gfrc2(i)=0.0D+00
 10   continue
c
c  calculate the forces
c
      do 50 jent=1,ment
         ione = first(jent)
         iend = last(jent)
         icat = cat(jent)
         if (icat.eq.0) goto 50
c
c   use work vectors
c
         do 15 i=ione,iend
           xwrk(i)   = xpt(i)
           ywrk(i)   = ypt(i)
           dlwrk(i)  = dl(i)
           swwrk(i)  = sw(i)
           cswrk(i)  = cs(i)
           swrk(i)   = s(i)
 15      continue
c
c  do different things for different types of things
c    affects the ends of the work arrays
c
c    open polygons - make sure the distance function
c      is ok
c
         if ( icat .eq. -3 ) then
           xwrk(ione-2)  = (2.0d0*xpt(ione+1)) - xpt(ione+2)
           xwrk(ione-1)  = (2.0d0*xpt(ione)) - xpt(ione+1)
           xwrk(iend+1)  = (2.0d0*xpt(iend)) - xpt(iend-1)
           xwrk(iend+2)  = (2.0d0*xpt(iend-1)) - xpt(iend-2)
           ywrk(ione-2)  = (2.0d0*ypt(ione+1)) - ypt(ione+2)
           ywrk(ione-1)  = (2.0d0*ypt(ione)) - ypt(ione+1)
           ywrk(iend+1)  = (2.0d0*ypt(iend)) - ypt(iend-1)
           ywrk(iend+2)  = (2.0d0*ypt(iend-1)) - ypt(iend-2)
           dlwrk(ione-1) = dl(iend)
           dlwrk(iend+1) = dl(ione)
           swwrk(ione-1) = sw(iend)
           swwrk(iend+1) = sw(ione)
           cswrk(ione-1) = cs(iend)
           cswrk(iend+1) = cs(ione)
           swrk(ione-1)  = s(iend)
           swrk(iend+1)  = s(ione)
c
c    closed polygons - make periodic in x and y
c
         elseif ( icat .eq. -2 ) then
           xwrk(ione-2) = xpt(iend-1)
           xwrk(ione-1) = xpt(iend)
           xwrk(iend+1) = xpt(ione)
           xwrk(iend+2) = xpt(ione+1)
           ywrk(ione-2) = ypt(iend-1)
           ywrk(ione-1) = ypt(iend)
           ywrk(iend+1) = ypt(ione)
           ywrk(iend+2) = ypt(ione+1)
           dlwrk(ione-1) = dl(iend)
           dlwrk(iend+1) = dl(ione)
           swwrk(ione-1) = sw(iend)
           swwrk(iend+1) = sw(ione)
           cswrk(ione-1) = cs(iend)
           cswrk(iend+1) = cs(ione)
           swrk(ione-1)  = s(iend)
           swrk(iend+1)  = s(ione)
c
c    tubes - make periodic in x only
c
         elseif ( icat .eq. -1 ) then
           xwrk(ione-2) = xpt(iend-1)-xlength
           xwrk(ione-1) = xpt(iend)-xlength
           xwrk(iend+1) = xpt(ione)+xlength
           xwrk(iend+2) = xpt(ione+1)+xlength
           ywrk(ione-2) = ypt(iend-1)
           ywrk(ione-1) = ypt(iend)
           ywrk(iend+1) = ypt(ione)
           ywrk(iend+2) = ypt(ione+1)
           dlwrk(ione-1) = dl(iend)
           dlwrk(iend+1) = dl(ione)
           swwrk(ione-1) = sw(iend)
           swwrk(iend+1) = sw(ione)
           cswrk(ione-1) = cs(iend)
           cswrk(iend+1) = cs(ione)
           swrk(ione-1)  = s(iend)
           swrk(iend+1)  = s(ione)
c
c    circles - make periodic in x and y
c
         else
           xwrk(ione-2) = xpt(iend-1)
           xwrk(ione-1) = xpt(iend)
           xwrk(iend+1) = xpt(ione)
           xwrk(iend+2) = xpt(ione+1)
           ywrk(ione-2) = ypt(iend-1)
           ywrk(ione-1) = ypt(iend)
           ywrk(iend+1) = ypt(ione)
           ywrk(iend+2) = ypt(ione+1)
           dlwrk(ione-1) = dl(iend)
           dlwrk(iend+1) = dl(ione)
           swwrk(ione-1) = sw(iend)
           swwrk(iend+1) = sw(ione)
           cswrk(ione-1) = cs(iend)
           cswrk(iend+1) = cs(ione)
           swrk(ione-1)  = s(iend)
           swrk(iend+1)  = s(ione)
         endif
c
c  Calculate stretching forces on entity jent
c
         do 20 i=ione,iend+1
           sm = dist(xwrk(i-1),ywrk(i-1),xwrk(i),ywrk(i))
           wrk(i) = (sm-dlwrk(i-1))*swwrk(i-1)/sm
 20      continue
         do 25 i=ione,iend
           gfrc1(i) = gfrc1(i)
     x              + (wrk(i)*(xwrk(i)-xwrk(i-1)))
     x              + (wrk(i+1)*(xwrk(i)-xwrk(i+1)))
           gfrc2(i) = gfrc2(i) 
     x              + (wrk(i)*(ywrk(i)-ywrk(i-1)))
     x              + (wrk(i+1)*(ywrk(i)-ywrk(i+1)))
 25      continue
c
c  Calculate bending forces
c
         do 30 i=ione-1,iend+1
            wrk(i) = ((xwrk(i+1)*(ywrk(i)-ywrk(i-1)))
     x              + (xwrk(i)*(ywrk(i-1)-ywrk(i+1)))
     x              + (xwrk(i-1)*(ywrk(i+1)-ywrk(i)))
     x              - cswrk(i))*swrk(i)
 30      continue
         do 35 i=ione,iend
            gfrc1(i) = gfrc1(i)
     x              + (wrk(i-1)*(ywrk(i-1)-ywrk(i-2)))
     x              + (wrk(i)*(ywrk(i-1)-ywrk(i+1)))
     x              + (wrk(i+1)*(ywrk(i+2)-ywrk(i+1)))
            gfrc2(i) =  gfrc2(i) 
     x              + (wrk(i-1)*(xwrk(i-2)-xwrk(i-1)))
     x              + (wrk(i)*(xwrk(i+1)-xwrk(i-1)))
     x              + (wrk(i+1)*(xwrk(i+1)-xwrk(i+2)))
 35      continue
c
c  Calculate tethering forces
c
         do 40 i=ione, iend
            gfrc1(i) =  gfrc1(i) + (sth(i)*(xpt(i)-xt(i)))
            gfrc2(i) =  gfrc2(i) + (sth(i)*(ypt(i)-yt(i)))
 40      continue
c
 50   continue
 9990 format(5f15.7)
      return
      end

c
      subroutine gradlink
c
      implicit none
      include 'ibp.inc'
      include 'links.inc'
c
c  local variables
c
      integer i
      double precision sm,xi,xp,xp1,yp,yp1,dist
c
c  compute forces
c
      do 10 i=1,nlink
        xp1 = xpt(n1(i))
        yp1 = ypt(n1(i))
        xp  = xpt(n2(i))
        yp  = ypt(n2(i))
        sm  = dist(xp1,yp1,xp,yp)
        xi  = (sm-rest(n1(i)))*stiff(n1(i))
        if (sm.gt.1.0d-8) then
          xi  = xi/sm
        else
          xi = 0.0d0
        endif
        gfrc1(n1(i))=xi*(xp1-xp)+gfrc1(n1(i))
        gfrc1(n2(i))=-xi*(xp1-xp)+gfrc1(n2(i))
        gfrc2(n1(i))=xi*(yp1-yp)+gfrc2(n1(i))
        gfrc2(n2(i))=-xi*(yp1-yp)+gfrc2(n2(i))
 10   continue
      return
      end
c
c
      subroutine fadd(nxp2,nyp1,f)
c
c  add the background force
c
      implicit none
      include 'domain.inc'
      include 'fluid.inc'
c
      integer nxp2,nyp1
      double precision f(0:nxp2,0:nyp1,1:2)
c
      integer i,j
c
      do j = ibtube, ittube
        do i = 1, nxp2-2
          f(i,j,1)= fbpx+f(i,j,1)
          f(i,j,2)= fbpy+f(i,j,2)
        enddo
      enddo
c
      return
      end
c
      subroutine chgcatb(listf,listb,i,p1,pz1,p2,pz2)
c
c    we have a 1-d array, listf, of platelets divided into
c    categories;  the number listf(j) gives the index of
c    the next platelet of listf in the same category as j
c    or if listf(j)=0, then j is the last platelet in this
c    category.  we also have a 1-d array, listb, which lists
c    the same categories as listf but with the ordering
c    within each category reversed compared with listf
c    this routine allows us to change a platelet from one
c    category to another
c    note:  the index j corresponds to platelet number j
c    in both lists
c
c     listf is the forward list described
c     listb is the reverse ordered list
c     i is the platelet whose category is to change;
c    p1 and pz1 are the initial and final pointers
c    respectively for the category i is in; and p2 is the
c    initial pointer for the category it will be placed in.
c    note:  initial and final are relative to listf, reverse
c    these for listb
c
      integer   i,pz2,p1,p2,pz1,listf,listb
      dimension listf(1),listb(1)
c
c   check if i is the only platelet in category 1
c
      if(listf(i).eq.(0).and.listb(i).eq.(0))go to 18
c
c     check if i is the last platelet in category 1
c
      if(listf(i).eq.0)go to 10
c
c   check if i is beginning of category 1
c
      if(listb(i).eq.0) go to 15
c
c    if none of these, then delete i from category 1
c   put contents of listf(i) in box whose address is listf(i)
c
      listf(listb(i))=listf(i)
c
c    change listb to reflect deletion of i from category 1
c    put contents of listb(i) in position in listb that had
c    address listb(i), namely that with index listf(listb(i))
c
c    listb(listf(listb(i)))=listb(i)
      listb(listf(i))=listb(i)
      go to 25
c
c    i is the last element in category 1; so need to move end
c    pointer to location which had address i and to put a zero
c    there;  on listb the ith element is just ignored
c
   10 continue
      listf(listb(i))=0
      pz1=listb(i)
      go to 25
c
c    i is first element in category 1, move pointer to 2nd
c    element, set listb element corresponding to this second
c    element to zero
c
   15 continue
      p1=listf(i)
      listb(listf(i))=0
      go to 25
c
c    i is the only element in category 1
c
   18 continue
      p1=0
      pz1=0
c
   25 continue
c
c   now add i to beginning of category 2 in listf;
c
      listb(i)=0
      if(p2.gt.0)go to 30
      pz2=i
      go to 33
   30 listb(p2)=i
   33 listf(i)=p2
      p2=i
      return
      end
c
      subroutine chgcate(listf,listb,i,p1,pz1,p2,pz2)
c
c    we have a 1-d array, listf, of platelets divided into
c    categories;  the number listf(j) gives the index of
c    the next platelet of listf in the same category as j
c    or if listf(j)=0, then j is the last platelet in this
c    category.  we also have a 1-d array, listb, which lists
c    the same categories as listf but with the ordering
c    within each category reversed compared with listf
c    this routine allows us to change a platelet from one
c    category to another
c    note:  the index j corresponds to platelet number j
c    in both lists
c
c     listf is the forward list described
c     listb is the reverse ordered list
c     i is the platelet whose category is to change;
c     p1 and pz1 are the initial and final pointers
c    respectively for the category i is in; and pz2 is the
c    final pointer for the category it will be placed in.
c    note:  initial and final are relative to listf, reverse
c    these for listb
c
      integer   i,pz2,p1,p2,pz1,listf,listb
      dimension listf(1),listb(1)
c
c   check if i is the only platelet in category 1
c
      if(listf(i).eq.(0).and.listb(i).eq.(0))go to 18
c
c     check if i is the last platelet in category 1
c
      if(listf(i).eq.0)go to 10
c
c   check if i is beginning of category 1
c
      if(listb(i).eq.0) go to 15
c
c    if none of these, then delete i from category 1
c   put contents of listf(i) in box whose address is listf(i)
c
      listf(listb(i))=listf(i)
c
c    change listb to reflect deletion of i from category 1
c    put contents of listb(i) in position in listb that had
c    address listb(i), namely that with index listf(listb(i))
c
c    listb(listf(listb(i)))=listb(i)
      listb(listf(i))=listb(i)
      go to 25
c
c    i is the last element in category 1; so need to move end
c    pointer to location which had address i and to put a zero
c    there;  on listb the ith element is just ignored
c
   10 continue
      listf(listb(i))=0
      pz1=listb(i)
      go to 25
c
c    i is first element in category 1, move pointer to 2nd
c    element, set listb element corresponding to this second
c    element to zero
c
   15 continue
      p1=listf(i)
      listb(listf(i))=0
      go to 25
c
c    i is the only platelet in category 1
c
   18 continue
      p1=0
      pz1=0
c
c   now add i to end of category 2 in listf; put address i at old end
c   put 0 at address i in listf; begin listb category 2 at i,
c   and put address of old beginning in location i in listb
c
   25 continue
      listf(i)=0
      if(pz2.gt.0) go to 30
      p2=i
      go to 33
   30 listf(pz2)=i
   33 listb(i)=pz2
      pz2=i
      return
      end
c
      subroutine map
c
c this subroutine constructs a map and inverse map
c from the list of indices of entities present in
c the domain to the numbers 1,2,3,...,nent.
c
c it maps the beginning and ending pointers (firstb,lastb) for xb
c to arrays (first,last) which are pointers for the arrays (x,y)
c
c it maps the coordinates of entities present from xb to (x,y)
c
c it maps the entity stiffnesses (sb,swb) and bending angles (csb)
c to arrays (s,sw) and (cs) respectively based on the map
c
c it constructs arrays n1,n2,stiff from the linked-list information
c on interentity links
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'links.inc'
c
c  local variables
c
      integer i,icnt,ient,j,jcnt,lcnt,nfst
      integer nst,nstb,num,npt
c
      icnt=0
      nfst=1
      do 10 ient=1,ment
         if(categor(ient).ne.0)then
            icnt=icnt+1
            pmap   (icnt)=ient
            invpmap(ient)=icnt
            first(icnt)  =nfst
            last (icnt)  =nfst+(lastb(ient)-firstb(ient))
            cat(icnt)    =categor(ient)
            nfst         =last(icnt)+1
         endif
 10   continue
c
c check that at end icnt = nent
c
      if(icnt.ne.nent)then
          write(6,*)' MAP  icnt=',icnt,' nent=',nent
          stop
      endif
c
c map coordinates
c
      do 20 j=1,nent
         ient=pmap(j)
         num =lastb(ient)-firstb(ient)+1
         nstb=firstb(ient)-1
         nst =first (j)-1
         do 15 i=1,num
            npt = nstb+i
            xpt (nst+i) = xb((2*npt)-1)
            ypt (nst+i) = xb(2*npt)
            xt  (nst+i) = xteth(npt)
            yt  (nst+i) = yteth(npt)
            dl  (nst+i) = dlp(npt)
            sw  (nst+i) = swp(npt)
            cs  (nst+i) = csp(npt)
            s   (nst+i) = sp (npt)
            sth (nst+i) = sthp(npt)
 15      continue
 20   continue
c
c map link information, construct n1,n2,stiff
c
c sweep through linked-list of links
c
      lcnt = lkl
      jcnt=0
 25   continue
        if(lcnt.eq.0)goto 30
          jcnt=jcnt+1
          n1(jcnt)    = first(invpmap(p1(lcnt)))+r1(lcnt)
          n2(jcnt)    = first(invpmap(p2(lcnt)))+r2(lcnt)
          lent1(jcnt) = p1(lcnt)
          lent2(jcnt) = p2(lcnt)
          rl1  (jcnt) = r1(lcnt)
          rl2  (jcnt) = r2(lcnt)
          stiff(jcnt) = stiffb(lcnt)
          rest (jcnt) = restb (lcnt)
          lcnt        = listlkf(lcnt)
        goto 25
 30   continue
c
      return
      end
c
      subroutine revmap
c
c this subroutine puts the coordinates (x,y) into the 
c proper place in the array xb
c
c  global include
c
      implicit none
      include 'ibp.inc'
c
c  local variables
c
      integer i,ient,j,nst,nstb,num
c
c map coordinates
c
      do j=1,nent
         ient=pmap(j)
         num =lastb(ient)-firstb(ient)+1
         nstb=firstb(ient)-1
         nst =first (j)-1
         do i=1,num
            xb(2*(nstb+i)-1) = xpt(nst+i)
            xb(2*(nstb+i)  ) = ypt(nst+i)
            usb1(nstb+i)     = us1(nst+i)
            usb2(nstb+i)     = us2(nst+i)
            gfrcb1(nstb+i)   = gfrc1(nst+i)
            gfrcb2(nstb+i)   = gfrc2(nst+i)
         enddo
      enddo
      return
      end
c
c------------------------------------------------------------------
c
      subroutine addmap(xpt,ypt,n,xadd,yadd,nadd,kpt)
c
c this subroutine puts the coordinates (x,y) into the proper place
c in the array xb
c
      implicit none
c
c  global variables
c
      integer n,nadd,kpt
      double precision xadd(nadd),yadd(nadd)
      double precision xpt(n),ypt(n)
c
c  local variables
c
      integer i
c
c map coordinates
c
      do i= 1, nadd
         xpt(kpt+i) = xadd(i)
         ypt(kpt+i) = yadd(i)
      enddo
      return
      end
c
c------------------------------------------------------------------
c
      subroutine delmap(xpt,ypt,n,xdel,ydel,ndel,kpt)
c
c this subroutine replaces the coordinates mapped by addmap
c
c  global include
c
      implicit none
c
c  global variables
c
      integer n,ndel,kpt
      double precision xdel(ndel),ydel(ndel)
      double precision xpt(n),ypt(n)
c
c  local variables
c
      integer i
c
c map coordinates
c
      do i= 1, ndel
         xdel(i) = xpt(kpt+i)
         ydel(i) = ypt(kpt+i)
      enddo
      return
      end
c
c------------------------------------------------------------------
c
      subroutine period(nxp2,nyp1,ph)
c
c extend ph periodically
c global include
c
      implicit none
c
      integer nxp2,nyp1
      double precision ph(0:nxp2,0:nyp1)
c
c  local variables
c
      integer i,j,nx,ny
c
      nx = nxp2-2
      ny = nyp1-1
c
      do 100 i=1,nx
         ph(i,0) = ph(i,ny)
         ph(i,nyp1) = ph(i,1)
 100  continue
      do 120 j=1,ny
         ph(0,j) = ph(nx,j)
         ph(nx+1,j) = ph(1,j)
 120  continue
c
c extend for corner points also
c
      ph(0    ,0   ) = ph(nx,ny)
      ph(nx+1 ,0   ) = ph( 1,ny)
      ph(0    ,nyp1) = ph(nx, 1)
      ph(nx+1 ,nyp1) = ph( 1, 1)
c
      return
      end
c
      subroutine init(nxp2,nyp1,f,u,sinsq,bzero,sfluid,a,z,rsf,
     x    mlink,mlkpt,mpts,mwhere,np,npdx,npdy,nring,nxx,nyy)

c  Modified initialization routine for Fogelsons code
c   incorporates output from little language
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'domain.inc'
      include 'fluid.inc'
      include 'graph.inc'
      include 'links.inc'
      include 'plat.inc'
      include 'active.inc'
c
c  global variables
c
      integer rsf,nxp2,nyp1,np,npdx,npdy,nring,nxx,nyy
      integer mlink,mlkpt,mpts,mwhere,nx,ny
      double precision sfluid(3),a
      double precision bzero(0:nxp2,0:nyp1)
      double precision f(0:nxp2,0:nyp1,1:2)
      double precision u(0:nxp2,0:nyp1,1:2)
      double precision z(0:nxp2,0:nyp1)
      double precision sinsq(0:nxp2,0:nyp1)
c
c  local variables
c
      character*100 type
      integer i,j,jpt,k,knp,listp,inp,inext,rsflag
      integer kk,jl,lk1,lk2,nl,lkmin,lkmax,ltype,getnum
      double precision ang,pi,stf,brk,rst
      double precision si,sj,tpvnx,tpvny,yy
c
c  initialize the namelist
c
      namelist /IBPTS/ rsflag,
     x    colorf,linef,fluidf,ibptsf,linkf,partf,
     x    xmin,xmax,ymin,ymax,dlt,h,mu,ro,umax,fbpx,fbpy,
     x    ntstart,ntmax,time,
     x    ccount,cfreq,lcount,lfreq,ivdens,usave,fsave,
     x    ipbx,ipby,ibtube,ittube,btubil,btubir,ttubil,ttubir,
     x    pwalk,iplat,maxlink,pltdif,pdens,scoh,scgap,
     x    sadh,sagap,sbreaka,sbreakc,stiffa,stiffc,rlinkc,
     x    rlinka

c
c  global velocity info
c
      data vmax, vmaxg / 0.d0, 0.d0 /
      data vimax, vimaxg / 0.d0, 0.d0 /
      data fmax, fmaxg / 0.d0, 0.d0 /
      data fimax, fimaxg / 0.d0, 0.d0 /
c
c  data from the ibis parser
c
      open (12,file='ibisdat/init.dat',status='old')
      read(12,IBPTS)
      close(12)
      rsf = rsflag

c
c  retrieve the domain information
c
      nx = nxp2-2
      ny = nyp1-1
      xlength = xmax - xmin
      ylength = ymax - ymin
      ycent   = ylength/2.0d0
c
c  compute grid information  -- CHECK ON ITUBE
c
      dlibp   = h/2
      ytop    = h*ittube
      ybot    = h*ibtube
c
c  get the fluid information
c
      if ( umax.ne.0.0 .and. rsf.eq.0 )  umax0 = 1
c
c typical platelet dimensions
c
      pi     = 4.D+00*datan(1.D+00)
      ang    = 2.D+00*pi/dble(nring)
      if (nring.gt.1) then
        diam = dlibp/dsin(.5D+00*ang)
      else
        diam = 1.0d0
      endif
c
c  rescale adhesion and cohesion data
c
      scoh    = scoh*diam
      scgap   = scgap*diam
      sadh    = sadh*diam
      sagap   = sagap*diam
      sbreaka = sbreaka*dlibp
      sbreakc = sbreakc*dlibp
      stiffa  = stiffa/dlibp
      stiffc  = stiffc/dlibp
c
c x portion of injured walls
c   xadhmin - minimum x injury location >= zero
c   xadhmin - maximum x injury location
c
      xadhmax = max(btubir,ttubir)*h
      if (btubil.gt.0 .and. ttubil.eq.0 ) then
        xadhmin = btubil*h
      elseif (btubil.eq.0 .and. ttubil.gt.0 ) then
        xadhmin = ttubil*h
      else
        xadhmax = min(btubil,ttubil)*h
      endif
c
c  load the tube injury locations
c
      afirstb(1) = btubil
      alastb(1) = btubir
      afirstb(2) = ttubil
      alastb(2) = ttubir
c
c number of entities
c
      open (100,file=ibptsf,status='unknown')
      read (100,*) jfil,jplat,jpoly,nlinke
      nent = jplat + jfil + jpoly
c
c filaments first, pointers into xb and spring information
c
      jpt = 0
      if (jfil .gt. 0) then
        pf  = 1
        pzf = jfil
        do 30, j = pf, pzf
          knp        = 0
          listb(j)   = j - 1
          listf(j)   = j + 1
          firstb(j)  = jpt + 1
          activ3(j)  = jpt + 1
          read(100,1000) type
          read(100,*) inp, iclose(j), categor(j), 
     x         activ1(j), activ2(j),
     x         xcm(j), ycm(j)
          xcm(j) = xcm(j) - xmin
          ycm(j) = ycm(j) - ymin
 20       continue
            jpt      = jpt + 1
            knp      = knp + 1
            read (100,*) inext, xb(2*jpt-1), xb(2*jpt),
     x                   xteth(jpt),  yteth(jpt),
     x                   dlp(jpt),    swp(jpt),
     x                   csp(jpt),    sp(jpt),
     x                   sthp(jpt)
            xb(2*jpt-1) = xb(2*jpt-1) - xmin
            xb(2*jpt) = xb(2*jpt) - ymin
            xteth(jpt) = xteth(jpt) - xmin
            yteth(jpt) = yteth(jpt) - ymin
            if ( inext.eq.1 ) goto 20
 25       continue
          if ( knp.ne.inp ) then
            write (6,*)
     x       'Filament point count does not match number of points'
          endif
          lastb(j)   = jpt
          activ4(j) = jpt
 30     continue
        listf(pzf)  =  0
      else
        pf  = 0
        pzf = 0
      endif
c
c polygon pointers
c  polygon links list parameters, pointers into xb
c   and spring information
c
      knp = 0
      if (jpoly .gt. 0) then
        pg  = jfil + 1
        pzg = pg + jpoly - 1
        listp = pzg
c 
c compute polygons positions
c
        do 45, j = pg, pzg
          knp        = 0
          listb(j)   = j - 1
          listf(j)   = j + 1
          firstb(j)  = jpt + 1
          centx(j)   = 0.0d0
          centy(j)   = 0.0d0
          activ3(j) = jpt+1
          read(100,1000) type
          read(100,*) inp, iclose(j), categor(j), 
     x         activ1(j), activ2(j), 
     x         xcm(j), ycm(j)
          xcm(j) = xcm(j) - xmin
          ycm(j) = ycm(j) - ymin
 35       continue
            jpt      = jpt + 1
            knp      = knp + 1
            read (100,*) inext, xb(2*jpt-1), xb(2*jpt),
     x                   xteth(jpt),  yteth(jpt),
     x                   dlp(jpt),    swp(jpt),
     x                   csp(jpt),    sp(jpt),
     x                   sthp(jpt)
            xb(2*jpt-1) = xb(2*jpt-1) - xmin
            xb(2*jpt) = xb(2*jpt) - ymin
            xteth(jpt) = xteth(jpt) - xmin
            yteth(jpt) = yteth(jpt) - ymin
            centx(j) = centx(j) + xb(2*jpt-1)
            centy(j) = centy(j) + xb(2*jpt)
            if ( inext.eq.1 ) goto 35
 40       continue
          if ( knp.ne.inp ) write (6,*)
     x       'Polygon point count does not match number of points'
c
c Finish the list data for polygons
c    centers are center of mass
c
          lastb(j)   = jpt
          activ4(j) = jpt
          centx(j)   = centx(j)/dble(knp)
          centy(j)   = centy(j)/dble(knp)
 45     continue
        listb(pg)  = 0
        listf(pzg) = 0
      else
        pg   = 0
        pzg  = 0
        listp = pzf
      endif
c
c activated platelet pointers - currently all platelets
c    links list parameters and pointers into xb
c
      if (jplat .gt. 0) then
        pac  = listp + 1
        pzac = pac + jplat - 1
        do 60 j=pac,pzac
          knp        = 0
          listb(j)   = j - 1
          listf(j)   = j + 1
          firstb(j)  = jpt + 1
          activ3(j) = jpt + 1
          read(100,1000) type
          read(100,*) inp, iclose(j), categor(j), 
     x         activ1(j), activ2(j), 
     x         xcm(j), ycm(j)
          xcm(j) = xcm(j) - xmin
          ycm(j) = ycm(j) - ymin
c
c  put in the immersed boundary points with spring info
c
 50       continue
            knp  = knp + 1
            jpt  = jpt + 1
            read (100,*) inext, xb(2*jpt-1), xb(2*jpt),
     x                   xteth(jpt),  yteth(jpt),
     x                   dlp(jpt),    swp(jpt),
     x                   csp(jpt),    sp(jpt),
     x                   sthp(jpt)
            xb(2*jpt-1) = xb(2*jpt-1) - xmin
            xb(2*jpt) = xb(2*jpt) - ymin
            xteth(jpt) = xteth(jpt) - xmin
            yteth(jpt) = yteth(jpt) - ymin
            if ( inext.eq.1 ) goto 50
 55       continue
          if ( knp.ne.inp ) then

            write (6,*)
     x       'Circle point count does not match number of points'
          endif
          lastb(j)   = jpt
          activ4(j) = jpt
          centx(j)   = xcm(j)
          centy(j)   = ycm(j)
 60     continue
        listb(pac)  = 0
        listf(pzac) = 0
      else
        pac  = 0
        pzac = 0
      endif
c
c unactivated platelet pointers - currently none
c
      pua  = 0
      pzua = 0
c
c linked platelet pointers - currently none
c
      pl   = 0
      pzl  = 0
c
c available space pointers and link list parameters
c
      pas  = nent+1
      pzas = ment
      do 65, j = pas, ment
        listf(j)   = j+1
        listb(j)   = j-1
        categor(j) = 0
        firstb(j)  = max0(lastb(j-1)+1,1)
        lastb (j)  = firstb(j)+(nring-1)
 65   continue
      listf(ment) = 0
      listb(pas)  = 0
      npts = jpt
c
      close (100)
c
c initialize link structure
c
      do j = 1, mpts
         status(j) = 0
      enddo
      do j = 1, mwhere
         numlink(j)=0
      enddo
c
      kk = 0
      k = 1
      if ( linkf(1:4) .ne. 'null' ) then
c
        open (12,file=linkf,status='old')
        if ( rsf .eq. 0 ) then
          do j = 1, nlinke
            read (12,*) lk1,lk2,nl,stf,brk,rst,ltype
            do jl = 1, nl
              listlkf(k) = k+1
              listlkb(k) = k-1
              k = k+1
            enddo
            lkmin = min0(getnum(lk1),getnum(lk2))
            lkmax = max0(getnum(lk1),getnum(lk2))
            if ( ltype .eq. 1 ) then
              call join (lkmin,lkmax,nl,stf,brk,rst,kk,mpts)
            elseif ( ltype .eq. 2 ) then
              call connect (lkmin,lkmax,nl,stf,brk,rst,kk)
            elseif ( ltype .eq. 3 ) then
              call spoke (getnum(lk1),getnum(lk2),nl,stf,brk,rst,kk)
            elseif ( ltype .ge. 4 ) then
              call bridge (lkmin,lkmax,nl,stf,brk,rst,kk,ltype)
            endif
          enddo
          listlkf(k) = 0
          listlkb(1) = 0
          nlink = kk
        elseif ( rsf .eq. 1 ) then
          read (12,*) nlink
          do j = 1, nlink
            read (12,*) q1(j),q2(j),p1(j),p2(j),
     x         r1(j),r2(j),stiffb(j),sbreak(j),restb(j)
            listlkf(j) = j+1
            listlkb(j) = j-1
            status(firstb(p1(j))+r1(j)) = 1
            status(firstb(p2(j))+r2(j)) = 1
            lcat(j) = 1
          enddo
          listlkf(k) = 0
          listlkb(1) = 0
        endif
c
        close(12)

      endif
c
      lkl   = min0(1,nlink)
      lkzl  = nlink
      lkas  = nlink+1
      lkzas = mlink
c
      do j = lkas, mlink
        listlkf(j) = j+1
        listlkb(j) = j-1
        lcat(j) = 0
        p1(j) = 0
        p2(j) = 0
        r1(j) = 0
        r2(j) = 0
        stiffb(j) = 0.0D+00
        restb (j) = 0.0D+00
        sbreak(j) = 0.0D+00
      enddo
      listlkf(mlink) = 0
      listlkb(lkas ) = 0
c
c  wall bound aggregate links
c  
      nas=1
      nzas=mlkpt
      do j=1,ment
         fstlnk(j)=0
         lstlnk(j)=0
      enddo
      do j=1,mlkpt
         nxtlnkf(j)=j+1
         nxtlnkb(j)=j-1
         pltlnk (j)=0
      enddo
      nxtlnkf(nzas)=0
      nxtlnkb(nas )=0
c
c  define parameters for energy minimization
c  
      lam=9.D+00/64.D+00 *(dlt**2)*dlibp/(h*h) *2.0D+00
c
c  Flow initialization stuff
c  trig functions for fft2d
c
      tpvnx=2.D+00*pi/dble(nx)
      tpvny=2.D+00*pi/dble(ny)
      do i=1,nx
        si=(dsin(tpvnx*(i-1)))**2
        do j=1,ny
          sj=(dsin(tpvny*(j-1)))**2
          sinsq(i,j)=-(4.D+00*(si+sj))
          bzero(i,j)=0.D+00
        enddo
      enddo
      sinsq(1,1)=1.D+00
      sinsq(1,ny/2+1)=1.D+00
      sinsq(nx/2+1,1)=1.D+00
      sinsq(nx/2 +1,ny/2+1)=1.D+00
c
c navier-stokes parameters
c
      sfluid(1)=dlt/(2*h)
      sfluid(2)=mu*dlt/(h*h)
      sfluid(3)=1.D+00/(2*h)
      a=ro+2*sfluid(2)
c
c initialize the pressure and velocity fields
c
      do j=1,nyp1
        do i=1,nxp2
          z(i,j)=0.D+00
        enddo
      enddo
c
      do k = 1, 2
        do j = 0, nyp1
          do i = 0, nxp2
            u(i,j,k)=0.0D+00
          enddo
        enddo
      enddo
c
c if starting with zero flow, skip this!
c
      fluxx=0.0D+00
      if (umax0.gt.0) then
        do 350 i = 1, nx
           do 340 j = ibtube, ittube
             yy       = (4.D+00*dble(j)*h/ylength - 2.D+00)
             u(i,j,1) = umax*(1.0D+00 - yy**2)
 340       continue
           u(i,0,1) = u(i,ny,1)
 350    continue
        do 360 j = 0, ny
          u(0,j,1) = u(nx,j,1)
          u(0,j,2) = u(nx,j,2)
 360    continue
      endif
c
c if fluid file provided, then read it
c
      if ( fluidf(1:4) .ne. 'null' ) then
        open (12, file=fluidf, status='old')
        do j = 1, ny
          do i = 1, nx
            read(12,*) u(i,j,1), u(i,j,2)
          enddo
        enddo
        close (12)
      endif
c
c compute maximum speed
c
      do j = 1, ny
        fluxx=fluxx + h*u(2,j,1)
        do i = 0, nx
           vmax = max(vmax,sqrt(u(i,j,1)**2 + u(i,j,2)**2))
        enddo
      enddo
c
c initialize background force density for poisuelle flow
c
      do 410 j=1,ibtube-1
        do 410 i=0,nx
          f(i,j,1)=0.0d0
          f(i,j,2)=0.0d0
 410  continue
      do 420 j=ibtube, ittube
        do 420 i=0,nx
          f(i,j,1)= fbpx
          f(i,j,2)= fbpy
 420  continue
      do 430 j=ittube+1,ny
        do 430 i=0,nx
          f(i,j,1)=0.0d0
          f(i,j,2)=0.0d0
 430  continue
c
c  graphics initialization
c
      i = 0
      if (npdx .eq. 0) then
        npx = nxx
      else
        npx = (nxx+1)/npdx
      endif
      if (npdy .eq. 0) then
        npy = nyy
      else
        npy = (nyy+1)/npdy
      endif
      pdx = xlength/dble(npx)
      pdy = ylength/dble(npy)
      nactm = 0
      if ( partf(1:4) .ne. 'null' ) then
        open (12,file=partf,status='old')
 490    continue
          i = i + 1
          read (12,*,end=500) px(i), py(i), pmark(i)
          px(i) = px(i) - xmin
          py(i) = py(i) - ymin
          nactm = nactm+1
          goto 490
 500    continue
        close (12)
      else
        do j = 1, npx
          do k = 1, npy
            i = i + 1
            px(i) = (dble(j)-0.5d0)*pdx
            py(i) = (dble(k)-0.5d0)*pdy
            pmark(i) = 2
            nactm = nactm+1
          enddo
        enddo
      endif
      do k = i, np
        px(k) = xlength/2.0d0
        py(k) = ylength/2.0d0
        pmark(k) = 0
      enddo
      dx = xlength/dble(nxx)
      dy = ylength/dble(nyy)

      do k = 1, 16
        icolr(k) = k-1
      enddo
      do k = 1, ment
        icoli(k) = mod(k-1,223)+32
      enddo
c
c  report some results
c
      write(6,*)' '
      write(6,*)' *** COMPUTATIONAL DOMAIN DATA *** '
      write(6,*)' *** xmin *** ',xmin
      write(6,*)' *** xmax *** ',xmax
      write(6,*)' *** ymin *** ',ymin
      write(6,*)' *** ymax *** ',ymax
      write(6,*)' *** delta t *** ',dlt
      write(6,*)' *** timesteps *** ',ntmax
      write(6,*)' *** end time *** ', ntmax*dlt
      write(6,*)' *** delta x_grid *** ', h
      write(6,*)' *** grid points *** ', nx, ny
      write(6,*)' '
      write(6,*)' *** IMMERSED BOUNDARY POINTS *** '
      write(6,*)' *** number of entities *** ', nent
      write(6,*)' *** number of points *** ', npts
      write(6,*)' '
      write(6,*)' *** FLUID DATA *** '
      write(6,*)' *** Umax *** ', umax
      write(6,*)' *** mu *** ', mu
      write(6,*)' *** rho *** ', ro
      write(6,*)' '
      write(6,*)' *** GRAPHICS DATA *** '
      write(6,*)' *** color plot size *** = ', nxx, nyy
      write(6,*)' *** color plot frequency *** = ', cfreq
      write(6,*)' *** line plot frequency *** = ', lfreq
      write(6,*)' '
c
c  platelet data
c
      if (iplat .eq. 1) then
        write(6,*)' '
        write(6,*)' *** INTER-ENTITY LINK DATA *** '
        write(6,*)' *** max number of links *** =', maxlink
        write(6,*)' *** plat-plat link strength *** =', stiffa
        write(6,*)' *** plat-wall link strength *** =', stiffc
        write(6,*)' *** adhesive link rest length *** =',rlinka
        write(6,*)' *** cohesive link rest length *** =',rlinkc
        write(6,*)' *** adhesive link break length *** =',sbreaka
        write(6,*)' *** cohesive link break length *** =',sbreakc
        write(6,*)' *** center of mass cohere length *** =',scoh
        write(6,*)' *** center of mass adhere length *** =',sadh
        write(6,*)' *** ib point dist cohere length *** =',scgap
        write(6,*)' *** ib point dist adhere length *** =',sagap
        write(6,*)' '
        write(6,*)' *** PLATELET DATA *** '
        write(6,*)' *** points on a ring *** = ',nring
        write(6,*)' *** walk length parameter *** = ',pltdif
        write(6,*)' *** platelet density *** = ', pdens
      endif
c
      return
 1008 format(7(1x,e15.7))
 1000 format (a4)
      end
c
      subroutine join(ient,jent,nl,stf,brk,rstl,kk,mpts)
c
c This subroutine forms nl links between entity ient and
c entity jent.  It creates at most one link per immersed
c boundary point.
c
c  global include
c
      implicit none
c
c     global common variables
c
      include 'domain.inc'
      include 'graph.inc'
      include 'ibp.inc'
      include 'links.inc'
c
c  global variables
c
      integer ient,jent,nl,kk,qm1,qm2,mpts
      double precision stf,brk,rstl
c
c  local variables
c
      integer ic,imin,jc,jmin,nc,nwhere
      double precision dist,sdist,smin
c
      do 10, ic = 1, mpts
        stat1(ic) = 0
 10   continue
      do 20, ic = 1, mpts
        stat2(ic) = 0
 20   continue
c
c Form the links.  Check to see the points are non-linked or 
c linked.  From the unlinked points on the two entities, find 
c the closest pair, one on each entity.  Link that pair.
c
      nwhere = ((ient-2)*(ient-1))/2 + jent
      numlink(nwhere) = nl
      do 80, nc = 1, nl
        imin=0
        jmin=0
        smin=1000.D+00
        do 60 ic=firstb(ient),lastb(ient)
          if ( stat1(ic).eq.0 ) then
            do 40 jc=firstb(jent),lastb(jent)
              if ( stat2(jc).eq.0 ) then
                sdist = dist(xb(2*ic-1),xb(2*ic),
     x                       xb(2*jc-1),xb(2*jc))
                if(sdist.lt.smin)then
                  smin = sdist
                  imin = ic - firstb(ient)
                  jmin = jc - firstb(jent)
                  qm1  = ic
                  qm2  = jc
                endif
              endif
 40         continue
          endif
 60     continue
c
c Form link between point imin on ient and point jmin on jent
c
        kk = kk + 1
        p1(kk) = ient
        r1(kk) = imin
        q1(kk) = qm1
        p2(kk) = jent
        r2(kk) = jmin
        q2(kk) = qm2
        stat1(qm1) = qm2
        stat2(qm2) = qm1
        stiffb(kk) = stf
        sbreak(kk) = brk*h
        restb (kk) = rstl*h
        lcat(kk) = 1
        status(firstb(ient)+imin) = 1
        status(firstb(jent)+jmin) = 1
 80   continue
c
      return
      end
c
      subroutine bridge(ient,jent,nl,stf,brk,rstl,kk,type)
c
c This subroutine forms a bridge work of links between 
c entity ient and entity jent.  It creates three links per 
c immersed boundary point.
c
c  global include
c
      implicit none
c
c     global common variables
c
      include 'domain.inc'
      include 'graph.inc'
      include 'ibp.inc'
      include 'links.inc'
c
c  global variables
c
      integer ient,jent,nl,kk,type
      double precision stf,brk,rstl
c
c  local variables
c
      integer ic,jc,nc,nwhere,incl,incr,inc,nloop
      double precision dist
c
c Form the 3 links from every point on ient to jent.
c Links are formed sequentially.
c
      nwhere = ((ient-2)*(ient-1))/2 + jent
      numlink(nwhere) = nl
c
      nloop = lastb(jent)-firstb(jent)
c
      do nc = 0, nloop
c
        ic = firstb(ient) + nc
c
c  if its not a closed structure handle it differently
c
        incl = -1
        incr = 1
        if (nc.eq.0 .and. type.gt.4) then
           incl = 0
           incr = 1
        elseif (nc.eq.nloop .and. type.gt.4) then
           incl = -1
           incr = 0
        endif        
c
c Create three links at every interior point by forming
c links between nearest points and next nearest points
c
        do inc = incl,incr
          jc = firstb(jent) + nc + inc
          if ( jc .lt. firstb(jent) ) jc = lastb(jent)
          if ( jc .gt. lastb(jent) ) jc = firstb(jent)
          kk = kk + 1
          p1(kk) = ient
          r1(kk) = ic - firstb(ient)
          q1(kk) = ic
          p2(kk) = jent
          r2(kk) = jc - firstb(jent)
          q2(kk) = jc
          lcat(kk) = 1
          stiffb(kk) = stf
          sbreak(kk) = brk*h
          restb (kk) = dist(xb(2*ic-1),xb(2*ic),
     x                       xb(2*jc-1),xb(2*jc))*rstl
        enddo
      enddo
c
      return
      end
c
      subroutine spoke(ient,jent,nl,stf,brk,rstl,kk)
c
c This subroutine forms a spoke work of links between 
c entity ient and entity jent.  Entity ient is assumed
c to have one immersed boundary point, and this routine
c creates one link for every immersed boundary point
c in entity jent
c
c  global include
c
      implicit none
c
c     global common variables
c
      include 'domain.inc'
      include 'graph.inc'
      include 'ibp.inc'
      include 'links.inc'
c
c  global variables
c
      integer ient,jent,nl,kk
      double precision stf,brk,rstl
c
c  local variables
c
      integer ic,jc,nc,nwhere
      double precision dist
c
      nwhere = ((ient-2)*(ient-1))/2 + jent
      numlink(nwhere) = nl
      ic = firstb(ient)
c
c Create one link at every point on jent
c
      do nc = 0, lastb(jent)-firstb(jent)

        jc = firstb(jent) + nc
        kk = kk + 1
        p1(kk) = ient
        r1(kk) = 0
        q1(kk) = ic
        p2(kk) = jent
        r2(kk) = jc - firstb(jent)
        q2(kk) = jc
        stiffb(kk) = stf
        sbreak(kk) = brk*h
        restb (kk) = dist(xb(2*ic-1),xb(2*ic),
     x                    xb(2*jc-1),xb(2*jc))*rstl
        lcat(kk) = 1

      enddo
c
      return
      end
c
      subroutine connect(ient,jent,nl,stf,brk,rstl,kk)
c
c This subroutine forms a one to one connection of links between 
c entity ient and entity jent.  It creates one link per 
c immersed boundary point.
c
      implicit none
c
c     global common variables
c
      include 'domain.inc'
      include 'graph.inc'
      include 'ibp.inc'
      include 'links.inc'
c
c  global variables
c
      integer ient,jent,nl,kk
      double precision stf,brk,rstl
c
c  local variables
c
      integer ic,jc,nc,nwhere,nloop
      double precision dist
c
c Form the 3 links from every point on ient to jent.
c Links are formed sequentially.
c
      nwhere = ((ient-2)*(ient-1))/2 + jent
      numlink(nwhere) = nl
c
      nloop = lastb(jent)-firstb(jent)
c
c Create one link at immersed boundry points
c
      do nc = 0, nloop
c
        ic = firstb(ient) + nc
c
c Form links between nearest points and next nearest points
c
        jc = firstb(jent) + nc
        kk = kk + 1
        p1(kk) = ient
        r1(kk) = ic - firstb(ient)
        q1(kk) = ic
        p2(kk) = jent
        r2(kk) = jc - firstb(jent)
        q2(kk) = jc
        stiffb(kk) = stf
        sbreak(kk) = brk*h
        restb (kk) = dist(xb(2*ic-1),xb(2*ic),
     x                    xb(2*jc-1),xb(2*jc))*rstl
        lcat(kk) = 1

      enddo
c
      return
      end
c
c
      integer function getnum(name)
c
c   given the name of an entity, return its number
c
      include 'ibp.inc'
c
      integer i,name
c
      do i = 1, ment
        if (activ2(i).eq.name) then
          getnum=i
          return
        endif
      enddo
c
      return
      end
c
      subroutine navs (nxp2,nyp1,u,f,p,psi,
     *        bzero,sinsq,z,
     *        b,c,r,work1,work2,yh,yp,
     *        a,dlt,h,ro,sfluid)
c
      implicit none
c
      integer nxp2, nyp1
      double precision a,dlt,h,ro,sfluid(3)
c
      double precision u(0:nxp2,0:nyp1,1:2)
      double precision f(0:nxp2,0:nyp1,1:2)
      double precision p(0:nxp2,0:nyp1)
      double precision psi(0:nxp2,0:nyp1)
c
      double precision bzero(0:nxp2,0:nyp1)
      double precision sinsq(0:nxp2,0:nyp1)
      double precision z(0:nxp2,0:nyp1)
c
      double precision b(0:nxp2,0:nyp1)
      double precision c(0:nxp2,0:nyp1)
      double precision r(0:nxp2,0:nyp1)
      double precision work1(0:nxp2,0:nyp1)
      double precision work2(0:nxp2,0:nyp1)
      double precision yh(0:nxp2,0:nyp1)
      double precision yp(0:nxp2,0:nyp1)
c
      double precision lamx(1)
      common /xlam/ lamx

      double precision lamy(1)
      common /ylam/ lamy
c
c  local variables
c
      integer i,j,nc,nx,ny,nxp1
      double precision pro,xl,yl
c
      nx = nxp2-2
      ny = nyp1-1
      nxp1 = nxp2-1
c
      do i=1,nx
        z(i,1   )=0.0D+00
        z(i,ny-1)=0.0D+00
      enddo
c
      do j=1,ny
        do i=0,nxp2
          bzero(i,j)=0.0D+00
        enddo
      enddo
c
c for each fixed y-value, we solve a periodic system
c of linear equations which will give us the first
c fractional step solution.
c
      do nc=1,2
c
        do j=1,ny
          do i=1,nx
            pro=ro*u(i,j,1)*sfluid(1)
            c(i,j)=-pro - sfluid(2)
            b(i,j)= pro - sfluid(2)
            r(i,j)= ro*u(i,j,nc)+dlt*f(i,j,nc)
          enddo
        enddo
c
c  we solve first with correct r.h.s. but zero b.c.'s
c
        call tridgx(nxp2,nyp1,a,b,c,r,yp,work1,work2)
c
c  we now solve with zero r.h.s. and b.c. u(0,y,t)=u(1,y,t)=1
c
        do j=1,ny
          z(1,j)=-c(1,j)
          z(nx-1,j)=-b(nx-1,j)
        enddo
        call tridgx(nxp2,nyp1,a,b,c,z,yh,work1,work2)
c
c the correct linear combo. of particular and homogeneous
c solutions is found by using the equation at endpoint...
c
        do j=1,ny
          xl=r(nx,j)-c(nx,j)*yp(nx-1,j)-b(nx,j)*yp(1,j)
          yl=c(nx,j)*yh(nx-1,j)+a+b(nx,j)*yh(1,j)
          lamy(j)=xl/yl
        enddo
c
        do j=1,ny
          f(nx,j,nc)=lamy(j)
          do i=1,nx-1
            f(i,j,nc)=lamy(j)*yh(i,j) + yp(i,j)
          enddo
        enddo
      enddo

      do j=1,ny
         z(1   ,j)=0.0D+00
         z(nx-1,j)=0.0D+00
      enddo
c
c now comes our second fractional step.  for each
c fixed value of x we calculate f(x,y,ncomp) by
c solving another periodic linear system in the
c same manner as above.......
c
      do nc=1,2
c
        do j=1,ny
          do i=1,nx
            pro=sfluid(1)*ro*u(i,j,2)
            c(i,j)=-pro -sfluid(2)
            b(i,j)= pro - sfluid(2)
            r(i,j)=ro*f(i,j,nc)
          enddo
        enddo
c
        call tridgy(nxp2,nyp1,a,b,c,r,yp,work1,work2)
c
        do i=1,nx
          z(i,1   )=-c(i,1)
          z(i,ny-1)=-b(i,ny-1)
        enddo
c
        call tridgy(nxp2,nyp1,a,b,c,z,yh,work1,work2)
c
        do i=1,nx
           xl=r(i,ny)-c(i,ny)*yp(i,ny-1)-b(i,ny)*yp(i,1)
           yl=c(i,ny)*yh(i,ny-1)+a+b(i,ny)*yh(i,1)
           lamx(i)=xl/yl
        enddo

        do i=1,nx
          do j=1,ny-1
            f(i,j,nc)=lamx(i)*yh(i,j)+yp(i,j)
          enddo
          f(i,ny,nc)=lamx(i)
        enddo
      enddo

c
c make f periodic in x
c
      do j=1,ny
        f(0   ,j,1)=f(nx,j,1)
        f(nxp1,j,1)=f(1 ,j,1)
        f(0   ,j,2)=f(nx,j,2)
        f(nxp1,j,2)=f(1 ,j,2)
      enddo
c
c div calculates the divergence of 2nd frac. step
c
      call div(sfluid(3),f,p,nxp2,nyp1)
c
c add sources and sinks into the fluid field
c
      do j = 1,ny
        do i = 0,nx
          p(i,j) = p(i,j) - psi(i,j)
        enddo
      enddo
c
      call fft2d(p,bzero,-1,nxp2,nyp1)
      do j=1,ny
        do i=1,nx
          p(i,j)=p(i,j)/sinsq(i,j)*h*h*4.D+00
          bzero(i,j)=bzero(i,j)/sinsq(i,j)*h*h*4.D+00
        enddo
      enddo
      call fft2d(p,bzero,+1,nxp2,nyp1)
      call newu(u,h,f,p,nxp2,nyp1)
c
      return
      end
c
c------------------------------------------------------------
c
      subroutine newu(u,h,f,p,nxp2,nyp1)
c
      integer i,j,nx,ny,nxp1
      double precision h,s
c
      integer nxp2,nyp1
      double precision p,u,f
c
      dimension u(0:nxp2,0:nyp1,2)
      dimension f(0:nxp2,0:nyp1,2)
      dimension p(0:nxp2,0:nyp1)
c
      nx = nxp2-2
      nxp1 = nxp2-1
      ny = nyp1-1
c
      s=1.D+00/(2*h)
      do i=2,ny-1
         u(1 ,i ,2)=f(1 ,i ,2)-s*(p(1  ,i+1)-p(1  ,i-1))
         u(nx,i ,2)=f(nx,i ,2)-s*(p(nx ,i+1)-p(nx ,i-1))
      enddo
c
      do i=2,nx-1
         u(i ,1 ,1)=f(i ,1 ,1)-s*(p(i+1,1  )-p(i-1,1  ))
         u(i ,ny,1)=f(i ,ny,1)-s*(p(i+1,ny )-p(i-1,ny ))
      enddo
c
      do i=2,nx-1
        do j=2,ny-1
          u(i,j,1)=f(i,j,1)-s*(p(i+1,j  )-p(i-1,j  ))
          u(i,j,2)=f(i,j,2)-s*(p(i  ,j+1)-p(i  ,j-1))
        enddo
      enddo
c
      do j=1,ny
         u(1 ,j ,1)=f(1 ,j ,1)-s*(p(2,j)-p(nx  ,j   ))
         u(nx,j ,1)=f(nx,j ,1)-s*(p(1,j)-p(nx-1,j   ))
      enddo
c
      do j=1,nx
         u(j ,1 ,2)=f(j ,1 ,2)-s*(p(j,2)-p(j   ,ny  ))
         u(j ,ny,2)=f(j ,ny,2)-s*(p(j,1)-p(j   ,ny-1))
      enddo
c
      do j=1,ny
          u(0   ,j,1)=u(nx,j,1)
          u(nxp1,j,1)=u(1 ,j,1)
          u(0   ,j,2)=u(nx,j,2)
          u(nxp1,j,2)=u(1 ,j,2)
      enddo
c
      do i=1,nx
          u(i,   0,1)   =u(i,ny,1)
          u(i,   0,2)   =u(i,ny,2)
      enddo
c
      u(0   ,0   ,1)=u(nx,ny,1)
      u(nxp1,0   ,1)=u(1 ,ny,1)
      u(0   ,0   ,2)=u(nx,ny,2)
      u(nxp1,0   ,2)=u(1 ,ny,2)
c
      return
      end
c
c------------------------------------------------------------
c
      subroutine div(s3,f,d,nxp2,nyp1)
c
      implicit none
c
      integer  i,j,nx,ny

      integer nxp2, nyp1
      double precision d,s3,f
      dimension   f(0:nxp2,0:nyp1,2),d(0:nxp2,0:nyp1)
c
      nx = nxp2-2
      ny = nyp1-1
c
      do i=2,nx-1
        do j=2,ny-1
           d(i,j)=s3*(f(i+1,j  ,1)-f(i-1,j  ,1)+
     +              f(i  ,j+1,2)-f(i  ,j-1,2))
        enddo
      enddo
c
      do j=2,ny-1
         d(1 ,j )=s3*(f(2  ,j  ,1)-f(nx  ,j   ,1)+
     +                f(1  ,j+1,2)-f(1   ,j-1 ,2))
         d(nx,j )=s3*(f(1  ,j  ,1)-f(nx-1,j   ,1)+
     +                f(nx ,j+1,2)-f(nx  ,j-1 ,2))
      enddo
c
      do i=2,nx-1
         d(i ,1 )=s3*(f(i+1,1  ,1)-f(i-1 ,1   ,1)+
     +                f(i  ,2  ,2)-f(i   ,ny  ,2))
         d(i ,ny)=s3*(f(i+1,ny ,1)-f(i-1 ,ny  ,1)+
     +                f(i  ,1  ,2)-f(i   ,ny-1,2))
      enddo
c
      d(1 ,1 )=s3*(f(2 ,1 ,1)-f(nx  ,1   ,1)
     +            +f(1 ,2 ,2)-f(1   ,ny  ,2))
      d(1 ,ny)=s3*(f(2 ,ny,1)-f(nx  ,ny  ,1)
     +            +f(1 ,1 ,2)-f(1   ,ny-1,2))
      d(nx,1 )=s3*(f(1 ,1 ,1)-f(nx-1,1   ,1)
     +            +f(nx,2 ,2)-f(nx  ,ny  ,2))
      d(nx,ny)=s3*(f(1 ,ny,1)-f(nx-1,ny  ,1)
     +            +f(nx,1 ,2)-f(nx  ,ny-1,2))
c
      return
      end
c
c------------------------------------------------------------
c
      subroutine fft2d(a,b,isign,nxp2,nyp1)
c
c*****two-dimensional fourier transform of a complex array a+ib.
c     (that is, the real part is stored in a and the immaginary
c     part is stored in b.)
c     the transform is computed in place, overwriting a and b.
c     vectorization is achieved as follows:
c     when computing transform in x-direction, vectorize over y;
c     when computing transform in y-direction, vectorize over x.
c
c     the data a and b are nx x ny,
c     where nx and ny are specified in the parameter statement below.
c     these nx x ny data are stored in arrays
c     with dimensions 0...nxp2, 1...ny
c
c     the argument isign may be any positive or negative integer (not 0)
c     its sign determines the sign of the exponent in the fourier transf
c
c     normalization is performed for either value of isign.
c     the normalization is division by sqrt(nx*ny).
c
c     although the program computes the complex fourier transform, the
c     arithmetic is written out in terms of the real and imaginary parts
c
c     the code is a vector version of the one-dimensional fft
c     program of cooley, lewis, and welch, ieee transactions e-12,
c     no. 1 (march 1965), cited by dahlquist & bjork, numerical methods,
c     prentice hall, englewood cliffs, new jersey, p416
c
      implicit none
c
      double precision a,b
      integer isign,nxp2,nyp1
      dimension a(0:nxp2,0:nyp1),b(0:nxp2,0:nyp1)
c
      integer  i,ip,iv,j,k,l,l2nx,l2ny,nx,ny
      integer  le,le1,nxm1,nym1,nxv2,nyv2
c
      double precision     ang,fnxny,rfnxny,dsign
      double precision     tu1,u1,u2,w1,w2
c
c
c  Fast Fourier Transform local variables
c
      double precision t1x(1)
      common /tx1/ t1x

      double precision t2x(1)
      common /tx2/ t2x

      double precision t3x(1)
      common /tx3/ t3x

      double precision t4x(1)
      common /tx4/ t4x

      double precision t1y(1)
      common /ty1/ t1y

      double precision t2y(1)
      common /ty2/ t2y

      double precision t3y(1)
      common /ty3/ t3y

      double precision t4y(1)
      common /ty4/ t4y
c
      dsign=dble(isign/iabs(isign))
c
      nx = nxp2-2
      ny = nyp1-1
c
      do i = 1, 100
        if (2**i.eq.nx) then
          l2nx = i
          go to 998
        endif
      enddo
 998  continue
      do i = 1, 100
        if (2**i.eq.ny) then
          l2ny = i
          go to 999
        endif
      enddo
 999  continue
c
      nxv2=nx/2
      nxm1=nx-1
      nyv2=ny/2
      nym1=ny-1
c
c*****fourier transform in x-direction:
c*****bit-reversal permutation:
c
      j=1
      do 7 i=1,nxm1
      if(i.ge.j)goto 5
      do 91 iv=1,ny
        t1y(iv)=a(j,iv)
        t3y(iv)=b(j,iv)
        t2y(iv)=a(i,iv)
        t4y(iv)=b(i,iv)
 91   continue
      do 92 iv=1,ny
        a(i,iv)=t1y(iv)
        b(i,iv)=t3y(iv)
        a(j,iv)=t2y(iv)
        b(j,iv)=t4y(iv)
 92   continue
    5 k=nxv2
    6 if(k.ge.j)goto 7
      j=j-k
      k=k/2
      goto 6
    7 j=j+k
c
c*****main loop:
c
      do 20 l=1,l2nx
      le=2**l
      le1=le/2
      u1=1.D+00
      u2=0.D+00
      ang=dsign*3.14159265358979D+00/le1
      w1=dcos(ang)
      w2=dsin(ang)
      do 20 j=1,le1
      do 10 i=j,nx,le
      ip=i+le1
      do 93 iv=1,ny
        t1y(iv)=a(i,iv)
        t3y(iv)=b(i,iv)
        t2y(iv)=a(ip,iv)*u1-b(ip,iv)*u2
        t4y(iv)=a(ip,iv)*u2+b(ip,iv)*u1
 93   continue
      do 94 iv=1,ny
        a(ip,iv)=t1y(iv)-t2y(iv)
        b(ip,iv)=t3y(iv)-t4y(iv)
        a(i ,iv)=t1y(iv)+t2y(iv)
        b(i ,iv)=t3y(iv)+t4y(iv)
 94   continue
   10 continue
      tu1=u1*w1-u2*w2
      u2 =u1*w2+u2*w1
   20 u1=tu1
c
c*****fourier transform in y-direction:
c*****bit-reversal permutation:
c
      j=1
      do 107 i=1,nym1
      if(i.ge.j)goto 105
      do 191 iv=1,nx
        t1x(iv)=a(iv,j)
        t3x(iv)=b(iv,j)
        t2x(iv)=a(iv,i)
        t4x(iv)=b(iv,i)
 191  continue
      do 192 iv=1,nx
        a(iv,i)=t1x(iv)
        b(iv,i)=t3x(iv)
        a(iv,j)=t2x(iv)
        b(iv,j)=t4x(iv)
 192  continue
  105 k=nyv2
  106 if(k.ge.j)goto 107
      j=j-k
      k=k/2
      goto 106
  107 j=j+k
c
c*****main loop:
c
      do 120 l=1,l2ny
      le=2**l
      le1=le/2
      u1=1.D+00
      u2=0.D+00
      ang=dsign*3.14159265358979D+00/le1
      w1=dcos(ang)
      w2=dsin(ang)
      do 120 j=1,le1
      do 110 i=j,ny,le
      ip=i+le1
      do 193 iv=1,nx
        t1x(iv)=a(iv,i)
        t3x(iv)=b(iv,i)
        t2x(iv)=a(iv,ip)*u1-b(iv,ip)*u2
        t4x(iv)=a(iv,ip)*u2+b(iv,ip)*u1
 193  continue
      do 194 iv=1,nx
        a(iv,ip)=t1x(iv)-t2x(iv)
        b(iv,ip)=t3x(iv)-t4x(iv)
        a(iv,i )=t1x(iv)+t2x(iv)
        b(iv,i )=t3x(iv)+t4x(iv)
 194  continue
  110 continue
      tu1=u1*w1-u2*w2
      u2 =u1*w2+u2*w1
  120 u1=tu1
c
c     normalization loop:
c
      fnxny=dsqrt(dble(nx*ny))
      rfnxny=1.0D+00/fnxny
      do 200 i=1,nx
      do 200 iv=1,ny
        a(i,iv)=a(i,iv)*rfnxny
        b(i,iv)=b(i,iv)*rfnxny
  200 continue
c
      return
      end
c
c------------------------------------------------------------
c
      subroutine tridgx(nxp2,nyp1,a,b,c,w,y,work1,work2)
c
      implicit none
      integer i,j,nx,ny
c
      integer nxp2, nyp1
      double precision a,b,c,w,y
      double precision work1,work2
c
      dimension b(0:nxp2,0:nyp1)
      dimension c(0:nxp2,0:nyp1)
      dimension y(0:nxp2,0:nyp1)
      dimension w(0:nxp2,0:nyp1)
c
      dimension work1(0:nxp2,0:nyp1)
      dimension work2(0:nxp2,0:nyp1)
c
      nx = nxp2-2
      ny = nyp1-1
c
      do j=1,ny
         work2(1,j) = a
      enddo

      do i=1,nx-2
        do j=1,ny
          y(i,j) = c(i+1,j)/work2(i,j)
          work2(i+1,j) = a - y(i,j)*b(i,j)
        enddo
      enddo

      do j=1,ny
        work1(1,j)=w(1,j)
      enddo

      do i=2,nx-1
        do j=1,ny
          work1(i,j)= w(i,j) - work1(i-1,j)*y(i-1,j)
        enddo
      enddo


      do j=1,ny
         y(nx-1,j)=work1(nx-1,j)/work2(nx-1,j)
      enddo

      do i=nx-2,1,-1
        do j=1,ny
          y(i,j)=( work1(i,j)-y(i+1,j)*b(i,j) )/work2(i,j)
        enddo
      enddo

      return
      end
c
c------------------------------------------------------------
c
      subroutine tridgy(nxp2,nyp1,a,b,c,w,y,work1,work2)
c
      implicit none
      integer i,j,nx,ny
c
      integer nxp2, nyp1
      double precision a,b,c,w,y
      double precision work1,work2
c
      dimension b(0:nxp2,0:nyp1)
      dimension c(0:nxp2,0:nyp1)
      dimension y(0:nxp2,0:nyp1)
      dimension w(0:nxp2,0:nyp1)
c
      dimension work1(0:nxp2,0:nyp1)
      dimension work2(0:nxp2,0:nyp1)
c
      nx = nxp2-2
      ny = nyp1-1
c
      do i=1,nx
         work2(i,1) = a
      enddo

      do j=1,ny-2
        do i=1,nx
          y(i,j) = c(i,j+1)/work2(i,j)
          work2(i,j+1) = a - y(i,j)*b(i,j)
        enddo
      enddo

      do i=1,nx
         work1(i,1)=w(i,1)
      enddo

      do j=2,ny-1
        do i=1,nx
          work1(i,j)=w(i,j)-work1(i,j-1)*y(i,j-1)
        enddo
      enddo

      do i=1,nx
         y(i,ny-1) = work1(i,ny-1)/work2(i,ny-1)
      enddo

      do j=ny-2,1,-1
        do i=1,nx
          y(i,j) = ( work1(i,j) - y(i,j+1)*b(i,j) ) / work2(i,j)
        enddo
      enddo

      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c
      subroutine cplot (it,alpha,nxx,nyy,mlink,np)
c
c  Produces a raw byte array of fluid simulation
c    Package
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'links.inc'
      include 'domain.inc'
      include 'graph.inc'
c
c  global variables
c
      integer it,nxx,nyy,mlink,np
      byte alpha(nxx,nyy)
c
c  local variables
c
      integer i,ient,j,kjx,kjy,bary
      double precision x1,x2,y1,y2
      character*1024 fnm
c
c  set the background color
c
      do 25 j = 1, nyy
        do 25 i = 1, nxx
          alpha(i,j) = icolr(1)
 25   continue
      do 30 j = 1, nyy
        alpha(1,j) = icolr(2)
        alpha(nxx,j) = icolr(2)
 30   continue
      do 35 i = 1, nxx
        alpha(i,1) = icolr(2)
        alpha(i,nyy) = icolr(2)
 35   continue
c
c  mark the entities.  first the closed entities
c
      do 50, ient = 1, ment
        if (iclose(ient).ge.1 .and. categor(ient).ne.0 ) 
     x          call cpoly(ient,alpha,nxx,nyy)
 50   continue
      do 55, ient = 1, ment
        if (iclose(ient).eq.0 .and. categor(ient).ne.0 ) 
     x          call cpoly(ient,alpha,nxx,nyy)
 55   continue
c
c  mark the links
c
      do 75 i=1,mlink
        if (lcat(i).eq.1) then
          if ( bary(xb(2*q1(i)-1),xb(2*q1(i)))+
     x       bary(xb(2*q2(i)-1),xb(2*q2(i))).eq.0 ) then
            x1 = xb(2*q1(i)-1)/dx
            x2 = xb(2*q2(i)-1)/dx
            y1 = xb(2*q1(i))/dx
            y2 = xb(2*q2(i))/dx
            call mline (x1,y1,x2,y2,icolr(4),alpha,nxx,nyy)
          endif
        endif
 75   continue
c
c  mark the particles
c
      do 100, i = 1, np
        if (pmark(i).ne.0) then
          kjx = max0(idint(px(i)/dx),1)
          kjx = min0(kjx,nxx)
          kjy = max0(idint(py(i)/dy),1)
          kjy = min0(kjy,nyy)
          alpha(kjx,kjy) = pmark(i)
        endif
 100  continue
c
c  Output the data
c
      write(fnm,'(A,".",i4.4)') colorf(1:index(colorf,' ')-1),it
      write(6,999) fnm
      open(unit=55,file=fnm,status='unknown',form='unformatted')
      write (55) alpha
      close(55)
c
c  Update information for visual.dat
c
      open(unit=60,file='ibisdat/cplot.dat',status='unknown')
      write (60,*) it
      close(60)
c
      return
 999  format('  Saving color data in ',A40)
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      subroutine cpoly (ient,alpha,nxx,nyy)
c
c  Plots the polygons
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'domain.inc'
      include 'graph.inc'
c
c  global variables
c
      integer ient,nxx,nyy
      byte alpha(nxx,nyy)
c
c  local variables
c
      integer ione,iend,ixx,iyy,j,iflag,bary
      double precision x1,x2,y1,y2
c
c run around the curve.
c 
      iflag = 0
      ione = firstb(ient)
      iend = lastb(ient)
      if ( categor(ient) .eq. 1) iend = iend-1
      do 20 j = ione, iend-1
        if ( bary(xb(2*j-1),xb(2*j)) +
     x       bary(xb(2*j+1),xb(2*j+2)) .eq.0 ) then
          x1 = xb(2*j-1)/dx
          x2 = xb(2*j+1)/dx
          y1 = xb(2*j)/dy
          y2 = xb(2*j+2)/dy
          call mline (x1,y1,x2,y2,icolr(5),alpha,nxx,nyy)
        else
          iflag = 1
        endif
 20   continue
c
c  if the object is closed then complete curve and
c  mark the interior of the polygon
c
      if (iclose(ient) .eq. 1) then
        if ( bary(xb(2*ione-1),xb(2*ione))+
     x       bary(xb(2*iend-1),xb(2*iend)) .eq.0 ) then
          x1 = xb(2*ione-1)/dx
          x2 = xb(2*iend-1)/dx
          y1 = xb(2*ione)/dy
          y2 = xb(2*iend)/dy
          call mline (x1,y1,x2,y2,icolr(5),alpha,nxx,nyy)
        else
          iflag = 1
        endif
        ixx = idint(centx(ient)/dx)
        iyy = idint(centy(ient)/dy)
        if ( iflag .eq. 0 ) 
     x     call minter (ixx,iyy,icoli(ient),alpha,nxx,nyy)
      elseif (iclose(ient) .eq. 2) then
        ixx = nxx/2
        iyy = nyy/2
        if ( iflag .eq. 0 ) 
     x     call minter (ixx,iyy,icoli(ient),alpha,nxx,nyy)
      endif
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      subroutine minter(ix,iy,coli,alpha,nxx,nyy)
c
c
c  mark the interior of the closed object
c
c  given that (iint,jint) is in the interior, ckeck the
c    four nearest neighbors.  If they are in the interior
c    then
c       1. mark them if they aren't already marked
c       2. store their location as new known interior points
c    Then take the last found unmarked point in the interior
c    and repeat.  The loop is a do-while type construction.
c
      implicit none
      include 'graph.inc'
c
      integer ix,iy,nxx,nyy
      byte alpha(nxx,nyy),coli
c
      integer colb,nlist,ii(4),jj(4),k
c
      nlist = 0
      data ii /1, -1, 0, 0/
      data jj /0, 0, 1, -1/
c
      colb = alpha(ix,iy)
      alpha(ix,iy) = coli
 40   continue
        do 50, k = 1, 4
          if ( alpha(ix+ii(k),iy+jj(k)) .eq. colb ) then
            alpha(ix+ii(k),iy+jj(k)) = coli
            nlist = nlist + 1
            list1(nlist) = ix+ii(k)
            list2(nlist) = iy+jj(k)
          endif
 50     continue
        if (nlist.eq.0) goto 60
        ix = list1(nlist)
        iy = list2(nlist)
        nlist = nlist - 1
        goto 40
 60   continue
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      subroutine mline (x1,y1,x2,y2,col,alpha,nxx,nyy)
c
c  Mark the boundary points making sure of no gap
c
c  global includes
c
      implicit none
      include 'graph.inc'
c
c  global variables
c
      integer nxx,nyy
      byte alpha(nxx,nyy),col
      double precision x1,x2,y1,y2
c
c  local variables
c
      integer j,k,n,nd
      double precision d,dist,f0,f1
c
      d  = dist(x1,y1,x2,y2)
      nd = idint(d+1.0d0)
      do 10, n = 0, nd
        f0 = dble(n)/dble(nd)
        f1 = 1.0d0 - f0
        j  = idint((f0*x1)+(f1*x2))
        k  = idint((f0*y1)+(f1*y2))
        alpha(j,k) = col
 10   continue
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      double precision function dist(x1,y1,x2,y2)
c
c  Calculate the distance between points
c
      double precision x1,y1,x2,y2
      dist = dsqrt(((x1-x2)**2)+((y1-y2)**2))
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      integer function bary(x,y)
c
c  Make sure (x,y) is in the domain
c
      implicit none
      include 'domain.inc'
c
      double precision x,y
c
c  local variables
c
      double precision distx,disty,area,b(4),bmin
c
c  distance from corner to x,y
c
      distx = xlength-x
      disty = ylength-y
c
c  rectangular barycentric coordinates
c
      area = xlength*ylength
      b(1) = x*y/area
      b(2) = distx*y/area
      b(3) = x*disty/area
      b(4) = distx*disty/area
c
c  check to see if they are all positive - i.e.
c  in the domain
c
      bary = 0
      bmin = dmin1(b(1),b(2),b(3),b(4))
      if (bmin .le. 0.0d0) bary = 1
c
      return
      end
c
c-------------------------------------------------------
c
      subroutine partwalk(np)
c
c  randomly move the fluid markers with gaussian random
c  numbers, mean zero, variance one.
c
c-------------------------------------------------------
c
      implicit none
      include 'graph.inc'
      include 'domain.inc'
c
      integer np
c
c  local variables
c
      integer i
      double precision ran1p,r1,r2,rg,var,tpi
c
c  random walk
c
      tpi = 8.0d0*atan(1.0d0)
      do i = 1, np
        var = 2.D+00*pwalk*dlt
        rg  = dsqrt(-2.D+00*var*dlog(max(ran1p(),1.d-8)))
        r1  = rg*dcos(tpi*ran1p())
        r2  = rg*dsin(tpi*ran1p())
        px(i) = px(i) + r1
        py(i) = py(i) + r2
      enddo
c
      return
      end
c
c-------------------------------------------------------
c
      subroutine part(np,npdx,npdy,nxx,nyy)
c
c  deal with the fluid markers at the boundaries
c    via method specified by user
c
c-------------------------------------------------------
c
      implicit none
      include 'graph.inc'
      include 'domain.inc'
c
      integer np,npdx,npdy,nxx,nyy
c
c  local variables
c
      integer i,pnew
      double precision xloc, yloc
c
c  the periodic case, just map particles one domain length
c    x-direction first
c
      if (ipbx.eq.0) then
        pfluxx = 0.0d0
        do 50, i = 1, np
          if (px(i).gt.xlength) px(i) = px(i)-xlength
          if (px(i).lt.0.0d0) px(i) = xlength + px(i)
 50     continue
      endif
c
c  y-direction
c
      if (ipby.eq.0) then
        pfluxy = 0.0d0
        do 100, i = 1, np
          if (py(i).gt.ylength) py(i) = py(i)-ylength
          if (py(i).lt.0.0d0) py(i) = ylength + py(i)
 100    continue
      endif
c
c  the flux case - remove particles that leave
c    x-direction first
c
      if (ipbx.eq.1) then
        do 150, i = 1, np
          if (px(i).gt.xlength .or. px(i).lt.0.0d0) then
            pmark(i) = 0
            px(i) = xlength/2.0d0
            nactm = nactm-1
          endif
 150    continue
      endif
c
c  y-direction
c
      if (ipby.eq.1) then
        do 200, i = 1, np
          if (py(i).gt.ylength .or. py(i).lt.0.0d0) then
            pmark(i) = 0
            py(i) = ylength/2.0d0
            nactm = nactm-1
          endif
 200    continue
      endif
c
c  add a new particles at left or right if its time
c
      npx = (nxx+1)/npdx
      npy = (nyy+1)/npdy
      pdx = xlength/dble(npx)
      pdy = ylength/dble(npy)

      if (ipbx.eq.1) then
        if ( abs(pfluxx) .gt. pdx ) then
          xloc = pdx/2.0d0
          if ( pfluxx.lt.0.0 ) xloc = xlength - (pdx/2.0d0)
          pfluxx = 0.0d0
          pnew = 0
          do 250, i = 1, np
            if (pmark(i).eq.0 .and. pnew.le.npy) then
              px(i) = xloc
              py(i) = (pnew*pdy) + (pdy/2.0d0)
              pmark(i) = 2
              nactm = nactm+1
              pnew = pnew + 1
            endif
 250      continue
        endif
      endif
c
c  add a new particles at top or bottom if its time
c
      if (ipby.eq.1) then
        if ( abs(pfluxy) .gt. pdy ) then
          yloc = pdy/2.0d0
          if ( pfluxy.lt.0.0 ) yloc = ylength - yloc
          pfluxy = 0.0d0
          pnew = 0
          do 300, i = 1, np
            if (pmark(i).eq.0 .and. pnew.lt.npx) then
              px(i) = (pnew*pdx) + (pdx/2.0d0)
              py(i) = yloc
              pmark(i) = 2
              nactm = nactm+1
              pnew = pnew + 1
            endif
 300      continue
        endif
      endif
c
      return
      end
c
c-------------------------------------------------------
c
      subroutine prtsrc(ntsrc,np)
c
c  deal with the fluid markers at the boundaries
c    via method specified by user
c
c-------------------------------------------------------
c
      implicit none
c
      include 'graph.inc'
      include 'domain.inc'
      include 'sink.inc'
c
c  global variables
c
      integer ntsrc,np
c
c  local variables
c
      integer i,j,isink,isource,ksrc
      double precision d,dist,dpx,dpy
c
c  find the sources and sinks
c
      isink = 0
      isource = 0
      do i=1,ntsrc
        if ( src(i) .lt. 0.0d0 ) then
          isink = isink + 1
          sink(isink) = i
        elseif ( src(i) .gt. 0.0d0 ) then
          isource = isource + 1
          source(isource) = i
        endif
      end do
c
c  check all the particles for distance from sinks
c    move from near sinks to near source if close 
c    enough to the sink
c
      ksrc = 0
      do i = 1, np
        if (pmark(i).ne.0) then
          do j = 1, isink
            d  = dist(xsourc(sink(j)),ysourc(sink(j)),px(i),py(i))
            if ( d .le. h/4.0d0 ) then
              ksrc = ksrc+1
              if (ksrc.gt.isource) ksrc = 1
              dpx = xsourc(sink(j))-px(i)
              dpy = ysourc(sink(j))-py(i)
              px(i) = xsourc(source(ksrc))+dpx
              py(i) = ysourc(source(ksrc))+dpy
            endif
          enddo
        endif
      enddo
c
      return
      end
c
c---------------------------------------------------------
c
      subroutine output(it,nxp2,nyp1,f,u,mlink,np)
c
c writes out data for plotting and other use
c
c  global include
c
      implicit none
c
c  global commons
c
      include 'ibp.inc'
      include 'links.inc'
      include 'plat.inc'
      include 'domain.inc'
      include 'graph.inc'
      include 'fluid.inc'
c
c  global variables
c
      integer it,nxp2,nyp1,mlink,np
      double precision f(0:nxp2,0:nyp1,2)
      double precision u(0:nxp2,0:nyp1,2)
c
c  local variables
c
      character*1024 fnm
      integer i,i1,i2,j, klink, kpart, kpoint, inside, iside, nx, ny
      double precision h2, vlmax, flmax, q, vlmin, flmin, xout, yout
      double precision x1, y1, x2, y2, maptox, maptoy
c
      nx = nxp2-2
      ny = nyp1-1
c
c  terminal message
c
      write(fnm,'(A,".*.",i4.4)') linef(1:index(linef,' ')-1),it
      write(6,998) fnm
c
c  local max/min initialize
c
      vlmax = 0.0d0
      flmax = 0.0d0
      vlmin = vmaxg
      flmin = fmaxg
c
c  write out field vectors file
c
      h2 = h*h
      write(fnm,'(A,".v.",i4.4)') linef(1:index(linef,' ')-1),it
      open(unit=18,file=fnm,status='unknown')
      do j = 1, ny, ivdens
        yout = ymin + (h*j)
        do i = 1, nx, ivdens
          xout = xmin + (h*i)
          write(18,1008) xout, yout, u(i,j,1), u(i,j,2),
     x        f(i,j,1)*h2, f(i,j,2)*h2
          vlmax = dmax1(vlmax,sqrt(u(i,j,1)**2 + u(i,j,2)**2)) 
          flmax = dmax1(flmax,h2*sqrt(f(i,j,1)**2 + 
     x                  f(i,j,2)**2)) 
          vlmin = dmin1(vlmin,sqrt(u(i,j,1)**2 + u(i,j,2)**2)) 
          flmin = dmin1(flmin,h2*sqrt(f(i,j,1)**2 + 
     x                  f(i,j,2)**2)) 
        enddo
      enddo
      close(18)
c
c  write out ibp file
c
      kpoint = 0
      write(fnm,'(A,".i.",i4.4)') linef(1:index(linef,' ')-1),it
      open(unit=18,file=fnm,status='unknown')
      do j=1,ment
        if ( categor(j).ne.0 ) then
          i1=firstb(j)
          i2=lastb (j)
          if ( i2-i1 .ge. 1 ) then
            do i=i1,i2
              xout = maptox(xb(2*i-1)+xmin)
              yout = maptoy(xb(2*i)+ymin)
              write(18,1006) xout,yout,usb1(i)*h2,usb2(i)*h2
     x            ,-gfrcb1(i),-gfrcb2(i)
              kpoint = kpoint + 1
              q = sqrt((usb1(i)*h2)**2 + (usb2(i)*h2)**2)
              vlmax = dmax1(vlmax,q)
              flmax = dmax1(flmax,sqrt(gfrcb1(i)**2 + gfrcb2(i)**2))
              vlmin = dmin1(vlmin,q)
              flmin = dmin1(flmin,sqrt(gfrcb1(i)**2 + gfrcb2(i)**2))
            enddo
          endif
        endif
      enddo
      close(18)
c
c  write out link file
c    first the intra entity links
c
      klink = 0
      write(fnm,'(A,".l.",i4.4)') linef(1:index(linef,' ')-1),it
      open(unit=18,file=fnm,status='unknown')
      do j=1,ment
        if ( categor(j).ne.0 ) then
          i1=firstb(j)
          i2=lastb (j)
          if ( i2-i1 .ge. 1 ) then
            do i=i1,i2-1
              x1 = xb(2*i-1)+xmin
              y1 = xb(2*i)+ymin
              x2 = xb(2*i+1)+xmin
              y2 = xb(2*i+2)+ymin
              iside = inside(x1,y1)+inside(x2,y2)
              if (iside.le.1) then
                write(18,1007) x1,y1,x2,y2
                klink = klink+1
              elseif (iside.eq.2) then
                write(18,1007) maptox(x1),maptoy(y1),
     x                 maptox(x2),maptoy(y2)
                klink = klink+1
              endif
            enddo
            if ( iclose(j).ne.0 ) then
              x1 = xb(2*i1-1)+xmin
              y1 = xb(2*i1)+ymin
              x2 = xb(2*i2-1)+xmin
              y2 = xb(2*i2)+ymin
              iside = inside(x1,y1)+inside(x2,y2)
              if (iside.le.1) then
                write(18,1007) x1,y1,x2,y2
                klink = klink+1
              elseif (iside.eq.2) then
                write(18,1007) maptox(x1),maptoy(y1),
     x                 maptox(x2),maptoy(y2)
                klink = klink+1
              endif
            endif
c
c  plot periodic lines
c
            if ( categor(j).eq.-1 ) then
              x1 = xb(2*i1-1)+xmin
              y1 = xb(2*i1)+ymin
              x2 = xb(2*i2-1)+xmin-xlength
              y2 = xb(2*i2)+ymin
              write(18,1007) x1,y1,x2,y2
              klink = klink+1
              write(18,1007) x1+xlength,y1,x2+xlength,y2
              klink = klink+1
            endif

          endif
        endif
      enddo
c
c  now the inter entity links
c
      do j=1,mlink
        if (lcat(j).eq.1) then
          x1 = xb(2*q1(j)-1)+xmin
          y1 = xb(2*q1(j))+ymin
          x2 = xb(2*q2(j)-1)+xmin
          y2 = xb(2*q2(j))+ymin
          iside = inside(x1,y1)+inside(x2,y2)
          if (iside.le.1) then
            write(18,1007) x1,y1,x2,y2
            klink = klink+1
          elseif (iside.eq.2) then
            write(18,1007) maptox(x1),maptoy(y1),
     x                     maptox(x2),maptoy(y2)
            klink = klink+1
          endif
        endif
      enddo
      close(18)
c
c  write out marker particles
c
      kpart = 0
      write(fnm,'(A,".m.",i4.4)') linef(1:index(linef,' ')-1),it
      open(unit=18,file=fnm,status='unknown')
      do j = 1, np
        if ( pmark(j) .ne. 0 ) then
          write(18,1004) px(j)+xmin,py(j)+ymin,pmark(j)
          kpart = kpart+1
        endif
      enddo
      close(18)
c
c  write out the header file
c
      write(fnm,'(A,".h.",i4.4)') linef(1:index(linef,' ')-1),it
      open(unit=18,file=fnm,status='unknown')
      write(18,1000) xmin, xmax, ymin, ymax, time
      write(18,1010) nx/ivdens, ny/ivdens, h
      write(18,1002) kpoint, 0, 2
      write(18,1001) 0, 0
      write(18,1001) nx*ny, 2
      write(18,1002) klink, kpart
      write(18,1000) vlmax, flmax
      write(18,1000) vlmin, flmin
      write(18,1000) vlmax, flmax
      write(18,1000) vlmin, flmin
      close(18)
      fmax = flmax
c
c update visual.dat
c
      open(unit=65,file='ibisdat/lplot.dat',status='unknown')
      write (65,*) it
      close(65)
c
      fmaxg = dmax1(flmax,fmaxg)
      vmaxg = dmax1(vlmax,vmaxg)
      umax = vmaxg
      open(19,file='ibisdat/vmax.dat',status='unknown')
      write(19,*) 'fvmax1   ',sngl(vmaxg)
      write(19,*) 'fvmax2   ',sngl(fmaxg)
      write(19,*) 'pvmax1   ',sngl(vmaxg)
      write(19,*) 'pvmax2   ',sngl(fmaxg)
      close(19)
c
c shutdown
c
      return
 998  format('  Saving line velocity data in ',A40)
 999  format('  Saving line force data in ',A40)
 1000 format(6(1x,e15.7))
 1001 format(2(1x,i6))
 1002 format(3(1x,i6))
 1003 format(5(1x,i6))
 1004 format(2(1x,e15.7),i6)
 1005 format(3(1x,i6),1x,e15.7)
 1006 format(6(1x,e15.7))
 1007 format(4(1x,e15.7))
 1008 format(6(1x,e15.7))
 1010 format(2(1x,i6),1x,e15.7)
      end
c
c----------------------------------------------------------------
c
      subroutine rsdata(rsfile,nxp2,nyp1,u,mlink,np)
c
c  restart data output
c  global include
c
      implicit none
c
c  global common variables
c
      include 'active.inc'
      include 'domain.inc'
      include 'fluid.inc'
      include 'graph.inc'
      include 'ibp.inc'
      include 'links.inc'
      include 'plat.inc'
c
c  global variable
c
      integer rsfile,nxp2,nyp1,mlink,np
      double precision u(0:nxp2,0:nyp1,2)
c
c  local variables
c
      integer ient,inum,ione,iend,jpt,ilink,ipart
      character*1024 fnm
c
c  open the restart data file
c
      write(6,*) ' Saving restart data'
c
      open(unit=rsfile,file='ibisdat/rsd',status='unknown')
      write (rsfile,*) 'error'
      close (rsfile)
c
      write(fnm,'("ibisdat/rsdata.",i2.2)') rsfile
      open(unit=rsfile,file=fnm,form='unformatted')
c
c  write the name list data
c
      write(rsfile) umax,nt,time,ccount,lcount
c
c  write the number of active entities
c
      inum = 0
      do ient = 1, ment
        if ( categor(ient).ne.0 ) inum = inum+1
      enddo
      write(rsfile) inum
      write(rsfile) jfil,jplat,jpoly,nlinke
c
c  write the entity data
c
      inum = 0
      do ient = 1, ment
        if ( categor(ient).ne.0 ) then
          ione = firstb(ient)
          iend = lastb(ient)
          write(rsfile)'rsrt'
          write(rsfile) iend-ione+1, iclose(ient), 
     x      categor(ient), activ1(ient), activ2(ient),
     x      xcm(ient)+xmin, ycm(ient)+ymin
          do jpt = ione, iend
            inum = inum+1
            imap(jpt) = inum
            write(rsfile) inum, 
     x        xb(2*jpt-1)+xmin, xb(2*jpt)+ymin,
     x        xteth(jpt)+xmin, yteth(jpt)+ymin,
     x        dlp(jpt), swp(jpt), csp(jpt), sp(jpt),
     x        sthp(jpt)
          enddo
        endif
      enddo
c
c  write the link data
c
      write(rsfile) nlink
      do ilink = 1, mlink
        if ( lcat(ilink).eq.1 ) then
          write(rsfile) imap(q1(ilink)),imap(q2(ilink)),
     x        p1(ilink),p2(ilink),r1(ilink),r2(ilink),
     x        stiffb(ilink),sbreak(ilink),restb(ilink)
        endif
      enddo
c
c  write the fluid velocities
c
      write (rsfile) u
c
c  write the particles
c
      do ipart = 1, np
        write(rsfile) px(ipart), py(ipart), pmark(ipart)
      enddo
      close(rsfile)
c
      open(unit=rsfile,file='ibisdat/rsd',status='unknown')
      write (rsfile,*) 'done'
      close(rsfile)
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      subroutine prout (nt,fluxx,fluxy,vmax,cfl,time)
c
c   inform the screen
c
      integer nt
      double precision fluxx,fluxy,vmax,cfl,time
c
      write(6,*)' -----------------------------------------------'
      write(6,*)' nt    = ',nt
      write(6,*)' time  = ',time
      write(6,*)' fluxx = ',fluxx
      write(6,*)' fluxy = ',fluxy
      write(6,*)' vmax  = ',vmax
      write(6,*)' cfl   = ',cfl
c
      return
      end
c
c
      subroutine ohnoo 
c
c   inform the screen that you blew it
c
      write(6,*)' '
      write(6,*)' '
      write(6,*)' -----------------------------------------------'
      write(6,*)' -----------------------------------------------'
      write(6,*)' -----------------------------------------------'
      write(6,*)' '       
      write(6,*)' The calculation is now unstable'
      write(6,*)'     Failure is imminent'
      write(6,*)' '
      write(6,*)'     Decrease the time step or spring constants'
      write(6,*)'     before running the calculation again'
      write(6,*)' '
      write(6,*)'     Saving the final state for inspection'
      write(6,*)' '
      write(6,*)' Adios'
      write(6,*)' '
      write(6,*)' -----------------------------------------------'
      write(6,*)' -----------------------------------------------'
      write(6,*)' -----------------------------------------------'
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      subroutine upflux(nxp2,nyp1,u)
c
c  update fluid fluxes
c
      implicit none
      include 'domain.inc'
      include 'fluid.inc'
      include 'graph.inc'
c
      integer nxp2, nyp1
      double precision u(0:nxp2,0:nyp1,1:2)
c
      integer i, j, nx
      double precision q,velxm,velym
c
      nx = nxp2-2
c
      vmax  = 0.0d0
      fluxx = 0.0d0
      fluxy = 0.0d0
      velxm = 0.0d0
      velym = 0.0d0
      do j = ibtube, ittube
        do i =  0, nx
          q = sqrt(u(i,j,1)**2 + u(i,j,2)**2)
          vmax = max(vmax,q)
        enddo
      enddo
c
      do j = ibtube, ittube
        fluxx = fluxx + h*u(2,j,1)
        velxm = velxm + u(2,j,1)
      enddo
c
      do j = 1, nx
        fluxy=fluxy + h*u(j,2,2)
        velym = velym + u(j,2,2)
      enddo
      pfluxx = pfluxx+(velxm*dlt/dble(ittube-ibtube))
      pfluxy = pfluxy+(velym*dlt/dble(nx))
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      subroutine pdebug (idebug)
c
c  print the debug messages
c
      implicit none
      include 'ibp.inc'
c
      integer idebug(0:4)
c
      integer i
c
      if (idebug(1).eq.1) then
        do i=1,npts
          write(100,9990) xpt(i),ypt(i),gfrc1(i),gfrc2(i)
        enddo
      endif
c
c the immersed boundary points and their link information
c
      if (idebug(2).eq.1) then
        do i=1,npts
          write(103,9991) xb(2*i-1),xb(2*i),xteth(i),yteth(i),
     x         dlp(i),swp(i),csp(i),sp(i),sthp(i)
        enddo
      endif
c
      return
 9990 format(8f15.7)
 9991 format(4f10.3,6e11.3)
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      integer function inside(x,y)
c
c  Is (x,y) is in the domain?
c
      include 'domain.inc'
c
      double precision x,y
c
      inside = 0
      if ( x.lt.xmin .or. x.gt.xmax ) inside = 1
      if ( y.lt.ymin .or. y.gt.ymax ) inside = 1
c 
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      double precision function maptox(x)
c
      include 'domain.inc'
c
c  Map x to the domain.
c
      double precision x
c
      maptox = x
      if ( x.lt.xmin ) maptox = x + xlength
      if ( x.gt.xmax ) maptox = x - xlength
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      double precision function maptoy(y)
c
      include 'domain.inc'
c
c  Map y to the domain.
c
      double precision y
c
      maptoy = y
      if ( y.lt.ymin ) maptoy = y + ylength
      if ( y.gt.ymax ) maptoy = y - ylength
c
      return
      end
c
      subroutine parea
c
      implicit none
      include 'ibp.inc'
c
c  local variables
c
      integer j, jent, jpt
      double precision areamax,areamin,c1,c2,d1,d2,aarea
c
      areamin= 1000.0D+00
      areamax=-1000.0D+00
      do 100 jent=1,ment
        if(categor(jent).ge.1.and.categor(jent).le.3)then
c
c compute area
c
           aarea = 0.0D+00
           do 20 j = 1, lastb(jent)-firstb(jent)
              jpt=firstb(jent)+j-1
              c1 = xb(2*jpt-1)-xcm(jent)
              c2 = xb(2*jpt  )-ycm(jent)
              d1 = xb(2*jpt+1)-xcm(jent)
              d2 = xb(2*jpt+2)-ycm(jent)
              aarea=aarea+.5D+00*dabs(c1*d2-c2*d1)
 20        continue
           jpt=firstb(jent)
           c1=xb(2*jpt-1)-xcm(jent)
           c2=xb(2*jpt  )-ycm(jent)
           jpt=lastb(jent)+1
           d1=xb(2*jpt-1)-xcm(jent)
           d2=xb(2*jpt  )-ycm(jent)
           aarea=aarea+.5D+00*dabs(c1*d2-c2*d1)
           areamin=dmin1(aarea/area(jent),areamin)
           areamax=dmax1(aarea/area(jent),areamax)
        endif
 100  continue
      write(16,101)areamin,areamax
 101  format(' areamin=',f13.7,' areamax=',f13.7)
      return
      end
c
      subroutine adhere(cpflag)
c
c this subroutine determines whether new links should be formed between
c platelet rings and wall filaments and if so, creates these links.
c
c this version permits multiple links between a given platelet
c and wall
c
c it expects filament 1 (entity 1) to be on bottom and 
c            filament 2 (entity 2) to be on top.
c
      implicit none
c
      include 'ibp.inc'
      include 'links.inc'
      include 'domain.inc'
      include 'plat.inc'
c
c  global variables
c
      integer cpflag
c
c  local variables
c
      integer i,ic,ient,ilink,imin,i1,i2
      integer j,jc,jent,jmin,j1,j2
      integer ni,nj,nwhere
      double precision dist,sdist,smin,xi,xj,yi,yj
c
c now see if x-coordinate of cm is in the x-interval which corresponds 
c to the injured part of the wall.
c
      do 200 ient=1,ment
         if(categor(ient).ge.1)then
         if(xcm(ient).gt.xadhmin.and.xcm(ient).lt.xadhmax)then
c
c see if platelet should adhere.
c
c first determine which wall is closer.
c
           if(ycm(ient).le.ycent)then
              jent = 1
           else
              jent = 2
           endif
c
c see if platelet is within sadh of nominal y-coor of closer wall
c
           if(jent.eq.1)then
              dist=ycm(ient)-ybot
           else
              dist=ytop-ycm(ient)
           endif
           if(dist.gt.sadh)goto 200
c
c next determine whether platelet ient is already linked to wall jent
c and, if so, how many links join them.
c
        if(categor(ient).eq.3)then
           i=ient
           j=jent
           nwhere=((i-2)*(i-1))/2 + j
           if(numlink(nwhere).ge.maxlink)goto 200
        endif
c
c find point on platelet and point on injured portion of
c this wall which are closest.
c
             i1=firstb(ient)
             i2=lastb (ient)
             ni=i2-i1+1
             j1=afirstb(jent)
             j2=alastb (jent)
             nj=j2-j1+1
             imin=0
             jmin=0
             smin=1000.D+00
             do 20 ic=1,ni
                i=i1 + ic - 1
             if(status(i+1).eq.0)then
                xi=xb(2*i-1)
                yi=xb(2*i  )
             do 15 jc=1,nj
                j=j1 + jc - 1
             if(status(j+1).eq.0)then
                xj=xb(2*j-1)
                yj=xb(2*j  )
                sdist =dsqrt((xi-xj)**2 + (yi-yj)**2)
                if(sdist.lt.smin)then
                     smin=sdist
                     imin=ic
                     jmin=jc + afirstb(jent)-firstb(jent)
                endif
             endif
 15          continue    
             endif
 20          continue
c
c now see if the distance between these nearest points is close enough.
c
             write(6,*)' smin=',smin,' sagap=',sagap
             if(smin.lt.sagap)then
c
c form adhesion link between point imin on platelet ient and point jmin 
c on filament jent.  note that jent < ient by our numbering convention.
c
                ilink =lkas
                p1(lkas)=jent
                r1(lkas)=jmin
                p2(lkas)=ient
                r2(lkas)=imin
                stiffb(lkas)=stiffa
                sbreak(lkas)=sbreaka
                restb (lkas)=rlinka
c
             status( firstb(ient)+imin)=1
             status( firstb(jent)+jmin)=1
c
                  write(6 ,31)ient,imin,jent,jmin,smin
                  write(16,31)ient,imin,jent,jmin,smin
c
             if(categor(ient).eq.3)then
c
c ient is linked, just update link-data structure
c
                  call chgcate(listlkf,listlkb,ilink,lkas,lkzas,
     +                         lkl,lkzl)
                  nwhere=((ient-2)*(ient-1))/2 + jent
                  if(numlink(nwhere).eq.0)then
c
c this link results in a new pair of coherent entities
c update coherent pair data structure
c
                     cpflag=1
                     i=nas
                     pltlnk(i)=jent
                     call chgcate(nxtlnkf,nxtlnkb,i,nas,nzas,
     +                  fstlnk(ient),lstlnk(ient))
                     j=nas
                     pltlnk(j)=ient
                     call chgcate(nxtlnkf,nxtlnkb,j,nas,nzas,
     +                  fstlnk(jent),lstlnk(jent))
                  endif
                  numlink(nwhere)=numlink(nwhere)+1

                  nlink=nlink+1
c
             elseif(categor(ient).eq.2)then
c
c ient is activated but not linked.  change category of ient and
c update link-data structure
c
                  categor(ient)=3
                  call chgcate(listf,listb,ient,pac,pzac,pl,pzl)
                  ilink =lkas
                  call chgcate(listlkf,listlkb,ilink,lkas,lkzas,
     +                         lkl,lkzl)
                  nwhere=((ient-2)*(ient-1))/2 + jent
                  if(numlink(nwhere).eq.0)then
c
c this link results in a new pair of coherent entities
c update coherent pair data structure
c
                     cpflag=1
                     i=nas
                     pltlnk(i)=jent
                     call chgcate(nxtlnkf,nxtlnkb,i,nas,nzas,
     +                  fstlnk(ient),lstlnk(ient))
                     j=nas
                     pltlnk(j)=ient
                     call chgcate(nxtlnkf,nxtlnkb,j,nas,nzas,
     +                  fstlnk(jent),lstlnk(jent))
                  endif
                  numlink(nwhere)=numlink(nwhere)+1
                  nlink=nlink+1
c
             elseif(categor(ient).eq.1)then
c
c ient is nonactivated.  change category of ient and
c update link-data structure
c
                  categor(ient)=3
                  call chgcate(listf,listb,jent,pac,pzac,pl,pzl)
                  ilink=lkas
                  call chgcate(listlkf,listlkb,ilink,lkas,lkzas,
     +                         lkl,lkzl)
                  nwhere=((ient-2)*(ient-1))/2 + jent
                  if(numlink(nwhere).eq.0)then
c
c this link results in a new pair of coherent entities
c update coherent pair data structure
c
                     cpflag=1
                     i=nas
                     pltlnk(i)=jent
                     call chgcate(nxtlnkf,nxtlnkb,i,nas,nzas,
     +                  fstlnk(ient),lstlnk(ient))
                     j=nas
                     pltlnk(j)=ient
                     call chgcate(nxtlnkf,nxtlnkb,j,nas,nzas,
     +                  fstlnk(jent),lstlnk(jent))
                  endif
                  numlink(nwhere)=numlink(nwhere)+1
                  nlink=nlink+1
             endif
             endif
         endif
         endif
 200  continue
c
   31 format('  adhere   ient=',i5,' imin=',i5,' jent=',i5,' jmin=',i5,
     +       ' smin=',f13.7)
      write(6,*)' adhere nlink=',nlink
      return
      end
c
      subroutine cohere(cpflag)
c
c this subroutine determines whether new links should be formed between
c platelet rings and if so, creates these links.
c
c this version allows multiple links between a given pair of platelets
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'links.inc'
      include 'plat.inc'
c
c  global variables
c
      integer cpflag
c
c  local variables
c
      integer i,ic,ient,ilink,imin,i1,i2
      integer j,jc,jent,jmin,j1,j2
      integer ni,nj,nbig,nsmall,nwhere
      double precision scm,sdist,smin
      double precision xi,xj,yi,yj
c
c  test for closeness of the cm for all pairs of platelet rings present
c  which are activated (categor=2) or linked (categor=3)
c
      do 200 ient=1,ment
      if(categor(ient).ge.2)then
         do 190 jent=ient+1,ment
         if(categor(jent).ge.2)then
            scm = dsqrt((xcm(ient)-xcm(jent))**2 
     +               + (ycm(ient)-ycm(jent))**2)
            if(scm.lt.scoh)then
c
c the cm platelets ient and jent are close enough to make linking a
c possibility.  
c
c check their state: non-linked or linked.
c
c if both linked, see if they are already linked to one-another.
c
c see also if the separation between the closest (unlinked) points 
c on the two platelets is small enough for a link to form.
c
        if(categor(ient).eq.(3).and.categor(jent).eq.(3))then
c
c both platelets are linked, see if they are already linked to
c one another.
c
             nsmall = min0(ient,jent)
             nbig   = max0(ient,jent)
             nwhere=((nbig-2)*(nbig-1))/2 + nsmall
             if(numlink(nwhere).ge.maxlink)goto 190
        endif
c
c compute gap between nearest unlinked points on the two platelets.
c
             i1=firstb(ient)
             i2=lastb (ient)
             ni=i2-i1+1
             j1=firstb(jent)
             j2=lastb (jent)
             nj=j2-j1+1
             imin=0
             jmin=0
             smin=1000.D+00
             do 70 ic=1,ni
                i=i1 + ic - 1
             if(status(i+1).eq.0)then
                xi=xb(2*i-1)
                yi=xb(2*i  )
             do 60 jc=1,nj
                j=j1 + jc - 1
             if(status(j+1).eq.0)then
                xj=xb(2*j-1)
                yj=xb(2*j  )
                sdist =dsqrt((xi-xj)**2 + (yi-yj)**2)
                if(sdist.lt.smin)then
                     smin=sdist
                     imin=ic
                     jmin=jc
                endif
             endif
 60          continue    
             endif
 70          continue
c
             if(smin.lt.scgap)then
c
c form link between point imin on ient and point jmin on jent
c
                ilink =lkas
                if(ient.lt.jent)then
                  p1(lkas)=ient
                  r1(lkas)=imin
                  p2(lkas)=jent
                  r2(lkas)=jmin
                else
                  p1(lkas)=jent
                  r1(lkas)=jmin
                  p2(lkas)=ient
                  r2(lkas)=imin
                endif
c
                stiffb(lkas)=stiffc
                sbreak(lkas)=sbreakc
                restb (lkas)=rlinkc
c
             status(firstb(ient)+imin)=1
             status(firstb(jent)+jmin)=1
c
             if(categor(ient).eq.(3).and.categor(jent).eq.(3))then
                write(6,* )' COHERE     LL'
c
c both are already linked, just update link-data structure
c
                  call chgcate(listlkf,listlkb,ilink,lkas,lkzas,
     +                         lkl,lkzl)
                  nwhere=((p2(ilink)-2)*(p2(ilink)-1))/2 + p1(ilink)
                  if(numlink(nwhere).eq.0)then
c
c this link results in a new pair of coherent entities
c update coherent pair data structure
c
                     cpflag=1
                     i=nas
                     pltlnk(i)=jent
                     call chgcate(nxtlnkf,nxtlnkb,i,nas,nzas,
     +                  fstlnk(ient),lstlnk(ient))
                     j=nas
                     pltlnk(j)=ient
                     call chgcate(nxtlnkf,nxtlnkb,j,nas,nzas,
     +                  fstlnk(jent),lstlnk(jent))
                  endif
                  numlink(nwhere)=numlink(nwhere)+1
                  nlink=nlink+1
                  write(6 ,31)ient,imin,jent,jmin,scm,smin
                  write(16,31)ient,imin,jent,jmin,scm,smin
c
             elseif(categor(ient).eq.(2).and.categor(jent).eq.(3))then
c
                write(6,* )' COHERE     AL'
c
c jent is already linked, ient is not linked.  change category of 
c ient and update link-data structure
c
                  write(6 ,31)ient,imin,jent,jmin,scm,smin
                  write(16,31)ient,imin,jent,jmin,scm,smin
                  categor(ient)=3
                  call chgcate(listf,listb,ient,pac,pzac,pl,pzl)
                  ilink =lkas
                  call chgcate(listlkf,listlkb,ilink,lkas,lkzas,
     +                         lkl,lkzl)
                  nwhere=((p2(ilink)-2)*(p2(ilink)-1))/2 + p1(ilink)
                  if(numlink(nwhere).eq.0)then
c
c this link results in a new pair of coherent entities
c update coherent pair data structure
c
                     cpflag=1
                     i=nas
                     pltlnk(i)=jent
                     call chgcate(nxtlnkf,nxtlnkb,i,nas,nzas,
     +                  fstlnk(ient),lstlnk(ient))
                     j=nas
                     pltlnk(j)=ient
                     call chgcate(nxtlnkf,nxtlnkb,j,nas,nzas,
     +                  fstlnk(jent),lstlnk(jent))
                  endif
                  numlink(nwhere)=numlink(nwhere)+1
                  nlink=nlink+1
c
             elseif(categor(ient).eq.(3).and.categor(jent).eq.(2))then
c
                write(6,* )' COHERE     LA'
c
c jent is already linked, ient is not linked.  change category of 
c ient and update link-data structure
c
                  write(6 ,31)ient,imin,jent,jmin,scm,smin
                  write(16,31)ient,imin,jent,jmin,scm,smin
                  categor(jent)=3
                  call chgcate(listf,listb,jent,pac,pzac,pl,pzl)
                  ilink=lkas
                  call chgcate(listlkf,listlkb,ilink,lkas,lkzas,
     +                         lkl,lkzl)
                  nwhere=((p2(ilink)-2)*(p2(ilink)-1))/2 + p1(ilink)
                  if(numlink(nwhere).eq.0)then
c
c this link results in a new pair of coherent entities
c update coherent pair data structure
c
                     cpflag=1
                     i=nas
                     pltlnk(i)=jent
                     call chgcate(nxtlnkf,nxtlnkb,i,nas,nzas,
     +                  fstlnk(ient),lstlnk(ient))
                     j=nas
                     pltlnk(j)=ient
                     call chgcate(nxtlnkf,nxtlnkb,j,nas,nzas,
     +                  fstlnk(jent),lstlnk(jent))
                  endif
                  numlink(nwhere)=numlink(nwhere)+1
                  nlink=nlink+1
c
             elseif(categor(ient).eq.(2).and.categor(jent).eq.(2))then
c
                write(6,* )' COHERE     AA'
c
c neither is linked.
c
                  write(6 ,31)ient,imin,jent,jmin,scm,smin
                  write(16,31)ient,imin,jent,jmin,scm,smin
                  categor(ient)=3
                  categor(jent)=3
                  call chgcate(listf,listb,ient,pac,pzac,pl,pzl)
                  call chgcate(listf,listb,jent,pac,pzac,pl,pzl)
                  ilink =lkas
                  call chgcate(listlkf,listlkb,ilink,lkas,lkzas,
     +                         lkl,lkzl)
                  nwhere=((p2(ilink)-2)*(p2(ilink)-1))/2 + p1(ilink)
                  if(numlink(nwhere).eq.0)then
c
c this link results in a new pair of coherent entities
c update coherent pair data structure
c
                     cpflag=1
                     i=nas
                     pltlnk(i)=jent
                     call chgcate(nxtlnkf,nxtlnkb,i,nas,nzas,
     +                  fstlnk(ient),lstlnk(ient))
                     j=nas
                     pltlnk(j)=ient
                     call chgcate(nxtlnkf,nxtlnkb,j,nas,nzas,
     +                  fstlnk(jent),lstlnk(jent))
                  endif
                  numlink(nwhere)=numlink(nwhere)+1
                  nlink=nlink+1
             endif
            endif
         endif
         endif
 190     continue
      endif
 200  continue   
c
   31 format('  cohere   ient=',i5,' imin=',i5,' jent=',i5,' jmin=',i5,
     +       ' scm =',f13.7,' smin=',f13.7)
      write(6,*)' cohere nlink=',nlink
c
       return
       end
c
      subroutine remove(cpflag,nring)
c
c this subroutine determines whether any platelet should be removed
c at the downstream end of the domain.
c the determination is based on the location of the platelet's
c center of mass.
c
c this subroutine also determines which links should be broken
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'links.inc'
      include 'domain.inc'
      include 'plat.inc'
c
c  global variables
c
      integer cpflag,nring
c
      integer i,ient,i1,is
      integer j,j1,j2
      integer lp,update,lps,np1,np2,npp,nwhere
c
c  local variables
c
      double precision sdist,smin
      double precision x1,x2,y1,y2
      double precision xp1,xp2,xremov
c
c
      xremov = xmax - 1.0D+00*diam
c
c next determine whether any nonactivated or activated, but
c not linked, platelets have xcm greater than xremov.
c if so, the platelet should be removed.
c
      do 20 ient=1,ment
      if(xcm(ient).gt.xremov)then
        if(categor(ient).eq.1.D+00)then
c
c remove this nonactivated platelet
c
           call chgcatb(listf,listb,ient,pua,pzua,pas,pzas)
           categor(ient)=0
           nent=nent-1
c
        elseif(categor(ient).eq.2)then
c
c remove this activated platelet
c
           call chgcatb(listf,listb,ient,pac,pzac,pas,pzas)
           categor(ient)=0
           nent=nent-1
        endif
      endif
 20   continue
c
c now check whether any linked platelet should be removed
c
c initialize changes array to 1
c
      do 30 j=1,ment
      changes(j)=1
 30   continue
c
c set changes=2 for any platelet which is currently linked.
c
      smin=10000.D+00
      lp=lkl
 35   continue
      if(lp.eq.0)go to 40
      changes(p1(lp))=2
      changes(p2(lp))=2
      lp=listlkf(lp)
      goto 35
 40   continue
c
c now sweep through the list of links.
c
      lp=lkl
 50   continue
c
      update = 0
      if(lp.eq.0)go to 210
      lps=listlkf(lp)
      np1=p1(lp)
      np2=p2(lp)
c
      if(categor(np1).eq.(-1))then
c
c link is a platelet-wall link
c
      xp2=xcm(np2)
      changes(np1)=1
c
c compute length of link
c
      j1 = firstb(np1)+r1(lp)-1
      j2 = firstb(np2)+r2(lp)-1
      x1 = xb(2*j1-1)
      y1 = xb(2*j1  )
      x2 = xb(2*j2-1)
      y2 = xb(2*j2  )
      sdist  = dsqrt((x1-x2)**2 + (y1-y2)**2)
      smin=dmin1(smin,sdist/sbreak(lp))
c
      if(xp2.gt.xremov)then
c
c np2 is beyond xremov:
c if not already changed, change category for np2 to as.
c set changes(np2)=3.
c leave changes(np1)=2.
c
          update=1
          status(firstb(np1)+r1(lp))=0
          if(changes(np2).ne.3)then
              categor(np2)=0
              call chgcatb(listf,listb,np2,pl,pzl,pas,pzas)
              nent=nent-1
              changes(np2)=3
              i1 = firstb(np2)
              do 58 j=1,nring
                 status(i1+j)=0
 58           continue
          endif
c
      elseif(sdist.gt.sbreak(lp))then
c
c length is greater than breaking length.
c break the link but do not remove the platelet
c
         status(j1+1)=0
         status(j2+1)=0
         update=1
      else
c
c platelet is not beyond xremov and tension is not 
c great enough to break the link.
c leave this link alone.
c
         changes(np2)=1
      endif
c
      else
c
c link is a platelet-platelet link
c
      xp1=xcm(np1)
      xp2=xcm(np2)
c
c compute length of link
c
      j1 = firstb(np1)+r1(lp)-1
      j2 = firstb(np2)+r2(lp)-1
      x1 = xb(2*j1-1)
      y1 = xb(2*j1  )
      x2 = xb(2*j2-1)
      y2 = xb(2*j2  )
      sdist  = dsqrt((x1-x2)**2 + (y1-y2)**2)
c
c check whether either of xp1 or xp2 is greater than xremov.  
c if so, break the link.
c
      if(xp1.gt.xremov.and.xp2.gt.xremov)then
c
c both are greater than xremov.
c
c if not already changed, change category for both platelets
c to as, update listf and categor; and set changes to 3
c
          update=1
          if(changes(np1).ne.3)then
              categor(np1)=0
              call chgcatb(listf,listb,np1,pl,pzl,pas,pzas)
              nent=nent-1
              changes(np1)=3
              i1 = firstb(np1)
              do 105 j=1,nring
                 status(i1+j)=0
 105          continue
          endif
c
          if(changes(np2).ne.3)then
              categor(np2)=0
              call chgcatb(listf,listb,np2,pl,pzl,pas,pzas)
              nent=nent-1
              changes(np2)=3
              i1 = firstb(np2)
              do 106 j=1,nring
                 status(i1+j)=0
 106          continue
          endif
c
      elseif(xp1.gt.xremov)then
c
c only np1 is beyond xremov:
c if not already changed, change category for np1 to as.
c set changes(np1)=3.
c leave changes(np2)=2.
c
          update=1
          status(firstb(np2)+r2(lp))=0
          if(changes(np1).ne.3)then
              categor(np1)=0
              call chgcatb(listf,listb,np1,pl,pzl,pas,pzas)
              nent=nent-1
              changes(np1)=3
              i1 = firstb(np1)
              do 107 j=1,nring
                 status(i1+j)=0
 107          continue
          endif
c
      elseif(xp2.gt.xremov)then
c
c only np2 is beyond xremov:
c if not already changed, change category for np2 to as.
c set changes(np2)=3.
c leave changes(np1)=2.
c
          update=1
          status(firstb(np1)+r1(lp))=0
          if(changes(np2).ne.3)then
              categor(np2)=0
              call chgcatb(listf,listb,np2,pl,pzl,pas,pzas)
              nent=nent-1
              changes(np2)=3
              i1 = firstb(np2)
              do 108 j=1,nring
                 status(i1+j)=0
 108          continue
          endif
c
      elseif(sdist.gt.sbreak(lp))then
c
c length is greater than breaking length.
c break the link but do not remove the platelets
c
         status(j1+1)=0
         status(j2+1)=0
         update=1
      else
c
c neither is beyond xremov and tension is not great enough
c to break the link.
c leave this link alone.
c
         changes(np1)=1
         changes(np2)=1
      endif
c
c end of if-then for wall-plt or plt-plt links
      endif
c
      if(update.eq.1)then
c
c update the link data structure: 
c      remove link lp. 
c      update p1,p2, and listlk.
c      update numlink
c      reduce nlink by 1
c
         nwhere=((p2(lp)-2)*(p2(lp)-1))/2 + p1(lp)
         np1=p1(lp)
         np2=p2(lp)
         p1(lp)=0
         p2(lp)=0
         r1(lp)=0
         r2(lp)=0
         call chgcatb(listlkf,listlkb,lp,lkl,lkzl,lkas,lkzas)
         numlink(nwhere)=numlink(nwhere)-1
         nlink=nlink-1
         if(numlink(nwhere).eq.0)then
c
c update coherent-pair data structure to reflect loss of pair
c
c remove np2 from np1's list
c
            cpflag=1
            i=fstlnk(np1)
            if(i.eq.0)goto 120
 110        continue
            is=nxtlnkf(i)
            if(pltlnk(i).eq.np2)then
              call chgcatb(nxtlnkf,nxtlnkb,i,fstlnk(np1),
     +             lstlnk(np1),nas,nzas)
              pltlnk(i)=0
              goto 120
            endif
            i=is
            goto 110
 120        continue
c
c remove np1 from np2's list
c
            i=fstlnk(np2)
            if(i.eq.0)goto 140
 130        continue
            is=nxtlnkf(i)
            if(pltlnk(i).eq.np1)then
              call chgcatb(nxtlnkf,nxtlnkb,i,fstlnk(np2),
     +             lstlnk(np2),nas,nzas)
              pltlnk(i)=0
              goto 140
            endif
            i=is
            goto 130
 140        continue
         endif
      endif
      lp=lps
      goto 50
  210 continue
c
c if changes(ient)=2,then categor(ient) should be 
c changed to 2 (activated).
c
      do 250 npp=1,ment
          if(changes(npp).eq.2)then
c
c this is a platelet which is in the domain but which is linked 
c to no other platelet.
c
               categor(npp)=2
               call chgcate(listf,listb,npp,pl,pzl,pac,pzac)
           endif
  250 continue
c
c      write(16,*)' remove    smin=',smin
      return
      end
c
      subroutine cmass
c
c this subroutine computes the center of mass for each ring in
c the the domain
c
      implicit none
c
c  global commons
c
      include 'ibp.inc'
c
c  local variables
c
      integer i,i1,i2,ient
c
c compute centers of mass (cm) for all platelet rings present
c in the domain
c
      do 10 ient=1,ment
         xcm(ient) = 0.0D+00
         ycm(ient) = 0.0D+00
         if(categor(ient).ge.1)then
            i1 = firstb(ient)
            i2 = lastb (ient)
            do 5 i=i1,i2
               xcm(ient) = xcm(ient) + xb(2*i-1)
               ycm(ient) = ycm(ient) + xb(2*i  )
 5          continue
            xcm(ient) = xcm(ient)/dble(i2-i1+1)
            ycm(ient) = ycm(ient)/dble(i2-i1+1)
         endif
 10      continue
         return
         end
c
      subroutine create(nring)
c
c this subroutine inserts new platelets at the upstream end
c of the computational domain.
c at present, we wish to achieve what would be a uniform 
c distribution of y-coordinates in the domain if the
c flow field were to remain parabolic.  to do this we
c use an insertion scheme with a parabolic distribution
c in y.  the scheme is also derived under the assumption that 
c the 2d flux should equal the real 3d flux.
c
      implicit none
c
c  global common variables
c
      include 'ibp.inc'
      include 'links.inc'
      include 'domain.inc'
      include 'fluid.inc'
      include 'plat.inc'
c
      integer nring
c
c  local variables
c
      integer i,i1,j,jc,jent,jx,jy,k
      integer kcount,ktry
      double precision ang,b,bprim,delta,ran1
      double precision eplat,fiy
      double precision sdiff,width,x,xdiff
      double precision y,ydiff
      double precision yleast,ymost,yn,yo,zran
      double precision poissn(4),xnew(4),ynew(4)
c
      data xnew /0.0d0,0.0d0,0.0d0,0.0d0/
      data ynew /0.0d0,0.0d0,0.0d0,0.0d0/
c
c platelet creation parameters
c
      eplat=fluxx*pdens*dlt
c
      if(eplat.lt.0.0D+00)return
c
      poissn(1)=dexp(-eplat)
      poissn(2)=poissn(1)*(1.D+00+ eplat)
      poissn(3)=poissn(1)*(1.D+00+ eplat + eplat*eplat/2.D+00)
      poissn(4)=poissn(3)+poissn(1)*(eplat*eplat*eplat)/6.D+00
c
      width=ytop-ybot
      ang=8.0D+00*datan(1.0D+00)/dble(nring)
      yleast = ybot + .70D+00*diam
      ymost  = ytop - .70D+00*diam
c 
      delta=1.0d-4
      kcount=0
c
      zran = ran1()
      if(zran.le.poissn(1))then
         k=0
      elseif(zran.le.poissn(2))then
         k=1
      elseif(zran.le.poissn(3))then
         k=2
      elseif(zran.le.poissn(4))then
         k=3
      else
         k=4
      endif
      write(6,*)' create k=',k
c
c see if one or more new platelets should be created.
c
      if(k.ge.1)then
c
      do 100 j=1,k
c
c   create a platelet
c
c   determine the y-coordinate of its center of mass.
c
      nent=nent+1
      if(nent.gt.ment)then
           write(16,1)nent
           stop 1
      endif
c
c use newton's method to determine y-coordinate of new platelet.
c
      ktry = 0
 3    continue
      ktry = ktry + 1
      if(ktry.gt.10)then
         write(6,*)' create:  could not fit new one'
         nent=nent-1
         goto 100
      endif
      kcount=0
      fiy=ran1()
           write(6,*)' fiy=',fiy
      yo=.5D+00*(ytop+ybot)
 5    continue
      zran=(yo-ybot)/width
      b=(zran**2)*(3.D+00-2.D+00*zran)-fiy
      bprim=6.D+00*zran*(1.D+00-zran)/width
      yn=yo - b/bprim
      if(dabs(yn-yo).lt.delta) go to 7
      yo=yn
      kcount=kcount+1
      if(kcount.gt.20)then
           write(6,*)' Newton failure kcount=',kcount
           go to 3
      endif
      go to 5
 7    continue
c
c newton's method converged
c
      y=yn
c
c see if platelet ring will be inside tube
c
      if(y.lt.yleast.or.y.gt.ymost)then
            write(6,*)' yleast/ymost   y=',y
            go to 3
      endif
c
c determine x-coordinate of platelet's center of mass
c
      x= .75D+00*diam + xmin
c
c see if platelet overlaps with an existing platelet
c
      do 50 jent=1,ment
         if(categor(jent).ge.1)then
            xdiff=xcm(jent)-x
            ydiff=ycm(jent)-y
            sdiff=dsqrt(xdiff**2 + ydiff**2)
            if(sdiff.le.(1.25D+00*diam))then
               write(6,*)' overlap    sdiff=',sdiff
               goto 3
            endif
         endif
 50   continue
      if(k.ge.2)then
      do 55 i=1,j-1
         xdiff=xnew(i)-x
         ydiff=ynew(i)-y
         sdiff=dsqrt(xdiff**2 + ydiff**2)
         if(sdiff.le.(1.25D+00*diam))then
            write(6,*)' overlap    sdiff=',sdiff
            goto 3
         endif
 55   continue
      endif
      xnew(j)=x
      ynew(j)=y
c
c assign to the new platelet the number of the first as, pas, 
c and put this platelet number at the end of unactivated category ua.
c
      i=pas
      write(6,*)' create     pas=',pas
c
c create unactivated platelets
c
c      call chgcate(listf,listb,i,pas,pzas,pua,pzua)
c      categor(i)=1
c
c create activated platelets
c
      call chgcate(listf,listb,i,pas,pzas,pac,pzac)
      categor(i)=2
c
c assign the locations of the points on the platelet ring
c
      write(6,*)' CREATE   x=',x,'      y=',y
      write(16,*)' CREATE   x=',x,'      y=',y
      i1=firstb(i)
      do 30 jc=1,nring
         jy = 2*(i1+jc-1)
         jx = jy - 1
         xb(jx) = x + .5D+00*diam*dcos(ang*dble(jc-1))
         xb(jy) = y + .5D+00*diam*dsin(ang*dble(jc-1))
         status(i1+jc)=0
 30   continue
c
 100  continue
      endif
      write(6,*)' create nent=',nent
c
    1 format(//,5x,'create   nent too big  nent=',i5)
      return
      end
c
c
      subroutine height(last1,last2)
c
c determine heights of wall bound aggregates
c
      implicit none
c
c  global commons
c
      include 'domain.inc'
      include 'ibp.inc'
      include 'plat.inc'
c
c  global variables
c
      integer last1,last2
c
c  local variables
c
      integer j
      double precision hbot,htop
c
      write(16,*)' ******* height'
c
c bottom wall
c
      hbot = -1000.d0
      if(last1.gt.1)then
         do 10 j=2,last1
            hbot=dmax1(hbot,ycm(wallbnd(j))-ybot)
 10      continue
         hbot = hbot + .5d0*diam
         write(16,*)' hbot=',hbot
      endif
c
c top wall
c
      htop = -1000.d0
      if(last2.gt.(last1+1))then
         do 20 j=last1+2,last2
            htop=dmax1(htop,ytop-ycm(wallbnd(j)))
 20      continue
         htop = htop + .5d0*diam
         write(16,*)' htop=',htop
      endif        

      return
      end
c
      function ran1()
c
c from Numerical Recipes (Press, Flannery, Teukoslsky, and Vetterling)
c page 196.
c
c returns a uniform random variable between 0.0 and 1.0
c
c set idum to any NEGATIVE value to initialize or re-initialize
c the sequence
c
      integer ia1,ia2,ia3,ic1,ic2,ic3,idum,iff,ix1,ix2,ix3
      integer j,m1,m2,m3
c
      double precision rp1,ran1,rm1,rm2
c
      common/fffff/iff,idum
      common/random/ rp1(97),ix1,ix2,ix3

      parameter(m1=259200,ia1=7141,ic1=54773,rm1=1.D+00/m1)
      parameter(m2=134456,ia2=8121,ic2=28411,rm2=1.D+00/m2)
      parameter(m3=243000,ia3=4561,ic3=51349)

c     data iff/0/
      save
c
      if(idum.lt.(0).or.iff.eq.(0))then
        iff=1
        ix1=mod(ic1-idum,m1)
        ix1=mod(ia1*ix1+ic1,m1)
        ix2=mod(ix1,m2)
        ix1=mod(ia1*ix1+ic1,m1)
        ix3=mod(ix1,m3)
c
        do 11 j=1,97
           ix1=mod(ia1*ix1+ic1,m1)
           ix2=mod(ia2*ix2+ic2,m2)
           rp1(j) = (dble(ix1)+dble(ix2)*rm2)*rm1
 11     continue
        idum = 1
      endif
      ix1=mod(ia1*ix1+ic1,m1)
      ix2=mod(ia2*ix2+ic2,m2)
      ix3=mod(ia3*ix3+ic3,m3)
      j = 1 + (97*ix3)/m3
      if(j.gt.(97).or.j.lt.(1))then
           write(6,*)' ran1    oops!   j=',j
           stop
      endif
      ran1 = rp1(j)
      rp1(j) = (dble(ix1)+dble(ix2)*rm2)*rm1
      return
      end

      double precision function ran1p()
c
c from Numerical Recipes (Press, Flannery, Teukoslsky, and Vetterling)
c page 196.
c
c returns a uniform random variable between 0.0 and 1.0
c
c set idum to any NEGATIVE value to initialize or re-initialize
c the sequence
c
      integer ia1,ia2,ia3,ic1,ic2,ic3,idump,iffp
      integer ix1p,ix2p,ix3p,j,m1,m2,m3
c
      double precision rm1,rm2,rp
c
      common/fffffp/iffp,idump
      common/randomp/ rp(97),ix1p,ix2p,ix3p
      parameter(m1=259200,ia1=7141,ic1=54773,rm1=1.D+00/m1)
      parameter(m2=134456,ia2=8121,ic2=28411,rm2=1.D+00/m2)
      parameter(m3=243000,ia3=4561,ic3=51349)
c     data iffp/0/
      save
c
      if(idump.lt.(0).or.iffp.eq.(0))then
        iffp=1
        ix1p=mod(ic1-idump,m1)
        ix1p=mod(ia1*ix1p+ic1,m1)
        ix2p=mod(ix1p,m2)
        ix1p=mod(ia1*ix1p+ic1,m1)
        ix3p=mod(ix1p,m3)
c
        do 11 j=1,97
           ix1p=mod(ia1*ix1p+ic1,m1)
           ix2p=mod(ia2*ix2p+ic2,m2)
           rp(j) = (dble(ix1p)+dble(ix2p)*rm2)*rm1
 11     continue
        idump = 1
      endif
      ix1p=mod(ia1*ix1p+ic1,m1)
      ix2p=mod(ia2*ix2p+ic2,m2)
      ix3p=mod(ia3*ix3p+ic3,m3)
      j = 1 + (97*ix3p)/m3
      if(j.gt.(97).or.j.lt.(1))then
           write(6,*)' ran1p    oops!   j=',j
           stop
      endif
      ran1p = rp(j)
      rp(j) = (dble(ix1p)+dble(ix2p)*rm2)*rm1
      return
      end
c
c NOTE:  put call cmass before call rw.
c
      subroutine rw
c
c platelet random walk
c
      implicit none
c
c  global commons
c
      include 'ibp.inc'
      include 'domain.inc'
      include 'plat.inc'
c
c  local variables
c
      integer i,i1,i2,ic,ient,j,j1,j2,jc,jent,ni,nj
      double precision aden,ang,anum
      double precision tpi,r1,r2
      double precision rad,radmax,ran1p,rg,wind
      double precision scmnom,s2vec1,s2vec2,scm
      double precision var,vec1x,vec1y,vec2x,vec2y
      double precision x,xi,xj,xjo
      double precision y,yi,yj,yjo
c
      tpi=8.D+00*datan(1.D+00)
      var=2.D+00*pltdif*dlt
c
      do 10 jent=1,ment
      if((categor(jent)-1)*(categor(jent)-2).eq.0)then
c
c calculate radius of platelet
c
         radmax=0.0D+00
         i1 = firstb(jent)
         i2 = lastb (jent)
         do 5 i=i1,i2
            rad=dsqrt((xcm(jent)-xb(2*i-1))**2 +
     +                (ycm(jent)-xb(2*i  ))**2)
            radmax=max(radmax,rad)
 5       continue
         radius(jent)=radmax
      endif
 10   continue

c
      do 20 jent=1,ment
      if((categor(jent)-1)*(categor(jent)-2).eq.0)then
c
c calculate the random step from a truncated symmetric
c gaussian distribution with standard deviation sqrt(var)
c
         r1=max(ran1p(),1.d-8)
         r2=ran1p()
         rg=dsqrt(-2.D+00*var*dlog(r1))
         ranx(jent) = rg*dcos(tpi*r2)
         rany(jent) = rg*dsin(tpi*r2)
         xcm (jent) = xcm(jent) + ranx(jent)
         ycm (jent) = ycm(jent) + rany(jent)
         i1 = firstb(jent)
         i2 = lastb (jent)
         do 15 i=i1,i2
            xb(2*i-1)=xb(2*i-1) + ranx(jent)
            xb(2*i  )=xb(2*i  ) + rany(jent)
 15      continue
      endif
 20   continue
c
      do 200 ient=1,ment
      if((categor(ient)-1)*(categor(ient)-2).eq.0)then
         do 190 jent=ient+1,ment
         if((categor(jent)-1)*(categor(jent)-2).eq.0)then
c
c compute separation between the centers of mass
c
            scm = dsqrt((xcm(ient)-xcm(jent))**2
     +                + (ycm(ient)-ycm(jent))**2)
            scmnom=radius(ient)+radius(jent)
            if(scm.lt.scmnom)then
c
c separation between centers of mass is close enough that overlap may
c have occured -- test using winding numbers:  if |wind| = 1, then
c rings overlap.
c
                i1=firstb(ient)
                i2=lastb (ient)
                ni=i2-i1+1
                j1=firstb(jent)
                j2=lastb (jent)
                nj=j2-j1+1
c
 30             continue
                do 70 ic=1,ni
                   i=i1 + ic - 1
                   xi=xb(2*i-1)
                   yi=xb(2*i  )
                   wind=0.0D+00
                   jc=1
                   j= j1 + jc - 1
                   xjo= xb(2*j-1)
                   yjo= xb(2*j  )
                   do 60 jc=2,nj
                      j=j1 + jc - 1
                      xj=xb(2*j-1)
                      yj=xb(2*j  )
                      vec1x=xjo-xi
                      vec1y=yjo-yi
                      vec2x=xj -xi
                      vec2y=yj -yi
                      s2vec1=vec1x**2 + vec1y**2
                      s2vec2=vec2x**2 + vec2y**2
                      anum = vec1x*vec2y - vec1y*vec2x
                      aden = dsqrt(s2vec1*s2vec2)
                      ang = dasin(anum/aden)
                      wind = wind + ang/tpi
                      xjo=xj
                      yjo=yj
 60                continue
                   jc=1
                   j= j1 + jc - 1
                   xj= xb(2*j-1)
                   yj= xb(2*j  )
                   vec1x=xjo-xi
                   vec1y=yjo-yi
                   vec2x=xj -xi
                   vec2y=yj -yi
                   s2vec1=vec1x**2 + vec1y**2
                   s2vec2=vec2x**2 + vec2y**2
                   anum = vec1x*vec2y - vec1y*vec2x
                   aden = dsqrt(s2vec1*s2vec2)
                   ang = dasin(anum/aden)
                   wind = wind + ang/tpi
c
                   if(dabs(wind).gt.0.5D+00)then
                      write(16,1001)jent,ient
c
c overlap detected, backstep on random steps for rings ient and jent
c
                      i1 = firstb(jent)
                      i2 = lastb (jent)
                      xcm(jent)=xcm(jent)  - .25D+00*ranx(jent)
                      ycm(jent)=ycm(jent)  - .25D+00*rany(jent)
                      do 45 i=i1,i2
                         xb(2*i-1)=xb(2*i-1) - .25D+00*ranx(jent)
                         xb(2*i  )=xb(2*i  ) - .25D+00*rany(jent)
 45                   continue
                      i1 = firstb(ient)
                      i2 = lastb (ient)
                      xcm(ient)=xcm(ient)  - .25D+00*ranx(ient)
                      ycm(ient)=ycm(ient)  - .25D+00*rany(ient)
                      do 50 i=i1,i2
                         xb(2*i-1)=xb(2*i-1) - .25D+00*ranx(ient)
                         xb(2*i  )=xb(2*i  ) - .25D+00*rany(ient)
 50                   continue
                      goto 30
                   endif
 70             continue
            endif
c
c no overlap detected between jent and ient
c
         endif
 190     continue
      endif
 200  continue
c
c check for overlap with walls
c
      do 500 jent=1,ment
      if((categor(jent)-1)*(categor(jent)-2).eq.0)then
         x=xcm(jent)
         y=ycm(jent)
c
c assume flat walls (modify for stenotic or bifurcating walls
c
         if    (y-radius(jent).le.ybot)then
            ient=1
c
c find points on wall close to platelet
c
            i1=firstb(ient)
            i2=lastb (ient)
            ni=i2-i1+1
            do 310 ic=1,ni
               i=i1 + ic - 1
               xi=xb(2*i-1)
               if(xi.gt.x)then
                  i1=max0(i1,i-10)
                  i2=min0(i2,i+10)
                  goto 311
               endif
 310        continue
 311        continue
            ni=i2-i1+1
            j1=firstb(jent)
            j2=lastb (jent)
            nj=j2-j1+1
            do 350 ic=1,ni
                i=i1 + ic - 1
                xi=xb(2*i-1)
                yi=xb(2*i  )
 315            continue
                wind=0.0D+00
                jc=1
                j= j1 + jc - 1
                xjo= xb(2*j-1)
                yjo= xb(2*j  )
                do 330 jc=2,nj
                   j=j1 + jc - 1
                   xj=xb(2*j-1)
                   yj=xb(2*j  )
                   vec1x=xjo-xi
                   vec1y=yjo-yi
                   vec2x=xj -xi
                   vec2y=yj -yi
                   s2vec1=vec1x**2 + vec1y**2
                   s2vec2=vec2x**2 + vec2y**2
                   anum = vec1x*vec2y - vec1y*vec2x
                   aden = dsqrt(s2vec1*s2vec2)
                   ang = dasin(anum/aden)
                   wind = wind + ang/tpi
                   xjo=xj
                   yjo=yj
 330            continue
                jc=1
                j= j1 + jc - 1
                xj= xb(2*j-1)
                yj= xb(2*j  )
                vec1x=xjo-xi
                vec1y=yjo-yi
                vec2x=xj -xi
                vec2y=yj -yi
                s2vec1=vec1x**2 + vec1y**2
                s2vec2=vec2x**2 + vec2y**2
                anum = vec1x*vec2y - vec1y*vec2x
                aden = dsqrt(s2vec1*s2vec2)
                ang = dasin(anum/aden)
                wind = wind + ang/tpi
c
                if(dabs(wind).gt.0.5D+00)then
c
c overlap detected, backstep on random step for ring jent
c
                      i1 = firstb(jent)
                      i2 = lastb (jent)
                      xcm(jent)=xcm(jent)  - .25D+00*ranx(jent)
                      ycm(jent)=ycm(jent)  - .25D+00*rany(jent)
                      do 345 i=i1,i2
                         xb(2*i-1)=xb(2*i-1) - .25D+00*ranx(jent)
                         xb(2*i  )=xb(2*i  ) - .25D+00*rany(jent)
 345                  continue
                      goto 315
                endif
 350        continue
c
         elseif(y+radius(jent).ge.ytop)then
            ient=2
c
c find points on wall close to platelet
c
            i1=firstb(ient)
            i2=lastb (ient)
            ni=i2-i1+1
            do 410 ic=1,ni
               i=i1 + ic - 1
               xi=xb(2*i-1)
               if(xi.gt.x)then
                  i1=max0(i1,i-10)
                  i2=min0(i2,i+10)
                  goto 411
               endif
 410        continue
 411        continue
            ni=i2-i1+1
            j1=firstb(jent)
            j2=lastb (jent)
            nj=j2-j1+1
            do 450 ic=1,ni
                i=i1 + ic - 1
                xi=xb(2*i-1)
                yi=xb(2*i  )
 415            continue
                wind=0.0D+00
                jc=1
                j= j1 + jc - 1
                xjo= xb(2*j-1)
                yjo= xb(2*j  )
                do 430 jc=2,nj
                   j=j1 + jc - 1
                   xj=xb(2*j-1)
                   yj=xb(2*j  )
                   vec1x=xjo-xi
                   vec1y=yjo-yi
                   vec2x=xj -xi
                   vec2y=yj -yi
                   s2vec1=vec1x**2 + vec1y**2
                   s2vec2=vec2x**2 + vec2y**2
                   anum = vec1x*vec2y - vec1y*vec2x
                   aden = dsqrt(s2vec1*s2vec2)
                   ang = dasin(anum/aden)
                   wind = wind + ang/tpi
                   xjo=xj
                   yjo=yj
 430            continue
                jc=1
                j= j1 + jc - 1
                xj= xb(2*j-1)
                yj= xb(2*j  )
                vec1x=xjo-xi
                vec1y=yjo-yi
                vec2x=xj -xi
                vec2y=yj -yi
                s2vec1=vec1x**2 + vec1y**2
                s2vec2=vec2x**2 + vec2y**2
                anum = vec1x*vec2y - vec1y*vec2x
                aden = dsqrt(s2vec1*s2vec2)
                ang = dasin(anum/aden)
                wind = wind + ang/tpi
c
                if(dabs(wind).gt.0.5D+00)then
c
c overlap detected, backstep on random step for ring jent
c
                      i1 = firstb(jent)
                      i2 = lastb (jent)
                      xcm(jent)=xcm(jent)  - .25D+00*ranx(jent)
                      ycm(jent)=ycm(jent)  - .25D+00*rany(jent)
                      do 445 i=i1,i2
                         xb(2*i-1)=xb(2*i-1) - .25D+00*ranx(jent)
                         xb(2*i  )=xb(2*i  ) - .25D+00*rany(jent)
 445                  continue
                      goto 415
                endif
 450        continue
c
         endif
      endif
 500  continue
 1001 format(' rw:  overlap   jent=',i4,' ient=',i4)
      return
      end
c
      subroutine wallagg(last1,last2)
c
c determine list of wall-bound platelets
c
      implicit none
c
c  global commons
c
      include 'ibp.inc'
      include 'links.inc'
      include 'plat.inc'
c
c  global variables
c
      integer last1,last2
c
c  local variables
c
      integer i,ient,is,j,jent,numbnd1,numbnd2
      integer pt1,pt2
c
c initialize
c
      do 5 j=1,ment
         visited(j)=0
         wallbnd(j)=0
 5    continue
c
c first bottom wall ient=1
c
      ient=1
      visited(ient)=1
      wallbnd(1   )=ient
      pt1=1
      pt2=2
 15   continue
      i=fstlnk(ient)
 20   continue
      if(i.eq.0)goto 30
      is=nxtlnkf(i)
      jent=pltlnk(i)
      if(visited(jent).eq.0)then
c
c check whether jent=2 (top wall)
c if so, do not add to list wallbnd
c
         if(jent.ne.2)then
            visited(jent)=1
            wallbnd(pt2 )=jent
            pt2=pt2+1
         endif
      endif
      i=is
      goto 20
 30   continue
      pt1=pt1+1
      if(pt1.eq.pt2)goto 40
      ient=wallbnd(pt1)
      goto 15
 40   continue
c
c numbnd = pt1 - (1 since pt1 is past end of list) 
c              - (1 to account for wall itself)
c
      last1  =pt1-1
      numbnd1=pt1-2
c
c      do 45 j=1,ment
c         write(16,46)j,wallbnd(j),visited(j)
c 45   continue
c 46   format(' j=',i3,' wb=',i5,' v=',i5)
      write(16,*)' numbnd1=',numbnd1
c
c now top wall ient=2
c
      ient=2
      visited(ient)=1
      wallbnd(pt1) =ient
      pt2=pt2+1
 115  continue
      i=fstlnk(ient)
 120  continue
      if(i.eq.0)goto 130
      is=nxtlnkf(i)
      jent=pltlnk(i)
      if(visited(jent).eq.0)then
         visited(jent)=1
         wallbnd(pt2 )=jent
         pt2=pt2+1
      endif
      i=is
      goto 120
 130  continue
      pt1=pt1+1
      if(pt1.eq.pt2)goto 140
      ient=wallbnd(pt1)
      goto 115
 140  continue
c
c numbnd = pt1 - (1 since pt1 is past end of list) 
c              - (1 to account for wall itself)
c
      last2  =pt1-1
      numbnd2=pt1-2-last1
      write(16,*)' numbnd2=',numbnd2
c
c can also add more code to determine platelets
c in each distinct free-floating aggregate
c
      return
      end
