c
      subroutine ibisdrv(nxp2,nyp1,u,f,fout,p,psi,bzero,
     *    sinsq,z,b,c,r,work1,work2,yh,yp,
     *    ments,mlink,mlkpt,mpts,mwhere,
     *    nring,np,npdx,npdy,nsrc,
     *    alpha,nxx,nyy)
c
c this program is the driver routine for IBIS simulations
c get global parameters
c
      implicit none
c
c get local parameter and global commons
c
      include 'ibp.inc'
      include 'active.inc'
      include 'links.inc'
      include 'domain.inc'
      include 'graph.inc'
      include 'fluid.inc'
      include 'plat.inc'
      include 'sink.inc'
c
c  global variables 
c
      integer cpflag,last1,last2,nring
      integer mpts,np,nsrc,ntsrc,nxp2,nxx,nyp1,nyy
      integer ments,mlink,mlkpt,mwhere,npdx,npdy
c
      byte alpha(nxx,nyy)
c
      double precision sfluid(3), a
c
      integer jx,jy
      common/sprdmap/ jx(16),jy(16)
c
      double precision u(0:nxp2,0:nyp1,1:2)
      double precision f(0:nxp2,0:nyp1,1:2)
      double precision p(0:nxp2,0:nyp1)
      double precision psi(0:nxp2,0:nyp1)
      double precision fout(0:nxp2,0:nyp1,1:2)
c
      double precision bzero(0:nxp2,0:nyp1)
      double precision sinsq(0:nxp2,0:nyp1)
      double precision z(0:nxp2,0:nyp1)
c
      double precision b(0:nxp2,0:nyp1)
      double precision c(0:nxp2,0:nyp1)
      double precision r(0:nxp2,0:nyp1)
      double precision work1(0:nxp2,0:nyp1)
      double precision work2(0:nxp2,0:nyp1)
      double precision yh(0:nxp2,0:nyp1)
      double precision yp(0:nxp2,0:nyp1)
c
c  command line variables
c
      character*1 cmdline
      integer*4 iargc,ncmdline
c
c  local variables
c
      integer i,j,k,idebug(0:4),rsflag,nx,ny
c
      data idebug / 0,0,0,0,0 /
c
      data iff,iffp,iplat /0,0,0/
c
      write(6,*)' '
      write(6,*)'IBIS numerical solver'
      write(6,*)'Version 1.0'
      write(6,*)' '
c
      nx = nxp2-2
      ny = nyp1-1
      ment = ments
c
c initialize immersed boundary points and fluid velocities
c
      call init(nxp2,nyp1,f,u,sinsq,bzero,sfluid,a,z,rsflag,
     x    mlink,mlkpt,mpts,mwhere,np,npdx,npdy,nring,nxx,nyy)
c
      time = ntstart*dlt
      nt = 0
c
c initialize array for spread/interp
c
      call sprdinit
c
c plot the initial data
c
      if (rsflag .eq. 0) then
        call cplot(ccount,alpha,nxx,nyy,mlink,np)
        call output(lcount,nxp2,nyp1,f,u,mlink,np)
      endif
c
c check command line for stop or debug options
c
      ncmdline = iargc()
      if (ncmdline .eq. 1) then
        call getarg(1,cmdline)
        if ( cmdline .eq. 'p' ) then
          stop
        elseif ( cmdline .eq. 'd' ) then
          idebug(0) = 1
          open (unit=12,file='ibisdat/debug.dat',status='old')
          do j=1,2
            read (12,*) idebug(j)
          enddo
          close(12)
        endif
      endif
c
c print some junk out on stdout
c
      write(6,*)' nent = ',nent, ' npts = ',npts
      call prout (ntstart,fluxx,fluxy,vmax,(dlt*vmax)-h,time)
c
cccccccccccccccccccccccc main loop ccccccccccccccccccccc
c
      do nt = ntstart+1, ntmax
c
c
        time = time + dlt
c
c compute centers of mass for use in create
c  and insert new rings if necessary
c
        if (iplat .eq. 1) then
          call cmass
          if(fluxx.gt.(0.0D+00))then
            call create(nring)
          endif
        endif
c
c active control of the immersed boundary points
c   pass stuff to user.
c
        do i=1,ment
          if ( activ1(i) .eq. 1 ) then
            call preact(i)
            call actmov
            call afact(i)
          endif
        enddo
c
c perform map of actual entities
c evaluate the force density at immersed boundary points
c
        call map
        call grad
        call gradlink
        if (idebug(0) .eq. 1 ) 
     x     call pdebug (idebug)
c
c calculate entity force density (force/length)
cdir$ ivdep
c
        do j=1,npts
          fs(2*j-1)=-gfrc1(j)
          fs(2*j)=-gfrc2(j)
        enddo
c
c spread this to grid to get fluid force density
c add background force to drive the flow
c
        call spread(npts,nxp2,nyp1,f)
        call fadd(nxp2,nyp1,f)
c
c save force density for output if appropriate
c extend fout periodically to 0:nx 0:nyp1
c
        if(mod(nt,lfreq).eq.0)then
          do j=1,ny
            do i=1,nx
              fout(i,j,1)=f(i,j,1)
              fout(i,j,2)=f(i,j,2)
            enddo
          enddo
          call period (nxp2,nyp1,fout(0,0,1))
          call period (nxp2,nyp1,fout(0,0,2))
        endif
c
c get the fluid sources and sinks, check for volume,
c   number of sources and sinks, and fix particles.
c
        call getsrc (ntsrc)
        if ( ntsrc.gt.0 ) then
          call chksrc(ntsrc,nsrc)
          call sprsrc(ntsrc,nxp2,nyp1,psi)
          call prtsrc(ntsrc,np)
        endif
c
c solve ns equations
c extend u periodically to 0:nx+1  0:nyp1
c compute maximum speed and fluxes
c
        call navs(nxp2,nyp1,u,f,p,psi,
     *        bzero,sinsq,z,
     *        b,c,r,work1,work2,yh,yp,
     *        a,dlt,h,ro,sfluid)
        call period (nxp2,nyp1,u(0,0,1))
        call period (nxp2,nyp1,u(0,0,2))
        call upflux (nxp2,nyp1,u)
c
c move entities by convection - add centers of the
c  entities to the point list and add the fluid marker
c  particles to the point list.  Then remove after moving.
c
        call addmap(xpt,ypt,mpts,centx,centy,nent,npts)
        call addmap(xpt,ypt,mpts,px,py,np,npts+nent)
        call xmove (npts+nent+np,nxp2,nyp1,u)
        call delmap(xpt,ypt,mpts,px,py,np,npts+nent)
        call delmap(xpt,ypt,mpts,centx,centy,nent,npts)
c
c move fluid marker particles via random walk
c evaluate the fluid marker positions
c
        if ( pwalk .gt. 0.0d0 ) call partwalk(np)
        call part(np,npdx,npdy,nxx,nyy)
c
c reverse map of coordinates
c
        call revmap
c
c check on the platelet progress
c
        if (iplat .eq. 1) then
          call chkplat(cpflag,last1,last2,mlink,nring)
        endif
c
c remove the forces on objects have left the domain
c  ADD SOMETHING HERE

c
c print data to screen...
c
        call prout (nt,fluxx,fluxy,vmax,(dlt*vmax)-h,time)
c
c  shut down if failure is about to happen
c
        if( abs((dlt*vmax)-h) .ge. 1.0d0 ) then
          call ohnoo
          call cplot(ccount+1,alpha,nxx,nyy,mlink,np)
          call output(lcount+1,nxp2,nyp1,fout,u,mlink,np)
          stop
        endif
c
c  output the color plot at correct times
c
        if (mod(nt,cfreq).eq.0) then
          ccount = ccount + 1
          call cplot(ccount,alpha,nxx,nyy,mlink,np)
        endif
c
c  output line drawing and restart files
c
        if(mod(nt,lfreq).eq.0)then
          lcount = lcount+1
          call output(lcount,nxp2,nyp1,fout,u,mlink,np)
          call rsdata(13,nxp2,nyp1,u,mlink,np)
        endif
c
c  debug as needed - the immersed boundary points, forces
c
c
cccccccccccccccccccccccc end main loop ccccccccccccccccc
c
      enddo
      return
 1008 format(7(1x,e15.7))
 9990 format(8f15.7)
 9991 format(4f10.3,6e11.3)
      end
c
c
      subroutine chkplat(cpflag,last1,last2,mlink,nring)
c
c  this routine gathers a cluster of subroutine calls for
c    platlets only
c
c get global parameters
c
      implicit none
c
c global commons
c
      include 'plat.inc'
      include 'links.inc'
c
c global variables
c
      integer cpflag,last1,last2,mlink,nring
c
c compute center of mass for each platelet in domain
c   and move single rings by random walk
c
      call cmass
      call rw
c
c make new links between rings and walls
c
      cpflag=0
      call adhere(cpflag)
c
      if(nlink.gt.mlink)then
          write(6,*)' NLINK GT MLINK'
          stop
      endif
c
c make new links between rings
c
      call cohere(cpflag)
c
      if(nlink.gt.mlink)then
          write(6,*)' NLINK GT MLINK'
          stop
      endif
c
c remove rings downstream (and break links)
c
      call remove(cpflag,nring)
c
c if linked pairs have changed, compute new
c list and number of wall-bound platelets
c
      if(cpflag.eq.1)then
        call wallagg(last1,last2)
      endif
c
c compute height of wall-bound aggregates
c
      call height(last1,last2)
c
      return
      end

