/*
/   Program ibpts
/
/     Reads the little language to create the initial
/     conditions for Fogelson's immersed boundary method
/     fluid dynamics code, and writes the input file
/     to IBM code.
/
*/
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define MAXLINE 100  /* maximum number of lines in input file */
#define MAXARG  100    /* maximum number of arguments on input line */
#define MAXCHAR 2000  /* maximum number of characters on input line */

#define max(A,B) ( (A) > (B) ? (A) : (B))
#define min(A,B) ( (A) > (B) ? (B) : (A))

char colorf[MAXLINE];
char linef[MAXLINE];
char fluidf[MAXLINE];
char partf[MAXLINE];
int IBact[MAXLINE][4];
int IBint[MAXLINE][MAXARG];
float IBdata[MAXLINE][MAXARG];
float defal[MAXLINE];
float work[2][MAXLINE];

int unused_names;
/*
/   Global variables
/
/
/  INTEGER DATA
/
/     IBint[0][k]  -  graphics information  - not passed to FORTRAN
/        IBint[0][0]  -  fluid marker density (% of ny)
/        IBint[0][1]  -  fluid marker density (% of nx)
/        IBint[0][2]  -  width of visualization output graphics
/        IBint[0][3]  -  height of visualization output graphics
/        IBint[0][4]  -  zoom factor
/        IBint[0][5]  -  increment on movies
/        IBint[0][6]  -  first file
/        IBint[0][7]  -  last file
/     IBint[0][k]  -  auxillary information
/        IBint[0][8]  -  filament counter
/        IBint[0][9]  -  circle counter
/        IBint[0][10] -  polygon counter
/        IBint[0][11] -  unused names
/        IBint[0][12] -  link counter
/     IBint[1][k]  -  domain information
/        IBint[1][0]  -  number of entities
/        IBint[1][2]  -  number of immersed boundary points
/        IBint[1][3]  -  number of executable commands in input code
/        IBint[1][4]  -  frequency of visualization output graphics
/        IBint[1][5]  -  line graphics output frequency
/        IBint[1][6]  -  vector output density
/        IBint[1][7]  -  2^nx
/        IBint[1][8]  -  2^ny
/        IBint[1][9]  -  number of linked entities
/     IBint[2][k]  -  grid information
/        IBint[2][1]  -  nx
/        IBint[2][2]  -  ny
/        IBint[2][3]  -  ntmax
/        IBint[2][4]  -  ntstart
/        IBint[2][5]  -  force save flag
/        IBint[2][6]  -  velocity save flag
/     IBint[3][k]  -  tube/particle information
/        IBint[3][1]  -  (=3) if a tube exists, (=0) if no tube
/        IBint[3][2]  -  number of points in the tube.
/        IBint[3][3]  -  (=0) if periodic particles in x (=1) if nuemann
/        IBint[3][4]  -  (=0) if periodic particles in y (=1) if nuemann
/        IBint[3][5]  -  number of sources/sinks
/     IBint[4][k]  -  platelet information
/        IBint[4][1]  -  maximum platelet links
/        IBint[4][2]  -  platelet indicator
/        IBint[4][3]  -  number of points on a ring
/
/     IBint[j][k]  -  entity information
/        IBint[j][0]  -  entity name (integer valued)
/        IBint[j][1]  -  entity type
/         (=2) open lines
/         (=3) tube
/         (=4) circle
/         (=5) singlets
/         (=6) initialize polygon
/           (=7) ray
/           (=8) arc
/           (=9) fixed arc
/           (=10) circle segment
/           (=11) singlet
/         (=14) ellipse
/         (=22) periodic lines
/         (=32) sinusoidal curve
/        IBint[j][2]  -  number of ibp points in entity
/        IBint[j][3]  -  (=0) if entity not closed, (=1) if closed
/        IBint[j][4]  -  number of segments in polygon
/        IBint[j][5]  -  activation status
/        IBint[j][6]  -  used for bending stiffness of endpoints
/        IBint[j][7]  -  used for bending stiffness of endpoints
/        IBint[j][8]  -  used for connecting angle calculations
/        IBint[j][15] -  used for endpoint tethering
/
/
/     IBact[1][1]  -  number of active entities
      IBact[k][1]  -  active entity flag
/         (=0) not active
/         (=1) active 
/
/
/  FLOATING POINT DATA
/
/     IBdata[1][k]  -  domain information
/        IBdata[1][1]  -  xmin
/        IBdata[1][2]  -  xmax
/        IBdata[1][3]  -  ymin
/        IBdata[1][4]  -  ymax
/        IBdata[1][5]  -  mu
/        IBdata[1][6]  -  rho
/        IBdata[1][7]  -  bottom tube injury x1
/        IBdata[1][8]  -  bottom tube injury x2
/        IBdata[1][9]  -  top tube injury x1
/        IBdata[1][10] -  top tube injury x2
/        IBdata[1][11] -  characteristic length
/        IBdata[1][12] -  characteristic time
/     IBdata[2][k]  -  grid information
/        IBdata[2][1]  -  nx
/        IBdata[2][2]  -  ny
/        IBdata[2][3]  -  delta t
/        IBdata[2][4]  -  h_x
/        IBdata[2][5]  -  h_y
/        IBdata[2][6]  -  min(h_x,h_y)
/        IBdata[2][7]  -  max initial flow rate
/        IBdata[2][8]  -  background x pressure grad
/        IBdata[2][9]  -  background y pressure grad
/        IBdata[2][10] -  platelet density
/        IBdata[2][11] -  max distance of particle random walk
/     IBdata[3][k]  -  tube information
/        IBdata[3][1]  -  tube xmin
/        IBdata[3][2]  -  tube ymin
/        IBdata[3][3]  -  tube xmax
/        IBdata[3][4]  -  tube ymax
/        IBdata[3][5]  -  platelet diffusion coef
/        IBdata[3][6]  -  link stiffness
/        IBdata[3][7]  -  bend stiffness
/        IBdata[3][8]  -  tether stiffness
/        IBdata[3][9]  -  rest length
/        IBdata[3][10] -  rest curvature
/     IBdata[4][k]  - adhere and cohere info
/        IBdata[4][1]  -  center of mass adhere dist
/        IBdata[4][2]  -  adhere gap
/        IBdata[4][3]  -  adhere stiffness
/        IBdata[4][4]  -  adhere break dist
/        IBdata[4][5]  -  adhere rest length
/        IBdata[4][6]  -  center of mass cohere dist
/        IBdata[4][7]  -  cohere gap
/        IBdata[4][8]  -  cohere stiffness
/        IBdata[4][9]  -  cohere break dist
/        IBdata[4][10] -  cohere rest length
/
/     IBdata[j][k]  -  entity information
/
/       for circles
/
/        IBdata[j][1]  -  x center
/        IBdata[j][2]  -  y center
/        IBdata[j][3]  -  radius
/        IBdata[j][4]  -  unused
/        IBdata[j][5]  -  unused
/        IBdata[j][6]  -  link stiffness
/        IBdata[j][7]  -  bend stiffness
/        IBdata[j][8]  -  tether stiffness
/        IBdata[j][9]  -  rest length
/        IBdata[j][10] -  rest curvature
/
/       for ellipse
/
/        IBdata[j][1]  -  x center
/        IBdata[j][2]  -  y center
/        IBdata[j][3]  -  major axis
/        IBdata[j][4]  -  minor axis
/        IBdata[j][5]  -  angular offset of major axis
/        IBdata[j][6]  -  link stiffness
/        IBdata[j][7]  -  bend stiffness
/        IBdata[j][8]  -  tether stiffness
/
/       for lines
/
/        IBdata[j][1]  -  x start point of segment
/        IBdata[j][2]  -  y start point of segment
/        IBdata[j][3]  -  x end point of segment
/        IBdata[j][4]  -  y end point of segment
/        IBdata[j][5]  -  unused
/        IBdata[j][6]  -  link stiffness
/        IBdata[j][7]  -  bend stiffness
/        IBdata[j][8]  -  tether stiffness
/        IBdata[j][9]  -  rest length
/        IBdata[j][10] -  rest curvature
/        IBdata[j][15] -  enpoint tether stiffness
/
/       for polygon
/
/        IBdata[j][1]  -  x start point of segment
/        IBdata[j][2]  -  y start point of segment
/        IBdata[j][3]  -  x end point of segment
/        IBdata[j][4]  -  y end point of segment
/        IBdata[j][5]  -  connect curvature between last and current seg.
/        IBdata[j][6]  -  link stiffness
/        IBdata[j][7]  -  bend stiffness
/        IBdata[j][8]  -  tether stiffness
/        IBdata[j][9]  -  rest length
/        IBdata[j][10] -  rest curvature
/        IBdata[j][11] -  rest curvature
/        IBdata[k][15] -  first to second rest length
/        IBdata[k][16] -  first to second spring strength
/        IBdata[k][17] -  first to second spring strength
/        IBdata[k][18] -  first bend strength
/        IBdata[k][19] -  first tether strength
/        IBdata[k][30] -  x coordinate of vector at start of seg
/        IBdata[k][31] -  y coordinate of vector at start of seg
/        IBdata[k][32] -  x coordinate of vector at end of seg
/        IBdata[k][33] -  y coordinate of vector at end of seg
/
/       for singlets
/
/        IBdata[j][1]  -  x coordinate
/        IBdata[j][2]  -  y coordinate
/        IBdata[j][3]  -  unused
/        IBdata[j][4]  -  unused
/        IBdata[j][5]  -  0.0d0
/        IBdata[j][6]  -  0.0d0
/        IBdata[j][7]  -  0.0d0
/        IBdata[j][8]  -  tether stiffness
/        IBdata[j][9]  -  0.0d0
/        IBdata[j][10] -  0.0d0
/
/       for sines
/
/        IBdata[j][1]  -  x start point of segment
/        IBdata[j][2]  -  y start point of segment
/        IBdata[j][3]  -  direction of the wave
/        IBdata[j][4]  -  wave length
/        IBdata[j][5]  -  amplitude
/        IBdata[j][6]  -  link stiffness
/        IBdata[j][7]  -  bend stiffness
/        IBdata[j][8]  -  tether stiffness
/        IBdata[j][9]  -  rest length
/        IBdata[j][10] -  rest curvature
/        IBdata[j][11] -  phase shift
/        IBdata[j][12] -  number of periods in wave
/        IBdata[j][15] -  endpoint tether stiffness
/
/    defal[][]  - default spring data - see the main program.
*/
