/* popt: gets options from command lines in ibpts */
int popt( int j, int k, char line[] )

{
  int i=0, l=0, err=0;
  float dr;
  char opt[10];
  extern int IBint[MAXLINE][MAXARG];
  extern int IBact[MAXLINE][4];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  IBint[k][0]   = 0;                 /* Name of the entity */
  IBint[k][2]   = 0;                 /* number of points */
  IBint[k][5]   = 0;                 /* activation status */
  IBdata[k][6]  = defal[j+1];        /* intra ibp link stiffness */
  IBdata[k][7]  = defal[j+2];        /* angle link stiffness */
  IBdata[k][8]  = defal[j+3];        /* tether spring stiffness */
  IBdata[k][9]  = defal[j+5];        /* ib point rest distance */
  IBdata[k][13] = defal[j+4];        /* rest angle */
  IBdata[k][14] = defal[j+5];        /* ib point rest distance */
  IBdata[k][17] = defal[j+4];        /* first rest angle */
  IBdata[k][19] = defal[j+3];        /* tether spring stiffness */
  IBact[k][1]   = 0;                 /* not an active element */
  

  while ((i=fdash(line)) >= 0)       /* parse the inputs as cases */
    {
      for (l=0; l<=MAXLINE - i; l++) line[l] = line[i+l];
      sscanf(line, "%s %e",opt, &dr);
      if ( err == 0 && opt[0] == '-' )
        {
          line[0] = ' ';
          switch (opt[1])     /* to add arguments, insert new cases */
            {
              case 'A':       /* active elements */
                IBact[1][1]++;
                IBact[k][1] = 1;
                IBint[k][0] = dr;
                break;
              case 'a':       /* rest angle */
                IBdata[k][13] = dr;
                break;
              case 'b':       /* bend stiffness */
                IBdata[k][7] = dr;
                break;
              case 'c':       /* first rest angle */
                IBdata[k][17] = dr;
                break;
              case 'd':       /* ibp initial dist */
                IBdata[k][9] = dr;
                break;
              case 'e':       /* endpoint tether stiffness */
                IBdata[k][19] = dr;
                break;
              case 'l':       /* rest distance */
                IBdata[k][14] = dr;
                break;
              case 'n':       /* number of points */
                IBint[k][2] = dr;
                break;
              case 'N':       /* Name of the entity */
                IBint[k][0] = dr;
                break;
              case 's':       /* strech stiffness */
                IBdata[k][6] = dr;
                break;
              case 't':       /* tether stiffness */
                IBdata[k][19] = dr;
                IBdata[k][8] = dr;
                break;
              case 'u':       /* activation status */
                IBint[k][5] = dr;
                if ( IBint[k][5] != 1 ) err = 1;
                break;
              case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
                break;
              default:
                err = 1;
            }
	}
    }

  IBdata[k][16] = IBdata[k][6];        /* first to second spring strength */
  IBdata[k][18] = IBdata[k][7];        /* first bend strength */

  if ( IBdata[k][9] < 0.0 ) err = 1;
  if ( IBdata[k][14] > 0.5 ) err = 1;
  IBint[k][5] = 2 - IBint[k][5];

  return err;
}

/* paopts: prints useage options on commands */
void paopts()

{
  printf ("    -a angle  = resting angle of springs \n");
  printf ("    -b bend   = bending springs stiffness \n");
  printf ("    -d distance > 0 = initial distance between ibp's");
  printf (" -- relative to h \n");
  printf ("    -l link   = intra ibp spring rest length");
  printf (" -- relative to h \n");
  printf ("    -s stiff  = intra ibp spring stiffness \n");
  printf ("    -t tether = tether spring stiffness \n");
  printf ("    -u act    = activation status \n");
}

/* pcircle: parse the circle statement, return status */
int pcircle ( char line[] )

{

  int err = 0, warn = 0, k = 0;
  float x0, y0, r, pi, th, h;
  char name[100];

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  IBint[1][3]++;                    /* increment count of objects */
  k = IBint[1][3];
  IBint[k][1] = 4;                  /* default ib point data */
  IBint[0][9]++;

  if ( wc(line) < 4 ) err = 1;      /* parse the arguments */
  else                              /* required arguments first */
    {
      sscanf(line, "%s %e %e %e", name, &x0, &y0, &r );

      IBdata[k][1] = x0;            /* x center */
      IBdata[k][2] = y0;            /* y center */
      IBdata[k][3] = r;             /* radius */

      if ( x0 < IBdata[1][1] ) err = 1;
      if ( x0 > IBdata[1][2] ) err = 1;
      if ( y0 < IBdata[1][3] ) err = 1;
      if ( y0 > IBdata[1][4] ) err = 1;
      if ( r < 0.0 ) err = 1;

      if ( x0 - r < IBdata[1][1] ) warn = 1;
      if ( x0 + r > IBdata[1][2] ) warn = 1;
      if ( y0 - r < IBdata[1][3] ) warn = 1;
      if ( y0 + r > IBdata[1][4] ) warn = 1;

      err = err + popt(10,k,line);   /* now the optional arguments */

      IBint[k][3] = 1;

    }

  if ( err == 0 )                   /* ibp parameters */
    {
       pi = 4.0*atan(1.0);
       if ( IBint[k][2] == 0 ){      /* number of ibp */
           h  = IBdata[k][14]*IBdata[2][6];
           th = 2.0*asin(h/(2.0*r));
           IBint[k][2] = (2.0*pi/th);
       }
       else
         {
           th = 2.0*pi/IBint[k][2];
           h  = 2.0*r*sin(th/2.0);
         }

                          /* resting length and local curvature */
       IBdata[k][9] = IBdata[k][9]*IBdata[2][6];
       if ( IBdata[k][13] == 0 ) 
         IBdata[k][10] = (h*h)*( sin(2.*th) - (2.*sin(th)) );
       else
         IBdata[k][10] = -h*h*sin(IBdata[k][13]);
    }       

  if ( err >= 1 )
    {
       printf ("\n*************** input error - line number %d \n",k);
       printf ("Usage: circle x0 y0 r [-options] \n");
       printf ("  (x0,y0) must be in the declared domain with r > 0\n");
       printf ("  [-options]\n");
       paopts();
    }

  if ( warn == 1 )
    {
       printf ("\n*************** input warning - line number %d \n",k);
       printf ("Usage: circle x0 y0 r [-options] \n");
       printf ("  a portion of the circle is outside the physical domain \n\n");
    }

  return err;
}

/* pellipse: parse the ellipse statement, return status */
int pellipse ( char line[] ){

  int err = 0, k = 0;
  float x0, y0, r1, r2, th, pi, dth, h, r;
  char name[100];

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  IBint[1][3]++;                    /* increment count of objects */
  k = IBint[1][3];
  IBint[k][1] = 14;                  /* default ib point data */
  IBint[0][9]++;
  IBint[k][3] = 1;

  pi = 4.0*atan(1.0);

  if ( wc(line) < 6 ) err = 1;      /* parse the arguments */
  else{                              /* required arguments first */

    sscanf(line, "%s %e %e %e %e %e", name, &x0, &y0, &r1, &r2, &th );

    if ( x0 < IBdata[1][1] ) err = 1;
    if ( x0 > IBdata[1][2] ) err = 1;
    if ( y0 < IBdata[1][3] ) err = 1;
    if ( y0 > IBdata[1][4] ) err = 1;
    if ( r2 < 0.0 ) err = 1;
    if ( r1 < 0.0 ) err = 1;
    if ( abs(th) > 90.0 ) err = 1;

    IBdata[k][1] = x0;            /* x center */
    IBdata[k][2] = y0;            /* y center */
    IBdata[k][3] = r1;            /* major axis */
    IBdata[k][4] = r2;            /* minor axis */
    IBdata[k][5] = th*pi/180.0;   /* angular offset */

    err = err + popt(10,k,line);   /* now the optional arguments */
  }

  if ( err == 0 ){                   /* ibp parameters */
       if ( IBint[k][2] == 0 ){      /* number of ibp */
         h  = IBdata[k][14]*IBdata[2][6];
         IBint[k][2] = 2.0*pi*sqrt( ((r1*r1) + (r2*r2))/2.0 )/h;
       }
  }

  if ( err >= 1 ){
    printf ("\n*************** input error - line number %d \n",k);
    printf ("Usage: ellipse x0 y0 a b theta [-options] \n");
    printf ("  (x0,y0) must be in the declared domain with a,b > 0\n");
    printf ("  (a,b) are the major and minor axes of the ellipse\n");
    printf ("  theta is the angular offset in degrees of the major axis\n");
    printf ("      and must satisfy -90 <= theta <= 90.\n");
    printf ("  [-options]\n");
    paopts();
  }

  return err;
}


/* pdomain: parse the domain statement, return status */
int pdomain ( char line[] )

{
  int err=0;
  float aspect, xmin, xmax, ymin, ymax;
  char name[100];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  if ( wc(line) != 5 )  err = 1;
  else 
    sscanf(line, "%s %e %e %e %e", name, &xmin, &xmax, &ymin, &ymax);

  if ( xmax <= xmin ) err = 1;
  if ( ymax <= ymin ) err = 1;

  if ( err == 0 )
    {
      IBint[1][1] = 1;
      IBdata[1][0] = 0;
      IBdata[1][1] = xmin;
      IBdata[1][2] = xmax;
      IBdata[1][3] = ymin;
      IBdata[1][4] = ymax;
      IBdata[3][2] = 0.0;
      IBdata[3][4] = ymax - ymin;
    }


  if ( err == 1 )
    {
      printf ("\n*********** input error - domain *****************\n");
      printf ("Usage: domain x_min x_max y_min y_max\n");
      printf ("  where x_min < x_max and y_min < y_max \n");
    }

  return err;
}

/* pfile: parses the file statement - treat like a polygon */
int pfile ( char line[], FILE *fpb )
{
  int i, j, junk, ndat, kdat, pflag =0, err=0;
  float x, y, xt, yt, s, r, b, a, t;
  float xn, yn, xnt, ynt, xot, yot, xo, yo;
  float xlt, ylt, xft, yft, xf,yf;
  float xaf,xab,yaf,yab;
  char filename[MAXLINE],name[MAXLINE],char_work[MAXLINE];
  FILE *fpi;

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  IBint[1][3]++;                     /* defaults for global parms */
  i = IBint[1][3];
  IBint[i][1] = 6;
  IBint[i][4] = 1;
  IBint[i+1][1] = 12;         
  IBdata[i][0]  = 0;
  IBint[0][10]++;
  IBint[1][3]++;                     /* defaults for global parms */

  if ( wc(line) == 1 ) err = 1;        /* parse the arguments */
  else{
    sscanf(line, "%s %s", name, filename );
    err = err + popt(60,i,line);       /* now the optional arguments */
  }
  sprintf(char_work,"pfile");
  if ( strncmp(name,char_work,5) == 0 ) pflag = 1;

  if ((fpi = fopen(filename,"r")) == NULL ) err = 2;
  if ( err == 0 ){
    fscanf(fpi,"%i %i",&ndat,&kdat);
    IBint[i+1][10] = 9;  
    IBint[i+1][2] = kdat;

    if (pflag == 1){
      if ( ndat >= 4 ){
        fscanf(fpi, "%e %e %e %e",&xf, &yf, &xft, &yft);
        for ( j=2; j<kdat; j++ ) getline(line,MAXLINE,fpi);
        fscanf(fpi, "%e %e %e %e",&x, &y, &xlt, &ylt);
      }
      if ( ndat == 2 ){
        fscanf(fpi, "%e %e", &xf, &yf);
        xft = xf; yft = yf;
        for ( j=2; j<kdat; j++ ) getline(line,MAXLINE,fpi);
        fscanf(fpi, "%e %e", &xlt, &ylt);
      }
      rewind(fpi);
      getline(line,MAXLINE,fpi);
    }

    if ( ndat == 9 ){
      for ( j=1; j<=kdat; j++ ){
        fscanf(fpi, "%e %e %e %e %e %e %e %e %e", 
                  &x, &y, &xt, &yt, &s, &r, &b, &a, &t);
        fprintf(fpb," %.13e %.13e %.13e %.13e %.13e",x,y,xt,yt,s);
        fprintf(fpb," %.13e %.13e %.13e %.13e\n",r,b,a,t);
      }
    }

    if ( ndat == 10 ){
      for ( j=1; j<=kdat; j++ ){
        fscanf(fpi, "%i %e %e %e %e %e %e %e %e %e", &junk,
                  &x, &y, &xt, &yt, &s, &r, &b, &a, &t);
        fprintf(fpb," %.13e %.13e %.13e %.13e %.13e",x,y,xt,yt,s);
        fprintf(fpb," %.13e %.13e %.13e %.13e\n",r,b,a,t);
      }
    }

    if ( ndat == 4 ){
      xot = xlt; yot = ylt;
      fscanf(fpi, "%e %e %e %e",&x, &y, &xt, &yt);
      fscanf(fpi, "%e %e %e %e",&xn, &yn, &xnt, &ynt);

      fprintf(fpb," %.13e %.13e %.13e %.13e",x,y,xt,yt);
      fprintf(fpb," %.13e",distance(xt,yt,xnt,ynt));
      fprintf(fpb," %.13e",IBdata[i][6]);
      if (pflag == 1){
        xaf = xnt-xt; yaf = ynt-yt; xab = xt-xot; yab = yt-yot;
        fprintf(fpb," %.13e",cross(xaf,yaf,xab,yab));
        fprintf(fpb," %.13e",IBdata[i][7]);
        fprintf(fpb," %.13e\n",IBdata[i][8]);
      }
      else{
        fprintf(fpb," 0.0 0.0");
        fprintf(fpb," %.13e\n",IBdata[i][19]);
      }

      for ( j=2; j<kdat; j++ ){
        xo = x; yo = y; xot = xt; yot = yt;
        x = xn; y = yn; xt = xnt; yt = ynt;
        fscanf(fpi, "%e %e %e %e",&xn, &yn, &xnt, &ynt);
        fprintf(fpb," %.13e %.13e %.13e %.13e",x,y,xt,yt);
        fprintf(fpb," %.13e",distance(xt,yt,xnt,ynt));
        fprintf(fpb," %.13e",IBdata[i][6]);
        xaf = xnt-xt; yaf = ynt-yt; xab = xt-xot; yab = yt-yot;
        fprintf(fpb," %.13e",cross(xaf,yaf,xab,yab));
        fprintf(fpb," %.13e",IBdata[i][7]);
        fprintf(fpb," %.13e\n",IBdata[i][8]);
      }

      xo = x; yo = y; xot = xt; yot = yt;
      x = xn; y = yn; xt = xnt; yt = ynt;
      xnt = xft; ynt = yft;
      fprintf(fpb," %.13e %.13e %.13e %.13e",x,y,xt,yt);
      if (pflag == 1){
        fprintf(fpb," %.13e",distance(xt,yt,xnt,ynt));
        fprintf(fpb," %.13e",IBdata[i][6]);
        xaf = xnt-xt; yaf = ynt-yt; xab = xt-xot; yab = yt-yot;
        fprintf(fpb," %.13e",cross(xaf,yaf,xab,yab));
        fprintf(fpb," %.13e",IBdata[i][7]);
        fprintf(fpb," %.13e\n",IBdata[i][8]);
      }
      else{
        fprintf(fpb," 0.0 0.0");
        fprintf(fpb," 0.0 0.0");
        fprintf(fpb," %.13e\n",IBdata[i][19]);
      }
    }

    if ( ndat == 2 ){

      xot = xlt; yot = ylt;
      fscanf(fpi, "%e %e",&x, &y);
      xt = x; yt = y;
      fscanf(fpi, "%e %e",&xn, &yn);
      xnt = xn; ynt = yn;

      fprintf(fpb," %.13e %.13e %.13e %.13e",x,y,xt,yt);
      fprintf(fpb," %.13e",distance(xt,yt,xnt,ynt));
      fprintf(fpb," %.13e",IBdata[i][6]);
      if (pflag == 1){
        xaf = xnt-xt; yaf = ynt-yt; xab = xt-xot; yab = yt-yot;
        fprintf(fpb," %.13e",cross(xaf,yaf,xab,yab));
        fprintf(fpb," %.13e",IBdata[i][7]);
        fprintf(fpb," %.13e\n",IBdata[i][8]);
      }
      else{
        fprintf(fpb," 0.0 0.0");
        fprintf(fpb," %.13e\n",IBdata[i][19]);
      }
      for ( j=2; j<kdat; j++ ){
        xo = x; yo = y; xot = xt; yot = yt;
        x = xn; y = yn; xt = xnt; yt = ynt;
        fscanf(fpi, "%e %e",&xn, &yn);
        xnt = xn; ynt = yn;
        fprintf(fpb," %.13e %.13e %.13e %.13e",x,y,xt,yt);
        fprintf(fpb," %.13e",distance(xt,yt,xnt,ynt));
        fprintf(fpb," %.13e",IBdata[i][6]);
        xaf = xnt-xt; yaf = ynt-yt; xab = xt-xot; yab = yt-yot;
        fprintf(fpb," %.13e",cross(xaf,yaf,xab,yab));
        fprintf(fpb," %.13e",IBdata[i][7]);
        fprintf(fpb," %.13e\n",IBdata[i][8]);
      }

      x = xn; y = yn; xt = xnt; yt = ynt;
      xnt = xft; ynt = yft;
      fprintf(fpb," %.13e %.13e %.13e %.13e",x,y,xt,yt);
      if (pflag == 1){
        fprintf(fpb," %.13e",distance(xt,yt,xnt,ynt));
        fprintf(fpb," %.13e",IBdata[i][6]);
        xaf = xnt-xt; yaf = ynt-yt; xab = xt-xot; yab = yt-yot;
        fprintf(fpb," %.13e",cross(xaf,yaf,xab,yab));
        fprintf(fpb," %.13e",IBdata[i][7]);
        fprintf(fpb," %.13e\n",IBdata[i][8]);
      }
      else{
        fprintf(fpb," 0.0 0.0");
        fprintf(fpb," 0.0 0.0");
        fprintf(fpb," %.13e\n",IBdata[i][19]);
      }
    }
  }

  IBint[i][2] = kdat;
  IBint[i+1][2] = kdat;
  IBint[i][3] = pflag;
  IBdata[i][1] = xf;
  IBdata[i][2] = yf;


  if ( err == 1 ){
    printf ("\n*********** input error - file *****************\n");
    printf ("Usage: file filename [-options]\n");
    printf ("  filename is a file containing immersed boundary data points \n");
    printf ("  and optional elastic link information.  See the IBIS program\n");
    printf ("  guide for details on the file structure and information.    \n");
    printf ("  [-options]\n");
    paopts();
  }

  return err;
}

/* pfluid: parse the fluid variables */
int pfluid ( char line[] )
{
  int err=0, i=0, j=0, icase=0, jcase=0;
  float dr;
  char name[100], base[100], opt[10], dash[10];

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern char fluidf[MAXLINE];

  while ((i=fdash(line)) >= 0)
    {

#if 0
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
#else
       (void)memmove(&line[0], &line[i], (MAXLINE - i + 1));
       line[MAXLINE-i] = '\0';
#endif
       sscanf(line, "%s %e",&opt[0], &dr);

       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'f':                /* mu -- viscosity */
                   sscanf(line, "%s %s",opt, base);
                   sprintf(dash,"'");
                   sprintf(fluidf,"%s%s%s",dash,base,dash);
                   break;
                case 'm':                /* mu -- viscosity */
                   IBdata[1][5] = dr;
                   break;
                case 'r':                /* rho -- density */
                   IBdata[1][6] = dr;
                   break;
                case 'u':                /* max initial velocity */
                   IBdata[2][7] = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                   break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 1 )
    {
       printf ("\n*************** input error in Fluid **************** \n");
       printf ("Usage: fluid [-options] \n");
       printf ("    -f fluid velocity file (default null)\n");
       printf ("    -m viscosity (default 1)\n");
       printf ("    -r density (default 1) \n");
       printf ("    -u max initial velocity (only with tubes) \n");
    }

  return err;
}

/* pbforce: parse the background force variables */
int pbforce ( char line[] )
{
  int err=0, i=0, j=0, icase=0, jcase=0;
  float dr;
  char name[100], base[100], opt[10];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'b':                /* x pressure gradient in tube */
                   if ( dr < IBdata[1][3] ) err = 1;
                   if ( dr > IBdata[1][4] ) err = 1;
                   IBdata[3][2] = dr - IBdata[1][3];
                   break;
                case 't':                /* x pressure gradient in tube */
                   if ( dr < IBdata[1][3] ) err = 1;
                   if ( dr > IBdata[1][4] ) err = 1;
                   IBdata[3][4] = dr - IBdata[1][3];
                   break;
                case 'x':                /* background x pressure gradient */
                   IBdata[2][8] = dr;
                   break;
                case 'y':                /* background y pressure gradient */
                   IBdata[2][9] = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                   break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 1 )
    {
       printf ("\n*************** input error in Fluid **************** \n");
       printf ("Usage: bforce [-options] \n");
       printf ("    -b y_b = apply x force from y_b to y_t \n");
       printf ("    -t y_t = apply x force from y_b to y_t \n");
       printf ("    -x background x pressure gradient \n");
       printf ("    -y background y pressure gradient \n");
    }

  return err;
}


/* pgrid: parse the grid statement, return status */
int pgrid ( char line[] )
{
  int err = 0, warn = 0, i = 1, j = 0, xgrid, ygrid, ntmax, ntstart;
  char name[100], opt[10];
  float dlt, dx, dy, yg;

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  ntstart = 0;

  if ( wc(line) < 3 ) err = 1;        /* parse the arguments */
  if ( wc(line) == 4 )
    {
      sscanf(line, "%s %d %e %d", name, &xgrid, &dlt, &ntmax );
      if ( ntmax < 0 ) err = 1;
    }
  else if ( wc(line) == 6 )
    {
      sscanf(line, "%s %d %e %d %s %d", name, &xgrid, &dlt, &ntmax, 
                  opt, &ntstart);
      if ( ntmax < 0 ) err = 1;
      if ( ntstart > ntmax ) err = 1;
    }
  else
    err = 1;

  if ( xgrid <= 0 ) err = 1;
  if ( xgrid >= 10 ) warn = 1;

  if ( err == 0 )                             /* load the data */
    {
      yg = (IBdata[1][4]-IBdata[1][3])/(IBdata[1][2]-IBdata[1][1]);
      yg = yg*pow(2,xgrid);
      ygrid = log(yg+0.001)/log(2.);
      IBint[2][1] = xgrid;
      IBint[2][2] = ygrid;
      IBint[1][7] = pow(2,xgrid);
      IBint[1][8] = pow(2,ygrid);
      IBdata[2][0] = 0;
      IBdata[2][1] = xgrid;
      IBdata[2][2] = ygrid;
      IBdata[2][3] = dlt;
      IBint[2][3] = ntmax;
      IBint[2][0] = ntstart;
      dx = (IBdata[1][2]-IBdata[1][1])/pow(2,IBdata[2][1]);
      dy = (IBdata[1][4]-IBdata[1][3])/pow(2,IBdata[2][2]);
      IBdata[2][4] = dx;
      IBdata[2][5] = dy;
      IBdata[2][6] = ( dx < dy ) ? dx : dy;       /* min xgrid,ygrid  */
    }

  if ( err == 1 )                              /* parsing error */
    {
      printf ("\n*********** input error - grid *****************\n");
      printf ("Usage: grid nx dlt ntmax -s ntstart \n");
      printf ("  where 2^nx is the number of grid points in x-direction\n");
      printf ("  dlt is the time step\n");
      printf ("  ntmax is the number of time steps ( >= 0 )\n");
      printf ("  ntstart is the first time ( < ntmax )\n");
    }

  if ( warn == 1 )
    {
      printf ("\n*********** input warning - grid *****************\n");
      printf ("Usage: grid nx dlt ntmax -s ntstart \n");
      printf ("  number of grid points exceeds 1024\n");
    }

  return err;
}

/* pline: parse the line statement, return status */
int pline ( char line[] )

{
  int err = 0, warn = 0, k = 0;
  float x0, y0, xn, yn, h;
  char name[100];

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  IBint[1][3]++;                    /* increment count of objects */
  k = IBint[1][3];
  IBint[k][1] = 2;                  /* default ib point data */
  IBint[k][3] = 0;                  /* open entity -- for graphics */

  if ( wc(line) < 4 ) err = 1;      /* parse the arguments */
  else                              /* required arguments first */
    {
      sscanf(line, "%s %e %e %e %e", name, &x0, &y0, &xn, &yn );

      IBdata[k][1] = x0;            /* x start point */
      IBdata[k][2] = y0;            /* y start point */
      IBdata[k][3] = xn;            /* x end point */
      IBdata[k][4] = yn;            /* y end point */

      if ( x0 < IBdata[1][1] ) err = 1;   /* is the line in the domain */
      if ( x0 > IBdata[1][2] ) err = 1;
      if ( y0 < IBdata[1][3] ) err = 1;
      if ( y0 > IBdata[1][4] ) err = 1;
      if ( xn < IBdata[1][1] ) err = 1;
      if ( xn > IBdata[1][2] ) err = 1;
      if ( yn < IBdata[1][3] ) err = 1;
      if ( yn > IBdata[1][4] ) err = 1;

                                     /* is the line periodic */
                                     /* chop this out if don't want */
      if ( x0 == IBdata[1][1] )
            if ( xn == IBdata[1][2] ) warn = 1;
      if ( xn == IBdata[1][1] )
            if ( x0 == IBdata[1][2] ) warn = 1;
      if ( y0 == IBdata[1][3] )
            if ( yn == IBdata[1][4] ) warn = 2;
      if ( yn == IBdata[1][3] )
            if ( y0 == IBdata[1][4] ) warn = 2;

      if ( warn == 1 ) if ( y0 == yn ) IBint[k][1] = 22;
      if ( warn == 2 ) if ( x0 == xn ) IBint[k][1] = 22;
      if ( IBint[k][1] == 22 )
        {
          IBint[0][8]++;            /* count as a filament if periodic */
        }
      else
        {
          IBint[0][10]++;           /* count as a polygon otherwise */
        }

      err = err + popt(70,k,line);   /* now the optional arguments */
    }

  if ( err == 0 )                   /* ibp parameters */
    {
       if ( IBint[k][2] == 0 )      /* number of ibp */
         {
           h  = IBdata[k][14]*IBdata[2][6];
           IBint[k][2] = (distance(x0,y0,xn,yn)/h)+1;
           h = distance(x0,y0,xn,yn)/(IBint[k][2]-1);
         }
       else
         {
           h = distance(x0,y0,xn,yn)/(IBint[k][2]-1);
         }

                          /* resting length and local curvature */
       IBdata[k][9] = h;
       IBdata[k][10] = 0.0;
       if ( IBint[k][1] == 22 ) IBint[3][2] = IBint[k][2]-1;
    }       

  if ( err == 1 )
    {
       printf ("\n*************** input error - line number %d \n",k);
       printf ("Usage: line x0 y0 xn yn [-options] \n");
       printf ("  (x0,y0), (xn,yn) must be in the declared domain \n");
       printf ("  [-options]\n");
       paopts();
    }
  if ( err == 2 )
    {
       printf ("\n*************** input warning - line number %d \n",k);
       printf ("line will be treated as periodic \n");
       printf ("  \n");
    }

  return err;
}

/* plink: parse the entity to entity link variables */
int plink ( char line[] )

{
  int err = 0, i, k = 0, l;
  int ent1=0, ent2=0, num=0, type=1;
  float rest, stiff, brake, dr;
  char name[100], opt[10];
  FILE  *fpl;

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  if ( wc(line) < 4 ) err = 1;      /* parse the arguments */
  else                              /* required arguments first */
    {
      sscanf(line, "%s %i %i %i", name, &ent1, &ent2, &num );
      if (ent1 <= 0) err=1;
      if (ent2 <= 0) err=1;
      if (num <= 0) err=1;
    }

  rest = defal[95];
  brake = defal[92];
  stiff = defal[91];

  while ((i=fdash(line)) >= 0)       /* parse the inputs as cases */
    {
      for (l=0; l<=MAXLINE - i; l++) line[l] = line[i+l];
      sscanf(line, "%s %e",opt, &dr);
      if ( err == 0 && opt[0] == '-' )
        {
          line[0] = ' ';
          switch (opt[1])     /* to add arguments, insert new cases */
            {
              case 'b':       /* breaking distance */
                brake = dr;
                break;
              case 'l':       /* rest distance */
                rest = dr;
                break;
              case 's':       /* strech stiffness */
                stiff = dr;
                break;
              case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
                break;
              default:
                err = 1;
	    }
	}
    }

  if ( err == 0 ) {
    IBint[1][9]++;
    fpl = fopen("ibisdat/link.dat","a");
    fprintf(fpl,"%i %i %i ",ent1, ent2, num);
    fprintf(fpl,"%e %e %e %i\n", stiff, brake, rest, type);
    fclose(fpl);
    IBint[0][12] = IBint[0][12] + num;
  }

  if ( err == 1 ) {
    printf ("\n*************** input error - link statement \n");
    printf ("Usage: link ent1 ent2 number [-options] \n");
    printf ("  -b breaking distance = (b*h)\n");
    printf ("  -r resting distance = (r*h)\n");
    printf ("  -s stiffness of the link \n");
    printf ("  ent1 > 0, ent2 > 0, num > 0 \n");
  }

  return err;

}
/* pblink: parse the entity to entity bridge variables */
int pblink ( char line[] )

{
  int err = 0, i, k = 0, l;
  int ent1=0, ent2=0, num=0, type,  ename, e2name;
  float rest, stiff, brake, dr;
  char name[100], opt[10];
  FILE  *fpl;

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  if ( wc(line) < 3 ) err = 1;      /* parse the arguments */
  else                              /* required arguments first */
    {
      sscanf(line, "%s %i %i", name, &ent1, &ent2 );
      if (ent1 <= 0) err=1;
      if (ent2 <= 0) err=1;
    }

  rest = defal[95];
  brake = defal[92];
  stiff = defal[91];

  type = 4;
  ename = getnum(ent1);
  e2name = getnum(ent2);
  if (IBint[ename][3] == 0) type = 5;
  num = 3*(min(IBint[ename][2],IBint[e2name][2]));
  if ( type == 5 ) num = num - 2;
  if ( IBint[ename][1] == 22 ) num = num - 3;


  if ( wc(line) > 3 ){
    while ((i=fdash(line)) >= 0){       /* parse the inputs as cases */
      for (l=0; l<=MAXLINE - i; l++) line[l] = line[i+l];
      sscanf(line, "%s %e",opt, &dr);
      if ( err == 0 && opt[0] == '-' ){
        line[0] = ' ';
        switch (opt[1]){     /* to add arguments, insert new cases */
          case 'b':       /* breaking distance */
            brake = dr;
            break;
          case 'l':       /* rest distance */
            rest = dr;
            break;
          case 's':       /* strech stiffness */
            stiff = dr;
            break;
          case '0': case '1': case '2': case '3': case '4':
          case '5': case '6': case '7': case '8': case '9':
            break;
          default:
            err = 1;
        }
      }
    }
  }

  if ( err == 0 ) {
    IBint[1][9]++;
    fpl = fopen("ibisdat/link.dat","a");
    fprintf(fpl,"%i %i %i ",ent1, ent2, num);
    fprintf(fpl,"%e %e %e %i\n", stiff, brake, rest, type);
    fclose(fpl);
    IBint[0][12] = IBint[0][12] + num;
  }

  if ( err == 1 ) {
    printf ("\n*************** input error - link statement \n");
    printf ("Usage: blink ent1 ent2 [-options] \n");
    printf ("  -b breaking distance = (b*h)\n");
    printf ("  -r resting distance = (r*h)\n");
    printf ("  -s stiffness of the link \n");
    printf ("  ent1 > 0, ent2 > 0 \n");
  }

  return err;

}

/* polink: parse the entity to entity connect variables */
int polink ( char line[] )

{
  int err = 0, i, k = 0, l;
  int ent1=0, ent2=0, num=0, type,  ename;
  float rest, stiff, brake, dr;
  char name[100], opt[10];
  FILE  *fpl;

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  if ( wc(line) < 3 ) err = 1;      /* parse the arguments */
  else                              /* required arguments first */
    {
      sscanf(line, "%s %i %i", name, &ent1, &ent2 );
      if (ent1 <= 0) err=1;
      if (ent2 <= 0) err=1;
    }

  rest = defal[95];
  brake = defal[92];
  stiff = defal[91];

  type = 2;
  ename = getnum(ent1);
  num = IBint[ename][2]-1;

  while ((i=fdash(line)) >= 0)       /* parse the inputs as cases */
    {
      for (l=0; l<=MAXLINE - i; l++) line[l] = line[i+l];
      sscanf(line, "%s %e",opt, &dr);
      if ( err == 0 && opt[0] == '-' )
        {
          line[0] = ' ';
          switch (opt[1])     /* to add arguments, insert new cases */
            {
              case 'b':       /* breaking distance */
                brake = dr;
                break;
              case 'l':       /* rest distance */
                rest = dr;
                break;
              case 's':       /* strech stiffness */
                stiff = dr;
                break;
              case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
                break;
              default:
                err = 1;
	    }
	}
    }

  if ( err == 0 ) {
    IBint[1][9]++;
    fpl = fopen("ibisdat/link.dat","a");
    fprintf(fpl,"%i %i %i ",ent1, ent2, num);
    fprintf(fpl,"%e %e %e %i\n", stiff, brake, rest, type);
    fclose(fpl);
    IBint[0][12] = IBint[0][12] + num;
  }

  if ( err == 1 ) {
    printf ("\n*************** input error - link statement \n");
    printf ("Usage: olink ent1 ent2 [-options] \n");
    printf ("  -b breaking distance = (b*h)\n");
    printf ("  -r resting distance = (r*h)\n");
    printf ("  -s stiffness of the link \n");
    printf ("  ent1 > 0, ent2 > 0 \n");
  }

  return err;

}

/* pslink: parse the entity to entity slink variables */
int pslink ( char line[] )

{
  int err = 0, i, k = 0, l;
  int ent1=0, ent2=0, num=0, type,  ename, eflag=1;
  float rest, stiff, brake, dr;
  char name[100], opt[10];
  FILE  *fpl;

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  if ( wc(line) < 3 ) err = 1;      /* parse the arguments */
  else                              /* required arguments first */
    {
      sscanf(line, "%s %i %i", name, &ent1, &ent2 );
      if (ent1 <= 0) err=1;
      if (ent2 <= 0) err=1;
    }

  rest = defal[95];
  brake = defal[92];
  stiff = defal[91];

  ename = getnum(ent1);
  if ( IBint[ename][1] == 5 ) eflag = 0;

  ename = getnum(ent2);
  if ( IBint[ename][2] == 1 ) {
    eflag = 0;
    ename = ent1;
    ent1 = ent2;
    ent2 = ename;
  }
  err = err + eflag;

  ename = getnum(ent2);
  num = IBint[ename][2];
  type = 3;

  while ((i=fdash(line)) >= 0)       /* parse the inputs as cases */
    {
      for (l=0; l<=MAXLINE - i; l++) line[l] = line[i+l];
      sscanf(line, "%s %e",opt, &dr);
      if ( err == 0 && opt[0] == '-' )
        {
          line[0] = ' ';
          switch (opt[1])     /* to add arguments, insert new cases */
            {
              case 'b':       /* breaking distance */
                brake = dr;
                break;
              case 'l':       /* rest distance */
                rest = dr;
                break;
              case 's':       /* strech stiffness */
                stiff = dr;
                break;
              case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
                break;
              default:
                err = 1;
	    }
	}
    }

  if ( err == 0 ) {
    IBint[1][9]++;
    fpl = fopen("ibisdat/link.dat","a");
    fprintf(fpl,"%i %i %i ",ent1, ent2, num);
    fprintf(fpl,"%e %e %e %i\n", stiff, brake, rest, type);
    fclose(fpl);
    IBint[0][12] = IBint[0][12] + num;
  }

  if ( err == 1 ) {
    printf ("\n*************** input error - link statement \n");
    printf ("Usage: slink ent1 ent2 [-options] \n");
    printf ("  -b breaking distance = (b*h)\n");
    printf ("  -r resting distance = (r*h)\n");
    printf ("  -s stiffness of the link \n");
    printf ("  ent1 > 0, ent2 > 0 \n");
    printf ("  ent1 or ent2 must have only one ib point \n");
  }

  return err;

}

/* ptube: parse the tube statement, return status 
/    currently the tube is assumed to be horizontal */

int ptube ( char line[] )

{
  int err=0, i=0, j=0;
  float xtr, ytr, xbl, ybl, dist, dr;
  char name[100], opt[10];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  IBint[3][1] = 3;                 /* default ib point data */
  IBint[0][8]++;                   /* count as two filaments */
  IBint[0][8]++;

  if ( wc(line) < 5 ) err = 1;       /* parse the arguments */
  else
    {
      sscanf(line,"%s %e %e %e %e", name, &xbl, &ybl, &xtr, &ytr );

      IBdata[3][1] = xbl;            /* bottom left coordinate - x */
      IBdata[3][2] = ybl;            /* bottom left coordinate - y */
      IBdata[3][3] = xtr;            /* top right coordinate - x */
      IBdata[3][4] = ytr;            /* top right coordinate - y */

      if ( IBdata[3][3] <= IBdata[3][1] ) err = 1;
      if ( IBdata[3][4] <= IBdata[3][2] ) err = 1;

      err = err + popt(30,3,line);
    }

  IBint[3][2] = (IBdata[3][3]-IBdata[3][1])/(IBdata[3][14]*IBdata[2][6]);
  IBdata[3][9] = IBdata[3][9]*IBdata[2][6];
  if ( IBdata[3][13] > 0 )
      IBdata[3][10] = -IBdata[3][9]*IBdata[3][9]*sin(IBdata[3][13]);

  if ( err == 1 )
    {
       printf ("\n*************** input error in tube **************** \n");
       printf ("Usage: tube x_bl y_bl x_tr y_tr [-options] \n");
       printf ("  x_bl,x_tr,y_bl,y_tr specify corners of the tube\n"); 
       printf ("    and, x_bl < x_tr, y_bl < y_tr \n"); 
       printf ("  [-options]\n");
       paopts();
    }

  return err;

}

/* pitube: parse the injured tube statement, return status */
int pitube ( char line[] )

{
  int err=0, i=0;
  float in11, in12, iwall;
  char name[100];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  if ( wc(line) != 4 ) err = 1;
  else
    sscanf(line, "%s %e %e %e", name, &iwall, &in11, &in12);

  if ( err == 0 )
    {
      i = 2*(iwall-1);
      IBdata[1][7+i] = in11;
      IBdata[1][8+i] = in12;
    }

  if ( err == 1 )
    {
      printf ("\n*********** input error - itube *****************\n");
      printf ("Usage: itube w x1 x2\n");
      printf ("  where w is the wall (1 or 2)\n");
      printf ("  x1 < x2 are the start and end points of the injury.\n");
      printf ("  and x1,x2 are between 0 and 1, relative to wall length.\n");
    }

  return err;
}

/* piwall: parse the injured wall statement, return status */
int piwall ( char line[] )

{
  int err=0, i=0;
  float in11, in12, iwall;
  char name[100];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  if ( wc(line) != 4 ) err = 1;
  else
    sscanf(line, "%s %e %e", name, &in11, &in12);

  if ( err == 0 ){

      if ( IBdata[1][7] == 0 ){
        IBdata[1][7] = in11;
        IBdata[1][8] = in12;
      }
      else if ( IBdata[1][9] == 0 ){
        IBdata[1][9] = in11;
        IBdata[1][10] = in12;
      }
      else {
        err = 2;
      }
  }

  if ( err == 1 )
    {
      printf ("\n*********** input error - iwall *****************\n");
      printf ("Usage: iwall x1 x2\n");
      printf ("  x1 < x2 are the start and end points of the injury.\n");
      printf ("  and x1,x2 are between 0 and 1, relative to wall length.\n");
    }
  if ( err == 2 )
    {
      printf ("\n*********** input error - iwall *****************\n");
      printf ("  Only two injured walls allowed.\n");
    }

  return err;
}

/* psing: parse the isolated point statement, return status */
int psing ( char line[] )
{
  int err = 0, j = 0, k = 0, l=0;
  float x0, y0, dr;
  char name[100], opt[10];

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  IBint[1][3]++;                    /* increment count of objects */
  k = IBint[1][3];
  IBint[k][1] = 5;                  /* default ib point data */
  IBint[k][2] = 1;                  /* default ib point data */

  IBint[0][10]++;                   /* count as a polygon */

  if ( wc(line) < 3 ) err = 1;
  else                              /* get the arguments */
    {
      sscanf(line, "%s %e %e", name, &x0, &y0 );
      if ( x0 < IBdata[1][1] ) err = 1;
      if ( x0 > IBdata[1][2] ) err = 1;
      if ( y0 < IBdata[1][3] ) err = 1;
      if ( y0 > IBdata[1][4] ) err = 1;
      IBdata[k][1] = x0;
      IBdata[k][2] = y0;
    }

  err = err + popt(20,k,line);   /* now the optional arguments */

  IBint[k][3] = 1;
  if ( IBdata[k][5] != 0.0 ) err = 2;
  if ( IBdata[k][6] != 0.0 ) err = 2;
  if ( IBdata[k][7] != 0.0 ) err = 2;
  if ( IBdata[k][9] != 0.0 ) err = 2;
  if ( IBdata[k][10] != 0.0 ) err = 2;

  if ( err == 1 )
    {
       printf ("\n*************** input error - line number %d \n",k);
       printf ("Usage: singlet x y [-t tether] \n");
       printf ("  (x,y) must be in the declared domain\n");
    }

  if ( err == 2 )
    {
       printf ("\n*************** input error - line number %d \n",k);
       printf ("Usage: singlet x y [-t tether] \n");
       printf ("  only tethering is allowed for singlets\n");
    }

  return err;
}

/* pspring: parse the spring variables */
int pspring ( char line[] )

{
  int err=0, i=0, j=0, icase=0, jcase=0;
  float dr;
  char name[100], opt[10];
  extern float defal[MAXLINE];

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  else
    {
      sscanf(line, "%s %s", name, opt);  /* which spring variables? */
      switch (opt[0])
        {
           case 'A':                /* all */
             icase = 0;
             break;
           case 'b':                /* bcurv and carc */
             icase = 4;
             break;
           case 'c':                /* circle */
             icase = 1;
             break;
           case 'f':                /* file */
             icase = 6;
             break;
           case 'l':                /* line */
             icase = 7;
             break;
           case 'L':                /* link */
             icase = 9;
             break;
           case 'r':                /* ray */
             icase = 5;
             break;
           case 's':                /* singlet */
             icase = 2;
             break;
           case 'S':                /* singlet */
             icase = 8;
             break;
           case 't':                /* tube */
             icase = 3;
             break;
           default:
             icase = 1;
             err = 1;
        }
    }

  icase = icase*10;
  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'a':                /* resting angle */
                   defal[icase+4] = dr;
                   break;
                case 'b':                /* bending stiffness */
                   defal[icase+2] = dr;
                   break;
                case 'l':                /* resting link length */
                   defal[icase+5] = dr;
                   break;
                case 's':                /* link stretching stiffness */
                   defal[icase+1] = dr;
                   break;
                case 't':                /* tether stiffness */
                   defal[icase+3] = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( icase == 0 )
    {
     for (icase=1; icase<=9; icase++) 
       {
          jcase = 10*icase;
          defal[jcase+1] = defal[1];
          defal[jcase+2] = defal[2];
          defal[jcase+3] = defal[3];
          defal[jcase+4] = defal[4];
          defal[jcase+5] = defal[5];
       }
    }

  if ( icase == 20 )
    {
     for (icase=4; icase<=5; icase++) 
       {
          jcase = 10*icase;
          defal[jcase+1] = defal[21];
          defal[jcase+2] = defal[22];
          defal[jcase+3] = defal[23];
          defal[jcase+4] = defal[24];
          defal[jcase+5] = defal[25];
       }
    }


  if ( err == 1 )
    {
       printf ("\n*************** input error in spring **************** \n");
       printf ("Usage: spring k [-options] \n");
       printf ("  k one argument from {A,b,c,f,l,L,r,s,S,t} \n");
       printf ("    for all, arc, circle, file, line, link, ray, singlet, ");
       printf ("sine, tube\n");
       printf ("  [-options]\n");
       printf ("    -a default resting angle \n");
       printf ("    -b default bending stiffness \n");
       printf ("    -d default intra ibp spring rest length \n");
       printf ("    -s default intra ibp spring stiffness \n");
       printf ("    -t default tether spring stiffness \n");
    }

  return err;
}

/* pvisual: parse the visualization variables */
int pvisual ( char line[] )

{
  int err=0, i=0, j=0, icase=0, jcase=0, pflag=0, cflag=0, Pflag=0;
  float dr;
  char name[100], base[100], cmap[100], path[100], opt[10], dash[10];
  extern int IBint[MAXLINE][MAXARG];
  extern char colorf[MAXLINE],linef[MAXLINE],partf[MAXLINE];
  FILE  *fpv;

  IBint[1][4] = 10;              /* default output frequency */
  IBint[0][0] = 2;               /* default particle density in y */
  IBint[0][1] = 2;               /* default particle density in x */
  IBint[0][2] = 2*IBint[1][7];   /* default file width */
  IBint[0][3] = 2*IBint[1][8];   /* default file height */
  IBint[0][4] = 1;               /* default zoom */
  IBint[0][5] = 1;               /* default increment */
  IBint[0][6] = 1;               /* default first file */
  IBint[3][3] = 0;               /* default periodic particles */
  IBint[3][4] = 0;               /* default periodic particles */
  sprintf(base,"image");
  sprintf(cmap,"ibisdat/cmap");

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'b':                /* base name */
                   sscanf(line, "%s %s",opt, base);
                   break;
                case 'c':                /* cmap name */
                   sscanf(line, "%s %s",opt, cmap);
                   break;
                case 'C':                /* compression */
                   cflag = 1;
                   break;
                case 'h':                /* height of file in pixels */
                   IBint[0][3] = dr;
                   break;
                case 'i':                /* movie increment */
                   IBint[0][5] = dr;
                   break;
                case 'n':                /* output frequency */
                   IBint[1][4] = dr;
                   break;
                case 'p':                /* path name */
                   pflag = 1;
                   sscanf(line, "%s %s",opt, path);
                   break;
                case 'w':                /* width of file in pixels */
                   IBint[0][2] = dr;
                   break;
                case 'z':                /* zoom factor */
                   IBint[0][4] = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 0 )
    {
      fpv = fopen("ibisdat/visual.dat","a");
      if ( pflag == 1 ) fprintf(fpv,"path %s \n",path);
      fprintf(fpv,"base %s \n",base);
      fprintf(fpv,"height %i \n",IBint[0][3]);
      fprintf(fpv,"width %i \n",IBint[0][2]);
      fprintf(fpv,"color %s \n",cmap);
      fprintf(fpv,"zoom %i \n",IBint[0][4]);
      fprintf(fpv,"increment %i \n",IBint[0][5]);
      fprintf(fpv,"first 0 \n");
      fprintf(fpv,"last 0 \n");
      if ( cflag == 0 ) fprintf(fpv,"ccompress yes\n",path);
      if ( cflag == 1 ) fprintf(fpv,"ccompress no\n",path);
      fclose(fpv);

      sprintf(dash,"'");
      sprintf(colorf,"%s%s/%s%s",dash,path,base,dash);
    }

  if ( err == 1 )
    {
       printf ("\n*************** input error in visual **************** \n");
       printf ("Usage: visual [-options] \n");
       printf ("  [-options]\n");
       printf ("    -b base file name (default image)\n");
       printf ("    -c color map file name (default ibisview/cmap)\n");
       printf ("    -C compress (gzip) the color files \n");
       printf ("    -f first file in movie (default 1)\n");
       printf ("    -h graphics file height (default 2*ny)\n");
       printf ("    -i movie frame increment (default 1)\n");
       printf ("    -n graphics output frequency (default 10) \n");
       printf ("    -p path name for output (default ./) \n");
       printf ("    -w graphics file width (default 2*nx)\n");
       printf ("    -z zoom factor (default 1)\n");
    }

  return err;
}
/* pparticle: parse the particle variables */
int pparticle ( char line[] )

{
  int err=0, i=0, j=0;
  char opt[10], dash[10], partname[MAXLINE];
  float dr;
  extern int IBint[MAXLINE][MAXARG];
  extern char partf[MAXLINE];
  FILE  *fpv;

  IBdata[2][11] = 0.0;

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'd':                /* fluid marker density in x and y*/
                   IBint[0][1] = dr;
                   IBint[0][0] = dr;
                   break;
                case 'P':                /* particle file name */
                   sscanf(line, "%s %s",opt, partname);
                   sprintf(dash,"'");
                   sprintf(partf,"%s%s%s",dash,partname,dash);
                   break;
                case 'r':                /* enable particle random walk */
                   IBdata[2][11] = dr;
                   break;
                case 'x':                /* particles in x */
                   IBint[3][3] = 1;
                   break;
                case 'X':                /* fluid marker density in x*/
                   IBint[0][1] = dr;
                   break;
                case 'y':                /* particles in y */
                   IBint[3][4] = 1;
                   break;
                case 'Y':                /* fluid marker density in y*/
                   IBint[0][0] = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 1 )
    {
       printf ("\n*************** input error in visual **************** \n");
       printf ("Usage: particle [-options] \n");
       printf ("  [-options]\n");
       printf ("    -d fluid marker density (default 2)\n");
       printf ("    -P particle file name \n");
       printf ("    -r max. dist. of random walk (default 0.0) \n");
       printf ("    -x nuemann particles in x\n");
       printf ("    -X particle density in x (default 2)\n");
       printf ("    -y nuemann particles in y\n");
       printf ("    -Y particle density in y (default 2)\n");
    }

  return err;
}

/* pgraph: parse the graph variables */
int pgraph ( char line[] )
{
  int err=0, i=0, j=0, icase=0, jcase=0, pflag=0, cflag=0;
  float dr;
  char name[100], base[100], cmap[100], path[100], opt[10], dash[10];
  extern char colorf[MAXLINE],linef[MAXLINE];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  FILE  *fpv;

  IBint[1][5] = 500;             /* default output frequency */
  IBint[1][6] = 1;               /* default vector output density */
  sprintf(base,"linep");
  IBint[2][5] = 2;               /* default no force output */
  IBint[2][6] = 1;               /* default velocity output */

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'b':                /* base name */
                   sscanf(line, "%s %s",opt, base);
                   break;
                case 'C':                /* compression */
                   cflag = 1;
                   break;
                case 'd':                /* output density */
                   IBint[1][6] = dr;
                   break;
                case 'f':                /* output force */
                   IBint[2][5] = 1;
                   break;
                case 'n':                /* output frequency */
                   IBint[1][5] = dr;
                   break;
                case 'p':                /* path name */
                   pflag = 1;
                   sscanf(line, "%s %s",opt, path);
                   break;
                case 'u':                /* output velocity */
                   IBint[2][6] = 2;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 0 )
    {
      fpv = fopen("ibisdat/visual.dat","a");
      if ( pflag == 1 ) fprintf(fpv,"lpth %s \n",path);
      fprintf(fpv,"lbse %s \n",base);
      fprintf(fpv,"llst 0 \n");
      if ( cflag == 0 ) fprintf(fpv,"lcompress yes\n",path);
      if ( cflag == 1 ) fprintf(fpv,"lcompress no\n",path);
      fprintf(fpv,"xmin %e \n", IBdata[1][1]);
      fprintf(fpv,"ymin %e \n", IBdata[1][3]);
      fprintf(fpv,"xmax %e \n", IBdata[1][2]);
      fprintf(fpv,"ymax %e \n", IBdata[1][4]);
      fprintf(fpv,"xlength %e \n", IBdata[1][2]-IBdata[1][1]);
      fprintf(fpv,"ylength %e \n", IBdata[1][4]-IBdata[1][3]);
      fprintf(fpv,"nfvector 2\n");
      fprintf(fpv,"fvector1 velocity\n");
      fprintf(fpv,"fvmax1 0.0001\n");
      fprintf(fpv,"fvmin1 0.0\n");
      fprintf(fpv,"fvector2 force\n");
      fprintf(fpv,"fvmax2 0.0001\n");
      fprintf(fpv,"fvmin2 0.0\n");
      fprintf(fpv,"npvector 2\n");
      fprintf(fpv,"pvector1 velocity\n");
      fprintf(fpv,"pvmax1 0.0001\n");
      fprintf(fpv,"pvmin1 0.0\n");
      fprintf(fpv,"pvector2 force\n");
      fprintf(fpv,"pvmax2 0.0001\n");
      fprintf(fpv,"pvmin2 0.0\n");
      fclose(fpv);
      sprintf(dash,"'");
      sprintf(linef,"%s%s/%s%s",dash,path,base,dash);
    }

  if ( err == 1 )
    {
       printf ("\n*************** input error in visual **************** \n");
       printf ("Usage: graph [-options] \n");
       printf ("  [-options]\n");
       printf ("    -b base file name (default linep)\n");
       printf ("    -C compress (gzip) the line files \n");
       printf ("    -d vector output density (default 1)\n");
       printf ("    -f output force vectors (default off)\n");
       printf ("    -n graphics output frequency (default 500) \n");
       printf ("    -p path name for output (default ./) \n");
       printf ("    -u output velocity vectors (default on)\n");
    }

  return err;
}



/* padhere: parse the adhere statement, return status */
int padhere ( char line[] )

{
  int err=0, i, j;
  float stiffa, sadh, sagap, sbreaka, rlinka, dr;
  char name[100], opt[10];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  sadh    = 1.7;
  sagap   = 0.8;
  stiffa  = 0.0;
  sbreaka = 8.0;
  rlinka  = 0.0;

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'h':
                   sadh = dr;
                   break;
                case 'g':
                   sagap = dr;
                   break;
                case 's':
                   stiffa = dr;
                   break;
                case 'b':
                   sbreaka = dr;
                   break;
                case 'r':
                   rlinka = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 0 )
    {
         IBdata[4][1] = sadh;
         IBdata[4][2] = sagap;
         IBdata[4][3] = stiffa;
         IBdata[4][4] = sbreaka;
         IBdata[4][5] = rlinka;
      
    }


  if ( err == 1 )
    {
      printf ("\n*********** input error - adhere *****************\n");
      printf ("Usage: adhere -b break -h center_dist -g gap -r rest -s stiff\n");
      printf ("  break -- breaking distance of adhesive link \n");
      printf ("  center_dist -- max distance of mass center to wall \n");
      printf ("  gap -- max distance of ib point to wall \n");
      printf ("  rest -- resting length of adhesive link \n");
      printf ("  stiff -- adhesive link stiffness \n");
      printf ("See the IBIS user guide for scaling these parameters \n");
    }

  return err;
}


/* pcohere: parse the cohere statement, return status */
int pcohere ( char line[] )

{
  int err=0, i, j;
  float stiffc, scoh, scgap, sbreakc, rlinkc, dr;
  char name[100], opt[10];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  scoh    = 1.7;
  scgap   = 0.8;
  stiffc  = 0.0;
  sbreakc = 8.0;
  rlinkc  = 0.0;

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'h':
                   scoh = dr;
                   break;
                case 'g':
                   scgap = dr;
                   break;
                case 's':
                   stiffc = dr;
                   break;
                case 'b':
                   sbreakc = dr;
                   break;
                case 'r':
                   rlinkc = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 0 )
    {
         IBdata[4][6] = scoh;
         IBdata[4][7] = scgap;
         IBdata[4][8] = stiffc;
         IBdata[4][9] = sbreakc;
         IBdata[4][10] = rlinkc;
      
    }


  if ( err == 1 )
    {
      printf ("\n*********** input error - cohere *****************\n");
      printf ("Usage: cohere -b break -h center_dist -g gap -r rest -s stiff\n");
      printf ("  break -- breaking distance of cohesive link \n");
      printf ("  center_dist -- max distance of mass center to center \n");
      printf ("  gap -- max distance of ib point to center \n");
      printf ("  rest -- resting length of cohesive link \n");
      printf ("  stiff -- cohesive link stiffness \n");
      printf ("See the IBIS user guide for scaling these parameters \n");
    }

  return err;
}


/* pplatelet: parse the adhere statement, return status */
int pplatelet ( char line[] )

{
  int maxlink, iplat, err=0, nring, i, j;
  float pltdif,pdens,dr;
  char name[100], opt[10];
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  iplat = 1;
  maxlink = 3;
  pltdif = 0.001;
  pdens  = 20.0;
  nring  = 20;

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0)
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'm':
                   maxlink = dr;
                   break;
                case 'd':
                   pltdif = dr;
                   break;
                case 'p':
                   pdens = dr;
                   break;
                case 'n':
                   nring = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                   break;
                default:
                   err = 1;
	      }
	  }
    }

  if ( err == 0 )
    {
         IBdata[3][5] = pltdif;
         IBdata[2][10] = pdens;
         IBint[4][1] = maxlink;
         IBint[4][2] = iplat;
         IBint[4][3] = nring;
      
    }


  if ( err == 1 )
    {
      printf ("\n*********** input error - platelet *****************\n");
      printf ("Usage: platelet -m maxlink -d pltdif -p pdens -n nring\n");
      printf ("  maxlink -- max number links per platelet \n");
      printf ("  pltdif -- platelet diffusion constant \n");
      printf ("  pdens -- platelet density \n");
      printf ("  nring -- number of ib points in a platelet ring \n");
    }

  return err;
}

/* psine: parse the sine statement, return status */
int psine ( char line[], FILE *fp )

{
  int err = 0, warn = 0, k = 0, i, j;
  float x0, y0, num_period, direction, period_length;
  float dist, amplitude, pshift, h, dr;
  char name[100], arg[100], opt[10];

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  IBint[1][3]++;                    /* increment count of objects */
  k = IBint[1][3];
  IBint[k][1] = 32;                 /* default ib point data */
  IBint[k][3] = 0;                  /* open entity -- for graphics */

  err = err + popt(80,k,line);      /* get the link arguments */

  direction = 0.0;
  period_length = 2.0*3.141592653;
  num_period = 1.0;
  amplitude = 0.0;
  pshift = 0.0;

  getcommand(line,MAXLINE,fp);      /* get the position arguments */ 
  while ((i=fdash(line)) >= 0)      /* parse the arguments */
    {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %e",opt, &dr);
       if ( err == 0 && opt[0] == '-' )
          {
            line[0] = ' ';
            switch (opt[1])              /* to add arguments, insert cases */
              {
                case 'x':                /* x start point */
                   x0 = dr;
                   break;
                case 'y':                /* y start point */
                   y0 = dr;
                   break;
                case 'd':                /* direction of the wave */
                   direction = dr;
                   break;
                case 'p':                /* period of the wavelength */
                   period_length = dr;
                   break;
                case 'l':                /* length in units of the period */
                   num_period = dr;
                   break;
                case 'a':                /* wave amplitude */
                   amplitude = dr;
                   break;
                case 'h':                /* phase shift at the first point */
                   pshift = dr;
                   break;
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    break;
                default:
                   err = 1;
	      }
	  }
    }


  IBdata[k][1] = x0;            /* x start point */
  IBdata[k][2] = y0;            /* y start point */
  IBdata[k][3] = direction;
  IBdata[k][5] = amplitude;
  IBdata[k][11] = pshift;
  IBdata[k][12] = num_period;

  if ( x0 < IBdata[1][1] ) err = 1;   /* is the line in the domain */
  if ( x0 > IBdata[1][2] ) err = 1;
  if ( y0 < IBdata[1][3] ) err = 1;
  if ( y0 > IBdata[1][4] ) err = 1;

  period_length = 2.0*3.141592653/period_length;  
  dist = num_period*period_length;
  IBdata[k][4] = period_length;
  IBdata[k][13] = dist;

  if ( err == 0 )                   /* ibp parameters */
    {
       if ( IBint[k][2] == 0 )      /* number of ibp */
         {
           h  = IBdata[k][14]*IBdata[2][6];
           IBint[k][2] = (dist/h)+1;
           h = dist/(IBint[k][2]-1);
         }
       else
         {
           h = dist/(IBint[k][2]-1);
         }

                          /* resting length and local curvature */
       IBdata[k][9] = h;
       IBdata[k][10] = 0.0;
    }       

  if ( err == 1 )
    {
       printf ("\n*************** input error - line number %d \n",k);
       printf ("Usage: sine [-options] \n");
       printf ("Usage:    -x x_start -y y_start -d direction -p period_length");
       printf ("-l num_period -a amplitude -h phase_shift \n");
       printf ("  (x,h) must be in the declared domain \n");
       printf ("[-options]\n");
       paopts();
    }
  return err;
}

/* psource: parse the source statement, return status */
int psource ( char line[] ){

  int dr, err=0, i, nsource = 0, j;
  char name[100], opt[10];
  extern int IBint[MAXLINE][MAXARG];

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0) {
       for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
       sscanf(line, "%s %i",opt, &dr);
       if ( err == 0 && opt[0] == '-' ) {
            line[0] = ' ';
            switch (opt[1]){
              case 'n':
                 nsource = dr;
                 break;
              case '0': case '1': case '2': case '3': case '4':
              case '5': case '6': case '7': case '8': case '9':
                  break;
              default:
                 err = 1;
	    }
       }
  }

  if ( err == 0 ) IBint[3][5] = nsource;

  if ( err == 1 ){
    printf ("\n*********** input error - source *****************\n");
    printf ("Usage: source -n nsource\n");
    printf ("  nsource -- number of fluid source/sink points \n");
  }

  return err;
}


/* punits: parse the units statement, return status */
int punits ( char line[] ){

  int err=0, i, j, tflag=0, uflag=0, nflag=1;
  float char_length=1.0, char_time=1.0, char_U=1.0, dr;
  char name[100], opt[10];

  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];

  if ( wc(line) < 2 ) err = 1;           /* parse the arguments */
  while ((i=fdash(line)) >= 0){
    for (j=0; j<=MAXLINE - i; j++) line[j] = line[i+j];
    sscanf(line, "%s %e",opt, &dr);
    if ( err == 0 && opt[0] == '-' ){
      line[0] = ' ';
      switch (opt[1]){              /* to add arguments, insert cases */
        case 'l':
          char_length = dr;
          break;
        case 't':
          tflag = 1;
          nflag = 0;
          char_time = dr;
          break;
        case 'u':
          uflag = 1;
          nflag = 0;
          char_U = dr;
          break;
        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
          break;
        default:
          err = 1;
      }
    }
  }

  if (tflag*uflag == 1) err = 1;
  if (char_length <= 0.0) err = 1;
  if (char_time <= 0.0) err = 1;
  if (char_U <= 0.0) err = 1;

  if ( err == 0 ){
    IBdata[1][11] = char_length;
    if ( nflag == 1 ) IBdata[1][12] = 1.0;
    if ( tflag == 1 ) IBdata[1][12] = char_time;
    if ( uflag == 1 ) IBdata[1][12] = char_length/char_U;
  }

  if ( err == 1 ){
    printf ("\n*********** input error - adhere *****************\n");
    printf ("Usage: units -l char_length -t char_time -u char_U\n");
    printf ("  char_length -- characteristic length of the simulation\n");
    printf ("  char_time -- characteristic time of the simulation\n");
    printf ("  char_U -- characteristic velocity of the simulation\n");
    printf ("Either char_time or char_U may be included but not both.\n");
    printf ("All characteristic constants must be positive.\n");
    printf ("See the IBIS user guide for scaling these parameters.\n");
  }

  return err;
}

/* pwheel: parse the wheel statement, return status */
int pwheel ( char line[] )

{
  int err = 0, warn = 0, k = 0, l;
  float x0, y0, r, pi, th, h;
  char name[100], sline[1000];
  FILE  *fpa;

  extern int unused_names;
  extern int IBint[MAXLINE][MAXARG];
  extern float IBdata[MAXLINE][MAXARG];
  extern float defal[MAXLINE];

  IBint[1][3]++;                    /* increment count of objects */
  k = IBint[1][3];
  IBint[k][1] = 4;                  /* default ib point data */
  IBint[k][0] = 0;

  IBint[0][9]++;                    /* count as a plat */

  if ( wc(line) < 4 ) err = 1;      /* parse the arguments */
  else                              /* required arguments first */
    {
      sscanf(line, "%s %e %e %e", name, &x0, &y0, &r );

      IBdata[k][1] = x0;            /* x center */
      IBdata[k][2] = y0;            /* y center */
      IBdata[k][3] = r;             /* radius */

      if ( x0 < IBdata[1][1] ) err = 1;
      if ( x0 > IBdata[1][2] ) err = 1;
      if ( y0 < IBdata[1][3] ) err = 1;
      if ( y0 > IBdata[1][4] ) err = 1;
      if ( r < 0.0 ) err = 1;

      if ( x0 - r < IBdata[1][1] ) warn = 1;
      if ( x0 + r > IBdata[1][2] ) warn = 1;
      if ( y0 - r < IBdata[1][3] ) warn = 1;
      if ( y0 + r > IBdata[1][4] ) warn = 1;

      err = err + popt(10,k,line);   /* now the optional arguments */

      IBint[k][3] = 1;

    }

  if ( err == 0 )                   /* ibp parameters */
    {
       pi = 4.0*atan(1.0);
       if ( IBint[k][2] == 0 )      /* number of ibp */
         {
           h  = IBdata[k][14]*IBdata[2][6];
           th = 2.0*asin(h/(2.0*r));
           IBint[k][2] = (2.0*pi/th);
         }
       else
         {
           th = 2.0*pi/IBint[k][2];
           h  = 2.0*r*sin(th/2.0);
         }

                          /* resting length and local curvature */
       IBdata[k][9] = IBdata[k][9]*IBdata[2][6];
       if ( IBdata[k][13] == 0 ) 
         IBdata[k][10] = (h*h)*( sin(2.*th) - (2.*sin(th)) );
       else
         IBdata[k][10] = -h*h*sin(IBdata[k][13]);
    }       


  if ( IBint[k][0] == 0 ) IBint[k][0] = ++IBint[0][11];

  for (l=0; l<100; l++) line[l] = ' ';
  IBint[0][11]++;  

  sprintf(sline,"singlet %f %f -t %f -N %i \n",x0,y0,IBdata[k][8],
         IBint[0][11]);
  err = err + psing(sline);
  sprintf(sline,"link %i %i\n", IBint[0][11], IBint[k][0]);
  err = err + pslink(sline);

  if ( err >= 1 )
    {
       printf ("\n*************** input error - line number %d \n",k);
       printf ("Usage: wheel x0 y0 r [-options] \n");
       printf ("  (x0,y0) must be in the declared domain with r > 0\n");
       printf ("  [-options]\n");
       paopts();
    }

  if ( warn == 1 )
    {
       printf ("\n*************** input warning - line number %d \n",k);
       printf ("Usage: wheel x0 y0 r [-options] \n");
       printf ("  a portion of the wheel is outside the physical domain \n\n");
    }

  return err;
}
