/*
 *  epaste.c
 *  This program displays a single output file
 *  from IBIS immersed boundary simulations.
 *
 *  Author:     David J. Eyre
 *              Department of Mathematics
 *              University of Utah
 *  Date:       November 12, 1996
 *
 *  Copyright @ 1996, David J. Eyre
 *
 */
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/file.h>
#include "aux.h"
#include <fcntl.h>

#define MAX 100
#define MAXC 1310720

char file_name[MAX], color_map[MAX];
unsigned char image[MAXC];

static GLint height = 128;
static GLint skip_bytes = 4;
static GLint width = 256;
static GLfloat zoom = 1.0;

void display(void){

    extern unsigned char image[MAXC];
    glDrawPixels(width,height,GL_COLOR_INDEX,GL_UNSIGNED_BYTE,image);
    glFlush();

}

void getInData (int argc, char** argv)
{
    extern int optind;
    extern char *optarg;
    extern char file_name[MAX], color_map[MAX];

    int i, err=0;

    sprintf(file_name,"image");
    sprintf(color_map,"cmap");
/*  
 *  Parse the input arguments
 */
    if (argc == 1) err = 1;
    while ((i = getopt(argc,argv,"b:c:f:h:i:l:s:w:z:")) != EOF) {
      switch (i) {
        case 'c':
          sprintf(color_map,optarg);
          break;
        case 'h':
          height = atoi(optarg);
          break;
        case 'i':
          sprintf(file_name,optarg);
          break;
        case 's':
          skip_bytes = atoi(optarg);
          break;
        case 'w':
          width = atoi(optarg);
          break;
        case 'z':
          zoom = atof(optarg);
          break;
        default:
          err = 1;
      }
    }
/*  
 *  Oops, there was a boo-boo
 */
    if (err == 1) {
      printf("Usage: epaste [-c color_map] [-h height] [-w width] ");
      printf("[-i file_name] [-s skip_bytes] [-z zoom]\n");
      printf("  Defaults: -h 128 -w 256 -z 1.0 -c cmap -s 4\n");
      printf("  To exit type q in the display window\n");
      exit(0);
    }
}

void idle (void)
{
    auxIdleFunc(0);
}

/*  Initialize the color map and background color
 */

void myinit (void)
{
    extern char file_name[MAX], color_map[MAX];
    extern unsigned char image[MAXC];

    FILE *fd;
    int i, err, fp;
    float r,g,b;

/*  
 *  get the colormap
 */

    fd = fopen(color_map,"r");
    for (i=0; i<=255; i++){
       fscanf(fd, "%e %e %e", &r, &g, &b);
       auxSetOneColor(i,r/255.,g/255.,b/255.);
    }
    glClearIndex(0);
    glClear( GL_COLOR_BUFFER_BIT );

/*  
 *  Read the data file
 */

    if ((fp=open(file_name,O_RDONLY,0)) == -1) {
       printf(" Can't find %s\n",file_name);
       exit(0);
    }
    read(fp, image, skip_bytes);
    if ( (err=read(fp, image, width*height)) < width*height ) {
       printf(" %s does not contain enough data\n",file_name);
       exit(0);
    }
    close(fp);

    glPixelZoom(zoom,zoom);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
}
 
/*  Ignore reshape events - just redisplay data.
 */

void myReshape(int w, int h) { display(); }

void exFunc(void){ exit(0); }

/*  Main Program
 *  Get the input data from the command line.
 *  Open window with initial window size and title bar.
 *  Use double buffer and color index display mode. 
 *  Handle input events.  Loop indefinitely.
 */

int main(int argc, char** argv) {
    getInData (argc,argv);
    auxInitDisplayMode (AUX_SINGLE | AUX_INDEX);
    auxInitPosition (0, 0, width*zoom, height*zoom);
    auxInitWindow (argv[0]);
    myinit ();
    auxReshapeFunc (myReshape);
    auxIdleFunc (idle);
    auxKeyFunc (AUX_q, exFunc);
    auxKeyFunc (AUX_Q, exFunc);
    auxMainLoop (display);
}
