/*
 *  ibistoppm.c - Convert IBIS output file to portable pixel map
 *
 *  Author:     David J. Eyre
 *              Department of Mathematics
 *              University of Utah
 *  Date:       December 9, 1996
 *
 *  Copyright @ 1996, David J. Eyre
 *
 */
#include <stdio.h>
#define MAXLINE 100
#define MAX 1500000

void main(int argc, char** argv)
{
  FILE *fd;
  char cmf[MAXLINE];
  int r[256], g[256], b[256], color, j;
  int red, green, blue;
  unsigned char cimage[3*MAX];

  int err = 0, i, skip_bytes = 4, h, w;
  extern int optind;
  extern char *optarg;

/* 
 * parse the input
 */
  sprintf(cmf,"cmap");
  if (argc == 1) err = 1;
  while ((i = getopt(argc,argv,"c:h:s:w:")) != EOF) {
    switch (i)
      {
        case 'c':
          sprintf(cmf,optarg);
          break;
        case 'h':
          h = atoi(optarg);
          break;
        case 's':
          skip_bytes = atoi(optarg);
          break;
        case 'w':
          w = atoi(optarg);
          break;
        default:
          err = 1;
      }
  }
/*  
 *  Error condition -- print usage
 */
  if (err == 1) 
   {
    fprintf(stderr,"Usage: ibistoppm -h height -w width -c colormap\n");
    fprintf(stderr,"   [-s skip_bytes (def=4)] < infile > outfile \n");
    exit(0);
   }

/*  
 *  Get the color map
 */
  fd = fopen(cmf,"r");
  for (i=0; i<=255; i++)
    {
       fscanf(fd, "%i %i %i", &red, &green, &blue);
       r[i] = red;   g[i] = green;   b[i] = blue;
    }
  fclose(fd);

/*  
 *  Read the data file
 */

  for (i=1; i<=skip_bytes; i++) {
     getc(stdin);
  }

  j=0;
  for (i=1; i<=h*w; i++) {
    color = getc(stdin);
    cimage[j++] = r[color];
    cimage[j++] = g[color];
    cimage[j++] = b[color];
    }

  fprintf(stdout, "P6\n");
  fprintf(stdout, "%i  %i\n", w, h);
  fprintf(stdout, "255\n");
  fwrite(cimage, h*w*3, 1, stdout);

  exit(0);
}

