/* 
 * ibiszoom.c - Read an IBIS output file and write a zoom file.
 * 
 * Author:      David J. Eyre
 *              Department of Mathematics
 *              University of Utah
 * Date:        Dec. 4, 1996
 *
 * Copyright @ 1996, David J. Eyre
 *
 */
#include <stdio.h>
#include <fcntl.h>

#define MAX 1048576
#define MAXLINE 100

void main(int argc, char** argv)
{
  unsigned char image[MAX], zimage[MAX];
  char fnm[MAXLINE];
  int i, j, k=0, err=0, xl, xr, yl, yr, fp, l, m;
  int skip_bytes=4, width=256, height=128;
  int out_skip_bytes=4;

  extern int optind;
  extern char *optarg;

/* 
 * parse the input
 */
 if (argc == 1) err = 1;
 while ((i = getopt(argc,argv,"h:i:s:t:w:u:v:x:y:")) != EOF) 
   {
    switch (i)
      {
        case 'h':
          height = atoi(optarg);
          break;
        case 'i':
          sprintf(fnm,optarg);
          break;
        case 's':
          skip_bytes = atoi(optarg);
          break;
        case 't':
          out_skip_bytes = atoi(optarg);
          break;
        case 'w':
          width = atoi(optarg);
          break;
        case 'u':
          xl = atoi(optarg);
          break;
        case 'v':
          yl = atoi(optarg);
          break;
        case 'x':
          xr = atoi(optarg);
          break;
        case 'y':
          yr = atoi(optarg);
          break;
        default:
          err = 1;
      }
   }

/* 
 * outta here if the input is bad
 */
  if (err == 1) 
   {
     fprintf(stderr,"Usage: ibiszoom -h height -w width ");
     fprintf(stderr,"-u bottom_left_x -v bottom_left_y \n");
     fprintf(stderr,"   -x top_right_x -y top_right_y ");
     fprintf(stderr,"[-s skip_bytes] [-t out_skip_bytes] ");
     fprintf(stderr,"-i infile > outfile \n");
     exit(0);
   }

/* 
 * deal with the image
 */
  if ((fp=open(fnm,O_RDONLY,0)) != -1){
     if ( skip_bytes>0 ) read(fp, image, skip_bytes);
     read(fp, image, width*height);
     for ( i=yl; i<yr; i++) {
       for(j=xl; j<xr; j++) {
        l = ((i-1)*width)+j;
        zimage[k++] = image[l];
       }
     }
  }

/* 
 * write the IBIS file
 */

  fwrite ( image, out_skip_bytes, 1, stdout );
  fwrite ( zimage, k, 1, stdout );

  exit(0);
 }
